pro firs_1565_obs,plt = plt,cont_wl = cont_wl,v2q = v2q, v2u = v2u,beta = beta,obsindex = obsindex,gainindex = gainindex,scan_range = scan_range,verbose = verbose,angle = angle,darkindex = darkindex,cal_file = cal_file,change_save = change_save,tel_cal = tel_cal,input_dir = input_dir,output_dir = output_dir,check_lines = check_lines,ccal_file = ccal_file,flat_file = flat_file,swap = swap,tmat_save = tmat_save,ooutfile = ooutfile,wavelength = wavelength,notelcor = notelcor,pos_lines = pos_lines,offang = offang,no_align = no_align, quit_column = quit_column,hair_up = hair_up, hair_down = hair_down,check_hairlines = check_hairlines,average_cal = average_cal,catch_dead = catch_dead,smooth_width = smooth_width,disp_scale = disp_scale,straight = straight,nopause=nopause,noxmat = noxmat,help = help

;+
;============================================================================
;
;	procedure : firs_1565_obs  
;
;	purpose   : reduce FIRS IR (1083 or 1565 nm) data
;
;	author    : cbeck@nso Apr 2013
;
;==============================================================================
;
;	Check number of arguments.
;
;==============================================================================
if n_elements(help) ne 0 then begin
	print
	print, "usage:   firs_1565_obs,plt = plt,cont_wl = cont_wl,v2q = v2q, v2u = v2u,beta = beta,obsindex = obsindex,gainindex = gainindex,scan_range = scan_range,verbose = verbose,angle = angle,darkindex = darkindex,cal_file = cal_file,change_save = change_save,tel_cal = tel_cal,input_dir = input_dir,output_dir = output_dir,check_lines = check_lines,ccal_file = ccal_file,flat_file = flat_file,swap = swap,tmat_save = tmat_save,ooutfile = ooutfile,wavelength = wavelength,notelcor = notelcor,pos_lines = pos_lines,offang = offang ,no_align = no_align, quit_column = quit_column,hair_up = hair_up, hair_down = hair_down,check_hairlines = check_hairlines,average_cal = average_cal,catch_dead = catch_dead,smooth_width = smooth_width,disp_scale = disp_scale,help = help"
	print
	print, "	 Reduce FIRS data of the IR channel."
        print, "         "
	print
	print, "	Arguments"
        print, "                none"
        print
        print, "         Keywords :"
        print, "              /plt : switch on plotting (recommended)"
        print, "              cont_wl : fltarr(2), [l0,l1] : wavelength pixel"
        print, "                        range to use for I->QUV cross-talk"
        print, "                        correction. If not provided, it will"
        print, "                        be set to the range between the two"
        print, "                        solar lines (1082.7/1083)"
        print, "              v2q, v2u: fltarr(1) each. Additional cross-talk"
        print, "                        correction coefficient for V->Q, V->U."
        print, "                        Q = Q - v2q * V"      
        print, "              beta    : parasitic light correction;" 
        print, "                        I = I-beta*<I>. Can remove false "
        print, "                        polarisation signal in the telluric lines"
        print, "              obsindex: number of observation file to be reduced"
        print, "                        in the source directory."
        print, "              gainindex:number of gain table to be used for the"
        print, "                        reduction in the target directory"
        print, "              scan_range: fltarr(2), [s0,s1]; scan steps to be"
        print, "                        reduced. If not provided, all steps"
        print, "                        will be reduced."
        print, "              /verbose : switch to activate extensive verbal"
        print, "                        output on the terminal."
        print, "              angle   : fltarr(1). Rotation angle for the beams."
        print, "                        Overrules value determined in gain table"
        print, "                        calculation."
        print, "              darkindex:number of dark current to be used for the"
        print, "                        reduction in the target directory"
        print, "              cal_file: number of calibration file to be used"
        print, "                        in the source directory."
        print, "              change_save: directory where the file information"
        print, "                        on the source directory was saved to."
        print, "              /tel_cal :  switch to de-activate telescope correction"
        print, "                        and I -> QUV correction."
        print, "              input_dir: source directory with raw data."
        print, "              output_dir: destination directory for reduced data."
        print, "              /check_lines: switch to enforce the manual selection"
        print, "                         of the spectral lines."
        print, "              ccal_file: use a different response function."
        print, "                         String with the complete filepath and"
        print, "                         filename of the xmat....sav file."
        print, "              flat_file: use a different gain table."
        print, "                         String with the complete filepath and"
        print, "                         filename of the gain.....sav file."
        print, "               /swap   : do a swap endian of the raw data."
        print, "               tmat_save:  String with the complete filepath and"
        print, "                         filename of the T-matrix parameter file."
        print, "               ooutfile: use different filename for output file."
        print, "                         Must include file path."
        print, "               wavelength: wavelength of data in AA"
        print, "               /notelcor: do not apply telescope correction." 
        print, "               pos_lines:fltarr(2), location of spectral lines in pixel. "
        print, "               offang: offset angle between telescope and instrument calibration."
        print, "               /no_align: do not use hairline for spatial alignment."
        print, "               /quit_column:switch to replace missing column at"
        print, "                         middle of IR CCD chip. Cosmetics."
        print, "               hair_up: fltarr(2), upper and lower range in pixels"
        print, "                        around position of hairline."
        print, "               hair_down: fltarr(2), upper and lower range in pixels"
        print, "                        around position of 2nd hairline."
        print, "                        Ignore, currently not used."
        print, "               /check_hairlines: switch to enforce the manual selection"
        print, "                         of the hair lines location."
        print, "               /average_cal: use only the polarimetric response"
        print, "                         from upper half of the spectrum"
        print, "                         (beam vignetting in calibration unit)"
        print, "               /catch_dead: try to remove dead pixels"
        print, "               smooth_width: fltarr(1). Width in pixels along"
        print, "                        the slit over which will be smoothed"
        print, "                        to remove spikes from flat field residuals."
        print, "               disp_scale: fltarr(1), change size of display"
        print, "                       windows on small screens, standard size"
        print, "                       is divided by / disp_scale"
        print, "               /help   : switch to active the help text above."
        print
return
endif
;==============================================================================
;-

; close all open file units to be on the safe side
close,/all

; load B/W color table
loadct,0,/silent

; smoothing width in pixels along slit to catch spikes
if n_elements(smooth_width) eq 0 then smooth_width = 1
; number of dark current to be used
if n_elements(darkindex) eq 0 then  darkindex = 0
; number of calibration file to be used
if n_elements(cal_file) eq 0 then  cal_file = 0
; number of gain table to be used
if n_elements(gainindex) eq 0 then gainindex = 0
; number of file to be reduced
if n_elements(obsindex) eq 0 then obsindex= 0 
; filename of save file with telescope parameters
if n_elements(tmat_save) eq 0 then tmat_save = '/home/cbeck/laptop_teneriffa/telcal_DST/Tmatrix_May2010_4708-14125.idl'
; scaling variable for display windows
if n_elements(disp_scale) ne 0 then disp_scale = float(disp_scale)

; filepath for SIR files, best ignore
if n_elements(filepath) eq 0 then filepath ='/export/cbeck/sir_collection/sir_gfpi_1/'

; restore save file with information on source directory, must be the
; first save file in the directory
if n_elements(change_save) ne 0 then erg = read_directory(change_save,filter ='*.sav') else erg = read_directory(input_dir,filter ='*.sav')

; check if file exists
outfile = erg.files
if n_elements(change_save) eq 0 then begin
   gfile = input_dir+outfile(0)  
   erg = check_existence(gfile,/read,/change)
   restore,erg(0)  
endif else begin
   gfile = change_save+outfile(0)
   erg = check_existence(gfile,/read,/change)
   restore,erg(0)  
endelse

; 16 cal   2 map  1 ff 0 dc

; get dark current file
ergdc = read_directory(output_dir,filter ='*dc*')
erg = check_existence(output_dir + ergdc.files(darkindex),/read,/change)
print,'Restoring dark current...',erg(0)
restore,erg(0)

; get gain table file
if n_elements(flat_file) eq 0 then begin
   erggain = read_directory(output_dir,filter ='*gain*')
   erg = check_existence(output_dir + erggain.files(gainindex),/read,/change)
   print,'Restoring gain table...',erg(0)
   restore,erg(0)
endif else begin
   erg = check_existence(flat_file,/read,/change)
   print,'Restoring gain table...',erg(0)
   restore,erg(0)
endelse

; get telescope matrix parameters
if n_elements(tel_cal) eq 0 and n_elements(notelcor) eq 0 then begin 
   erg = check_existence(tmat_save,/read,/change)
   print,'Restoring telescope parameters...',erg(0)
   restore,erg(0)
   tparams = tt
   tparams(3) = 90.
   if n_elements(offang) ne 0 then tparams(3) = offang
endif

; overrule rotation angle from flat field if value is provided
if n_elements(angle) ne 0 then aangle = aangle - aangle + angle

; spectral offset between the beams, used to get the spectral lines at
; the same place in all beams. Taken from gain table save file.
posref = mean(pposref)
oofset = fltarr(n_slits)
for k = 0,n_slits - 1 do begin & oofset(k) = mean(pposref(k,*,*)) - mean(pposref(0,*,*)) & endfor


; get instrument calibration file
calbasefile = bbasefile1(cal_file)  
outfile1 = output_dir+'xmat.'+calbasefile+'sav'
if n_elements(ccal_file) eq 0 then begin
   erg = check_existence(outfile1,/read,/change)
   print, 'Restoring instrument calibration...',erg(0)
   restore,erg(0)
endif else begin
   erg = check_existence(ccal_file,/read,/change)
   print, 'Restoring instrument calibration ',erg(0)
   restore,erg(0)
endelse


print,'Extrapolating X-matrix to full slit...'
; normalize and invert instrument response function X
xxinv = xxmat - xxmat
for kk = 0,n_separations -1 do begin
    for k = 0,n_slits-1 do begin
       if n_elements(average_cal) eq 0 then xmat = reform(xxmat(kk,k,*,*)) else  xmat = total(reform(xxmat(3:n_separations-1,k,*,*)),1)/(n_separations-3.) 
       dmod=invert(xmat,/double)
       xxmat(kk,k,*,*) = xmat
       xxinv(kk,k,*,*) = dmod
    endfor
 endfor

ppos = fltarr(n_separations)
for k = 0,n_separations-1 do begin & ppos(k) = (yyr(k)+yyr(k+1))/2. & endfor
full_xmat = fltarr(n_slits,dy_final+1,4,4)
for k = 0,n_slits-1 do begin & xmats = reform(xxmat(*,k,*,*)) & erg = extrapolate_xmat(ppos,xmats,dy_final+1,plt=0) & full_xmat(k,*,*,*) = erg & endfor
full_xxinv = fltarr(n_slits,dy_final+1,4,4)
for k = 0,n_slits-1 do begin & xmats = reform(xxinv(*,k,*,*)) & erg = extrapolate_xmat(ppos,xmats,dy_final+1,plt=0) & full_xxinv(k,*,*,*) = erg & endfor

;for k = 0,dy_final do begin & full_xxinv(0,k,*,*) = [ [1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1] ] & endfor


; get input raw data file
obsbasefile = bbasefile1(obsindex)
ergobs = read_directory(input_dir,filter ='*'+obsbasefile+'*')
obsfiles = ergobs.files

erg = check_existence(input_dir+obsfiles(0),/read)
if erg(1) ne 1 then begin
   print,'!firs_1565_obs.pro: cannot read observation file ',input_dir+obsfiles(0)
   print,'Please check input file.'
   print,'Aborting.'
   return
endif

; check number of steps available
nsteps = n_elements(obsfiles)

; Search for repeated scans. Attention: file name convention in FIRS
; supposedly has changed over the years.
if nsteps gt 1 then begin
   maxrep = fltarr(nsteps)
   for k = 0,nsteps-1 do begin & eerg = string2num(obsfiles(k),divider ='.') & maxrep(k) = eerg(n_elements(eerg)-1) & endfor

      maxrep = max(maxrep)

      if maxrep ne 0 then begin
         
         obsfiles = strarr(nsteps)

         iiindex = 0.

         for k = 0,maxrep do begin & iindex = k & if iindex lt 10 then maxreptext = '.000'+num2string(fix(iindex)) &  if iindex ge 10 and iindex lt 100 then maxreptext = '.00'+num2string(fix(iindex))  &  if iindex ge 100 and iindex lt 1000 then maxreptext = '.0'+num2string(fix(iindex)) &  if iindex ge 1000 then maxreptext = '.'+num2string(fix(iindex))  & ergobs = read_directory(input_dir,filter ='*'+obsbasefile+'*'+maxreptext) & ffiless = ergobs.files & obsfiles(iiindex:iiindex+n_elements(ffiless)-1) = ffiless & iiindex = iiindex+n_elements(ffiless) & endfor

      endif
   endif

; name of reduced data file
if n_elements(ooutfile) eq 0 then outfile = output_dir +obsbasefile+'reduced.dat' else outfile = ooutfile

erg = check_existence(outfile,/write,/change)
outfile = erg(0)

; file of overview maps
outfile2 = outfile+'.sav'

; set scan range to full size or use explicit scan range
if n_elements(scan_range) eq 0 then scan_range = [0,nsteps-1]

; number of steps to be reduced
nsteps_total = scan_range(1)-scan_range(0)+1.

print,'Reducing ',nsteps_total,' scan steps from ', scan_range(0),' to ', scan_range(1)

; storage arrays for line positions and image shifts
sshifts = fltarr(n_slits,nsteps_total,2)
llinepos = fltarr(n_slits,nsteps_total,2)

; 2D overview maps I_c, |QUV|
ic = fltarr(nsteps_total*n_slits,dy_final+1)
totq = fltarr(nsteps_total*n_slits,dy_final+1)
totu = fltarr(nsteps_total*n_slits,dy_final+1)
totv = fltarr(nsteps_total*n_slits,dy_final+1)

; maps for I -> QUV cross-talk
ii2q =  fltarr(nsteps_total*n_slits,dy_final+1)
ii2u =  fltarr(nsteps_total*n_slits,dy_final+1)
ii2v =  fltarr(nsteps_total*n_slits,dy_final+1)

; variable for storing time of the scan step
ttime = fltarr(nsteps_total)

; variables for storing the variables needed for the telescope model
ttel_azimuth= fltarr(nsteps_total)
ttel_elevation= fltarr(nsteps_total)
ttable_angle= fltarr(nsteps_total)

; determine screen size
ssize = get_screen_size()

; do plotting
if n_elements(plt) ne 0 then begin

   if dx_final*2 gt ssize(0) or dy_final*2 gt ssize(1) then img_scale_disp = 2. else  img_scale_disp = 1.

   if n_elements(disp_scale) ne 0 then img_scale_disp = disp_scale

   if dx_final lt 800 then begin
   for k = 0,n_slits -1 do begin & window,k,xs = 4*dx_final/img_scale_disp,ys = 500,title = 'Stokes spectra slit '+num2string(k) & endfor
   endif else begin
   for k = 0,n_slits -1 do begin & window,k,xs = 2*dx_final/img_scale_disp,ys = 2*500/img_scale_disp,title = 'Stokes spectra slit '+num2string(k) & endfor
   endelse
   if  n_slits*nsteps_total*2 gt 1000 then xxssize = 1000 else xxssize =  n_slits*nsteps_total*2
   window,4,xs = xxssize,ys = 600,title = '2D maps'
endif

; switches
erg = 0.
erg1 = 0.

; for spectroscopic data: int_only = 1, is determined from image size
; later on
int_only = 0.

print,'Starting reduction...'
print,'Writing to output file ',outfile

; open output file
openw,outunit,outfile,/get_lun

; loop over scan steps
for k = scan_range(0),scan_range(1) do begin

; get system time for estimate of duration
   oldt = systime(1)

   error_status = 0
; read in data
   ffxread,input_dir+obsfiles(k),dum,hdr,error_status = error_status

   if error_status eq 0 and k gt scan_range(0) then begin
      print,'! firs_1565_obs: file error at step ',k,' file: ',obsfiles(k)
      print,' Replacing data with previous scan step.'
      dum = dumold
   endif
   
   if error_status eq 1 then begin
      dumold = dum
   endif
   
; do swap endian if needed
 if n_elements(swap) ne 0 then dum = swap_endian(dum)

; check size of image in raw data. If only spectroscopy, blow up to
; QUV == 0
 ee = size(dum)
 if k eq scan_range(0) then print,'Dimensions of data: ',ee
 
 if ee(3) eq 8 then begin
    dum(*,*,0) = .5*( dum(*,*,0)+dum(*,*,4))
    dum(*,*,1) = .5*( dum(*,*,1)+dum(*,*,5))
    dum(*,*,2) = .5*( dum(*,*,2)+dum(*,*,6))
    dum(*,*,3) = .5*( dum(*,*,3)+dum(*,*,7))
    dum = dum(*,*,0:3)
 endif
 
 if ee(0) eq 2 then begin
    ddum = fltarr(ee(1),ee(2),4)
    ddum(*,*,0) = dum
    ddum(*,*,1) = dum
    ddum(*,*,2) = dum
    ddum(*,*,3) = dum
    dum = ddum
    int_only = 1.
 endif

; remove missing column(s)
 if n_elements(quit_column) ne 0 then begin
 for kkk = 0.,3 do begin & dum(510+kkk,*,*) = dum(509,*,*)+kkk/3.*(dum(514,*,*)-dum(509,*,*)) &  endfor 
endif


yyear = string2num(fxpar(hdr,'DATE_OBS'),divider = '/')

if yyear(0) eq -1 then yyear = 2015 


if n_elements(bad_pix) ne 0 and yyear(0) ge 2015 then begin

ddum = dum

for kkk = 0,3 do begin & ddum(*,*,kkk) =  ddum(*,*,kkk) -  dc & endfor

fff = where(bad_pix eq 1)
;medwidth = 3.

for kkk = 0.,n_elements(fff)-1 do begin 
   xx0 = fff(kkk) mod ee(1)
   yy0 = fff(kkk) / ee(1)

   xxmin = xx0 - medwidth
   yymin = yy0 - medwidth
   xxmax = xx0+ medwidth
   yymax = yy0+ medwidth

   if xxmin lt 0 then begin
      xxmin = 0
      xxmax = 2* medwidth+1
   endif
   if yymin lt 0 then begin
      yymin = 0
      yymax = 2* medwidth+1
   endif

   if xxmax gt ee(1)-1 then begin
       xxmin = ee(1)-1-(2* medwidth+1)
       xxmax = ee(1)-1
   endif

   if yymax gt ee(2)-1 then begin
       yymin = ee(2)-1-(2* medwidth+1)
       yymax = ee(2)-1
   endif

for kkkk = 0,3 do begin & ddum(xx0,yy0,kkkk) = median(reform(ddum(xxmin:xxmax,yymin:yymax,kkkk))) & endfor


endfor

dum = ddum

endif



; convert to IQUV      
   getiquv,dum,ii,qq,uu,vv
   ii = float(ii)
   if n_elements(bad_pix) eq 0 then ii = ii - float(dc)

   fff = where(ii gt 0 and finite(ii) eq 1)
   ww = where(ii le 0 or finite(ii) ne 1)
   if ww(0) ne -1 then ii(ww) = 1. ; mean(ii(fff))

  ; ff = where(ii le 0)
  ; if ff(0) ne -1 then ii(ff) = 1.
  ; ff = where(finite(ii) ne 1)
  ; if ff(0) ne -1 then ii(ff) = 1.
  
   dum(*,*,0) = ii
   dum(*,*,1) = float(qq)
   dum(*,*,2) = float(uu)
   dum(*,*,3) = float(vv)
    
   ; define array to store beams  
   ibeams = fltarr(n_slits,2,dx+1,dy+1,4)

; cut out beams without alignment for n slits
for kk = 0,n_slits-1 do begin &  ii_temp = ii & ibeams(kk,0,*,*,0) = ii_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1))  & ibeams(kk,1,*,*,0) = ii_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & qq_temp = qq & ibeams(kk,0,*,*,1) = qq_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,1) = qq_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & uu_temp = uu & ibeams(kk,0,*,*,2) = uu_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,2) = uu_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & vv_temp = vv & ibeams(kk,0,*,*,3) = vv_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,3) = vv_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & endfor

; cut out the aligned part beams for n slits
   ibeams1 = fltarr(n_slits,2,dx_final+1,dy_final+1,4)

for kk = 0,n_slits-1 do begin & ibeams1(kk,0,*,*,*) =  ibeams(kk,0,posx_final(kk,0):posx_final(kk,1), posy_final(kk,0):posy_final(kk,1),*) & ibeams1(kk,1,*,*,*) =  ibeams(kk,1,posx_final(kk,2):posx_final(kk,3), posy_final(kk,2): posy_final(kk,3),*) & endfor

   ibeams1old = ibeams1

; define array to store beams  
;ibeams = fltarr(n_slits,2,dx+1,dy+1,4)
; cut out the part beams for n slits
;for kk = 0,n_slits-1 do begin &  ii_temp = reform(dum(*,*,0)) & ibeams(kk,0,*,*,0) = ii_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1))  & ibeams(kk,1,*,*,0) = ii_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & qq_temp = reform(dum(*,*,1)) & ibeams(kk,0,*,*,1) = qq_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,1) = qq_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & uu_temp = reform(dum(*,*,2)) & ibeams(kk,0,*,*,2) = uu_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,2) = uu_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & vv_temp = reform(dum(*,*,3)) & ibeams(kk,0,*,*,3) = vv_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,3) = vv_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & endfor

; cut out the aligned part beams for n slits
 ;ibeams1 = fltarr(n_slits,2,dx_final+1,dy_final+1,4)
;for kk = 0,n_slits-1 do begin & ibeams1(kk,0,*,*,*) =  ibeams(kk,0,posx_final(kk,0):posx_final(kk,1), posy_final(kk,0): posy_final(kk,1),*) & ibeams1(kk,1,*,*,*) =  ibeams(kk,1,posx_final(kk,2):posx_final(kk,3), posy_final(kk,2): posy_final(kk,3),*) & endfor

; apply gain correction
for kk = 0,n_slits -1 do begin &  for kkk = 0,3 do begin & ibeams1(kk,0,*,*,kkk) =  ibeams1(kk,0,*,*,kkk) / gaintable(kk,0,*,*) & ibeams1(kk,1,*,*,kkk) =  ibeams1(kk,1,*,*,kkk) / gaintable(kk,1,*,*)  & endfor & endfor

; apply beam rotation
for kk = 0,n_slits -1 do begin &  for kkk = 0,3 do begin & ibeams1(kk,0,*,*,kkk) =  rot(reform(ibeams1(kk,0,*,*,kkk)),aangle(kk),/interp) & ibeams1(kk,1,*,*,kkk) =  rot(reform(ibeams1(kk,1,*,*,kkk)),aangle(kk),/interp)  & endfor & endfor

; rotate gaintable to have later spectra with hairlines not removed
; for subpixel alignment
gain1 = gaintable
for kk = 0,n_slits -1 do begin & gain1(kk,0,*,*) = rot(reform(gaintable(kk,0,*,*)),aangle(kk),/interp) & gain1(kk,1,*,*) = rot(reform(gaintable(kk,1,*,*)),aangle(kk),/interp) & endfor

; at first scan step, determine location of spectral lines
   if ( (k-scan_range(0))) eq 0 then begin

; average profile
      iline = total(reform(ibeams1(0,0,*,*,0)),2)

if max(iline) eq 0 then iline = iline + 1.
 iline = abs(iline/float(max(iline))-1)

 ; search for lines automatically 
 n = length(iline)
 ilobes = fltarr(n-4)
 threshh = .2
 ilobeindex = 0.
 for kkk = 5,n-5 do begin
 if iline(kkk) gt max(iline(kkk+1:kkk+4)) and iline(kkk) gt max(iline(kkk-4:kkk-1)) and iline(kkk) gt threshh then begin
 ilobes(ilobeindex) = kkk
 ilobeindex = ilobeindex+1
endif
 endfor

 if n_elements(verbose) ne 0 then print,ilobeindex,' spectral lines found.'

wdel17 = 0.


; location of solar lines for |QUV| maps
if n_elements(pos_lines) ne 0 then begin
   ilobes = pos_lines
   ilobeindex = n_elements(ilobes)
endif

; mark center of lines manually
if ilobeindex ne 2 or n_elements(check_lines) ne 0 then begin
   print,' Automatic detection of spectral lines failed.'
   print,' Please select lines manually:'
   window,17,title = 'Please mark 2 solar spectral lines'
   wdel17 = 1.
   !p.multi = 0
   plot,iline 
   ilobeindex = 0.
   ilobes = fltarr(n-4)
   while ilobeindex ne 2 do begin
      wset,17
      cursor,x,y,/data,/wait
      if !mouse.button eq 1 then begin
         ilobes(ilobeindex) = x
         ilobeindex = ilobeindex+1
         print,'Line ',ilobeindex
         print,'at ',x
         oplot,[x,x],[0,1]
         wait,.5
      endif
      if ilobeindex eq 2 then begin
         ans = read_keyboard('Selection okay y(es)/n(o) ')
         if ans eq 'n' then begin
            ilobeindex = 0
            plot,iline    
         endif
      endif
   endwhile
endif

ilobes = ilobes(0:ilobeindex-1)
ilobes = ilobes(sort(ilobes))

; location of spectral line to be used for alignment purposes
; should be set to Si 1082.7 nm for He 1083 nm data and to 1564.8 nm for
; 1565 nm data
range_telu = [-15,15]+ilobes(0)

range_telu = [-5,5]+ilobes(0)

if range_telu(0) lt 0 then range_telu(0) = 0.
if range_telu(1) gt dx_final then range_telu(1) = dx_final

ilo = range_telu(0)
ihi = range_telu(1)

pos_solar = [ilobes(0)-30,ilobes(0)+30.]
         
if pos_solar(0) lt 0 then pos_solar(0) = 0.
if pos_solar(1) gt dx_final then pos_solar(1) = dx_final
   
; wavelength range for I-> QUV correction
if n_elements(cont_wl) eq 0 then cont_wl = [ilobes(0)+30,ilobes(1)-30]  
ccont_wl = cont_wl(sort(cont_wl))

if ccont_wl(0) lt 0 then ccont_wl(0) = 0.
if ccont_wl(1) gt dx_final then ccont_wl(1) = dx_final

;if n_slits gt 1 then begin
 ;           posslit = mean(poslines(kk,*,*,*))
 ;           if n_elements(cont_wl) eq 0 then ccont_wl = [posslit - 15,posslit + 15] else ccont_wl = cont_wl         
;         endif else begin
;            if n_elements(cont_wl) eq 0 then ccont_wl = [ilobes(0)+30,ilobes(1)-30] else ccont_wl = cont_wl
;         endelse

      if n_elements(hair_up) eq 0 and n_elements(no_align) eq 0 then begin

; average profile
         iline = total(transpose(reform(ibeams1(0,0,*,*,0)*gain1(0,0,*,*))),2)

         if max(iline) eq 0 then iline = iline + 1.
         iline = abs(iline/float(max(iline))-1)
   
                                ; search for hair lines automatically 
         n = length(iline)
         ilobes = fltarr(n-4)
         threshh = .5
         ilobeindex = 0.
         for kkk = 5,n-5 do begin
            if iline(kkk) gt max(iline(kkk+1:kkk+4)) and iline(kkk) gt max(iline(kkk-4:kkk-1)) and iline(kkk) gt threshh then begin
               ilobes(ilobeindex) = kkk
               ilobeindex = ilobeindex+1
            endif
         endfor
   
         if n_elements(verbose) ne 0 then print,ilobeindex,'  hairline found.'
   
         wdel17 = 0.
         ans = 1.
; mark center of hair lines manually
         if ilobeindex eq 0 or ilobeindex gt 2 or n_elements(check_hairlines) ne 0 then begin
            print
            print,' Automatic detection of hairlines failed.'
            ans = read_keyboard('How many hairlines 1/2 : ',/number)
            if ans eq 'nonumber' or ans eq 0 then ans = 1.
            print,' Please select hair lines manually:'
            window,17,title ='Please mark hair lines'
            wdel17 = 1.
            !p.multi = 0
            plot,iline 
            ilobeindex = 0.
            ilobes = fltarr(n-4)
            while ilobeindex ne ans do begin
               wset,17
               cursor,x,y,/data,/wait
               if !mouse.button eq 1 then begin
                  ilobes(ilobeindex) = x
                  ilobeindex = ilobeindex+1
                  print,'Hairline ',ilobeindex
                  print,'at ',x
                  oplot,[x,x],[0,1]
                  wait,.5
               endif
               if ilobeindex eq ans then begin
                  aans = read_keyboard('Selection okay y(es)/n(o) ')
                  if aans eq 'n' then begin
                     ilobeindex = 0
                     plot,iline    
                  endif
               endif
            endwhile
         endif
         ilobes = ilobes(0:ilobeindex-1)
         ilobes = ilobes(sort(ilobes))
   
   ; location of hair line to be used for alignment purposes
         if ans eq 1 then hair_up = [ilobes(0)-5,ilobes(0)+5] else begin
            hair_up = [ilobes(1)-5,ilobes(1)+5]
            hair_down = [ilobes(0)-5,ilobes(0)+5]
            if hair_down(0) lt 0 then hair_down(0) = 0
         endelse
         if hair_up(1) gt dy_final then hair_up(1) = dy_final

         if n_elements(plt) ne 0 then begin
            window,17,title ='Hairlines'
            plot,iline
            oplot,[0,0]+hair_up(0),[0,2]
            oplot,[0,0]+hair_up(1),[0,2]
            if ans eq 2 then begin
               oplot,[0,0]+hair_down(0),[0,2]
               oplot,[0,0]+hair_down(1),[0,2]
            endif
            print,ans,' hairlines found.'
            if n_elements(nopause) eq 0 then paused
         endif
      endif 
   endif

; do precise subpixel alignment

if n_elements(no_align) eq 0 then begin

   for kk = 0,n_slits -1 do begin 

      im1a = reform(ibeams1(kk,0,*,*,0))
      im2a = reform(ibeams1(kk,0,*,*,1))
      im3a = reform(ibeams1(kk,0,*,*,2))
      im4a = reform(ibeams1(kk,0,*,*,3))

      im1b = reform(ibeams1(kk,1,*,*,0))
      im2b = reform(ibeams1(kk,1,*,*,1))
      im3b = reform(ibeams1(kk,1,*,*,2))
      im4b = reform(ibeams1(kk,1,*,*,3))

; align beams with subpixel precision
      i1atemp  = im1a*gain1(kk,0,*,*)
      i1atemp  = transpose(i1atemp)

      im1atemp = transpose(im1a)
      im2atemp = transpose(im2a)
      im3atemp = transpose(im3a)
      im4atemp = transpose(im4a)
     
      i1btemp  = im1b*gain1(kk,1,*,*)
      i1btemp  = transpose(i1btemp)

      im1btemp = transpose(im1b)
      im2btemp = transpose(im2b)
      im3btemp = transpose(im3b)
      im4btemp = transpose(im4b)

; align in spatial direction with hair line 
      deskew1,i1atemp, im2atemp, im3atemp, im4atemp,hair_up(0),hair_up(1),0,dx_final,0,dx_final,2,/straight
      i1atemp  = im1a*gain1(kk,0,*,*)
      i1atemp  = transpose(i1atemp)
      dum =  i1atemp -  i1atemp 
      deskew1,i1atemp,im1atemp,dum,dum,hair_up(0),hair_up(1),0,dx_final,0,dx_final,2,/straight
  
      wave_shift_spinor,i1atemp,i1btemp, im2btemp, im3btemp, im4btemp,hair_up(0),hair_up(1),0,dx_final,2,/straight
      i1btemp  = im1b*gain1(kk,1,*,*)
      i1btemp  = transpose(i1btemp)
      wave_shift_spinor,i1atemp,i1btemp,im1btemp,dum,dum,hair_up(0),hair_up(1),0,dx_final,2,/straight

; align in spectral direction with spectral line
      im1a = transpose(im1atemp)
      im2a = transpose(im2atemp)
      im3a = transpose(im3atemp)
      im4a = transpose(im4atemp)
     
      im1b = transpose(im1btemp)
      im2b = transpose(im2btemp)
      im3b = transpose(im3btemp)
      im4b = transpose(im4btemp)

      wave_shift_spinor,im1a,im1b,im2b,im3b,im4b,ilo,ihi,0,dy_final,2,straight = straight


      ibeams1(kk,0,*,*,0) = im1a 
      ibeams1(kk,0,*,*,1) = im2a 
      ibeams1(kk,0,*,*,2) = im3a 
      ibeams1(kk,0,*,*,3) = im4a 

      ibeams1(kk,1,*,*,0) = im1b 
      ibeams1(kk,1,*,*,1) = im2b 
      ibeams1(kk,1,*,*,2) = im3b 
      ibeams1(kk,1,*,*,3) = im4b 

   endfor

endif

; storage array for merged beams
   ibeams_merged =  fltarr(n_slits,dx_final+1,dy_final+1,4)

; apply demodulation
   for kk = 0,n_slits -1 do begin    
  
      im1a = reform(ibeams1(kk,0,*,*,0))
      im2a = reform(ibeams1(kk,0,*,*,1))
      im3a = reform(ibeams1(kk,0,*,*,2))
      im4a = reform(ibeams1(kk,0,*,*,3))
      
      im1b = reform(ibeams1(kk,1,*,*,0))
      im2b = reform(ibeams1(kk,1,*,*,1))
      im3b = reform(ibeams1(kk,1,*,*,2))
      im4b = reform(ibeams1(kk,1,*,*,3))
      
      i1a = im1a
      i2a = im2a
      i3a = im3a
      i4a = im4a
      
      i1b = im1b
      i2b = im2b
      i3b = im3b
      i4b = im4b
      
; do subpixel alignemt by correlation if hairlines are not present
; and/or spectral lines hidden by fringes
if n_elements(no_align) ne 0 then begin
   erg = shc(i1a,i1b,/filt,/interp)
   if total(abs(erg)) lt 10 then begin
      i1b = shift_frac(i1b,erg(0),erg(1),'C')
      i2b = shift_frac(i2b,erg(0),erg(1),'C')
      i3b = shift_frac(i3b,erg(0),erg(1),'C')
      i4b = shift_frac(i4b,erg(0),erg(1),'C')
   endif
endif

; merge the beams to convert to IQUV
      ii1 = (i1a+i1b)/2.
      qq1 = (i2a-i2b)/2.
      uu1 = (i3a-i3b)/2.
      vv1 = (i4a-i4b)/2.

      ii1_temp = ii1-ii1
      qq1_temp = ii1-ii1
      uu1_temp = ii1-ii1
      vv1_temp = ii1-ii1


if n_elements(noxmat) eq 0 then begin

; apply inverse response function
      for kkk =0,dy_final do begin
         ii1_temp(*,kkk) = full_xxinv(kk,kkk,0,0)*ii1(*,kkk)+full_xxinv(kk,kkk,1,0)*qq1(*,kkk)+full_xxinv(kk,kkk,2,0)*uu1(*,kkk)+full_xxinv(kk,kkk,3,0)*vv1(*,kkk)
         qq1_temp(*,kkk) = full_xxinv(kk,kkk,0,1)*ii1(*,kkk)+full_xxinv(kk,kkk,1,1)*qq1(*,kkk)+full_xxinv(kk,kkk,2,1)*uu1(*,kkk)+full_xxinv(kk,kkk,3,1)*vv1(*,kkk)
         uu1_temp(*,kkk) = full_xxinv(kk,kkk,0,2)*ii1(*,kkk)+full_xxinv(kk,kkk,1,2)*qq1(*,kkk)+full_xxinv(kk,kkk,2,2)*uu1(*,kkk)+full_xxinv(kk,kkk,3,2)*vv1(*,kkk)
         vv1_temp(*,kkk) = full_xxinv(kk,kkk,0,3)*ii1(*,kkk)+full_xxinv(kk,kkk,1,3)*qq1(*,kkk)+full_xxinv(kk,kkk,2,3)*uu1(*,kkk)+full_xxinv(kk,kkk,3,3)*vv1(*,kkk)
      endfor

      ii1 = ii1_temp
      qq1 = qq1_temp
      uu1 = uu1_temp
      vv1 = vv1_temp

   endif else begin
      ii1 = ii1
      qq1 = ii1-ii1
      uu1 = ii1-ii1
      vv1 = ii1-ii1
   endelse



      vold = vv1
      uold = uu1
      qold = qq1
      
         if n_elements(notelcor) eq 0 and n_elements(tel_cal) eq 0 then begin

; apply telescope correction
   
         tel_azimuth = FXPAR(hdr,'DST_AZ')
         tel_elevation = FXPAR(hdr,'DST_EL')
         table_angle = FXPAR(hdr,'DST_TBL')
         tel_p = FXPAR(hdr,'DST_PEE')

         ttel_azimuth(k) =  tel_azimuth
         ttel_elevation(k) = tel_elevation
         ttable_angle(k) = table_angle

         tel_time = FXPAR(hdr,'TIME_OBS')
         if tel_time eq 0 then begin
            tel_time = FXPAR(hdr,'DST_TIME')
            tel_time = string2num(tel_time,divider ='T:')
            tel_time =  tel_time(1:*)
         endif else tel_time = string2num(tel_time,divider =':')
         tel_time = tel_time(0)+tel_time(1)/60.+tel_time(2)/3600.
         ttime(k) = tel_time


; get wavelength from file header if not provided
         if n_elements(wavelength) eq 0 then begin
            wavelength =  FXPAR(hdr,'WAVELENG')
            wavelength = string2num(wavelength,divider ='In')
            wavelength =  wavelength * 10.
         endif
         
         if ( (k-scan_range(0))) eq 0 and kk eq 0 then begin
            print
            print,' firs_1565_obs: wavelength of data is: ',wavelength
            print,' Use keyword "wavelength" if the header info is incorrect !'
            print
         endif

       ;  print,tparams(3)
         ;print
         ;print, table_angle,tel_p,tel_azimuth,tel_elevation,tparams(3)
         ;print

; get telescope matrix
         tmtx  = ibis_tmtx(tparams, wavelength, tel_azimuth, tel_elevation, table_angle)
         tmtx = tmtx/tmtx(0,0)
         Tinv = INVERT(tmtx,/double) 

; apply inverse telescope matrix      
            ii2  = Tinv(0,0)*ii1+Tinv(1,0)*qq1+Tinv(2,0)*uu1+Tinv(3,0)*vv1
            qq2  = Tinv(0,1)*ii1+Tinv(1,1)*qq1+Tinv(2,1)*uu1+Tinv(3,1)*vv1
            uu2  = Tinv(0,2)*ii1+Tinv(1,2)*qq1+Tinv(2,2)*uu1+Tinv(3,2)*vv1
            vv2  = Tinv(0,3)*ii1+Tinv(1,3)*qq1+Tinv(2,3)*uu1+Tinv(3,3)*vv1

; correct QU for parallactic angle
            DST_earth_latitude = 32.786 * !dtor
            r_frame_sphtri, DST_earth_latitude , tel_azimuth * !dtor, tel_elevation * !dtor,siteha, sitedc, neg_parallactic
            parallactic = -neg_parallactic

            rotn = !pi + parallactic - tel_p * !dtor
            crot = cos(-2.*rotn) 
            srot = sin(-2.*rotn)

            temp0 =  crot*qq2 + srot*uu2
            temp1 = -srot*qq2 + crot*uu2
         
            qq1 = temp0
            uu1 = temp1
            vv1 = vv2
            ii1 = ii2
         endif
 
; remove intensity < 0
      ff = where(ii1 lt 0)
      if ff(0) ne -1 then ii1(ff) = 0.

; store calibrated IQUV 
      ibeams_merged(kk,*,*,0) = ii1
      ibeams_merged(kk,*,*,1) = qq1
      ibeams_merged(kk,*,*,2) = uu1
      ibeams_merged(kk,*,*,3) = vv1
   endfor

; correct for I -> QUV cross talk
     for kk = 0,n_slits -1 do begin
     
      ii1 = reform( ibeams_merged(kk,*,*,0))
  
; correct I for parasitic light. Must be done before I->QUV cross-talk
; correction !
      if n_elements(beta) ne 0 then ii1 = ii1 - beta*mean(ii1)

      qq1 = reform( ibeams_merged(kk,*,*,1))
      uu1 = reform( ibeams_merged(kk,*,*,2))
      vv1 = reform( ibeams_merged(kk,*,*,3))

; shift beams to have spectral line at same location
       if n_slits gt 1 then begin
          ii1 = shift_frac(ii1,  -oofset(kk)  ,0,'C')
          qq1 = shift_frac(qq1,  -oofset(kk)  ,0,'C')
          uu1 = shift_frac(uu1,  -oofset(kk)  ,0,'C')
          vv1 = shift_frac(vv1,  -oofset(kk)  ,0,'C')
       endif

; catch dead pixels
if n_elements(catch_dead) ne 0 then begin

pp = 0
if pp eq 1 then begin
for kkk = 0,dy_final do begin
   iline = reform(ii1(*,kkk))
   qline = reform(qq1(*,kkk))
   uline = reform(uu1(*,kkk))
   vline = reform(vv1(*,kkk))
   val = median(iline)
   gg = iline - median(iline,7)
   fff = where(abs(gg) le .1*val)
   ff1 = where(abs(gg) gt .1*val)


   if fff(0) ne -1 and n_elements(fff) ne (dx_final + 1.) then begin

      ff1 = [ff1-2,ff1-1,ff1,ff1+1,ff1+2]
      ff1 = ff1(sort(ff1))
      ff1 = unique(ff1)
      fff = findgen(dx_final+1.)
   for hh = 0,n_elements(ff1)-1 do begin & fff(ff1(hh)) = -1. & endfor
      hh = where(fff ne -1.)
      fff = fff(hh)
      
      if fff(0) ne -1 and n_elements(fff) gt 10 then begin
         iline = spline(fff,iline(fff),findgen(dx_final+1),/double)
         qline = spline(fff,qline(fff),findgen(dx_final+1),/double)
         uline = spline(fff,uline(fff),findgen(dx_final+1),/double)
         vline = spline(fff,vline(fff),findgen(dx_final+1),/double)
      endif

    ii1(*,kkk) = iline
    qq1(*,kkk) = qline
    uu1(*,kkk) = uline
    vv1(*,kkk) = vline
 endif

endfor
endif


ttemp = (abs(qq1-median(qq1,10))+abs(uu1-median(uu1,10))+abs(vv1-median(vv1,10)))

iref = median(ii1)


badpixels = where(ttemp gt 0.1*iref)
ssize1 = size(ii1)


for kkk = 0.,n_elements(badpixels)-1 do begin 

  x0 = badpixels(kkk) mod ssize1(1) 
y0 = badpixels(kkk) / ssize1(1)


xmin = x0 & ymin = y0 & xmax = x0+10. & ymax = y0+10.

   if xmax ge ssize1(1) then begin
      xmin = x0 -10
      xmax = x0+10.
   endif

   if ymax ge ssize1(2) then begin
      ymin = y0-10
      ymax = y0+10.
   endif


   ii1(x0,y0) = median(ii1(xmin:xmin+10,ymin:ymin+10))
   qq1(x0,y0) = median(qq1(xmin:xmin+10,ymin:ymin+10))
   uu1(x0,y0) = median(uu1(xmin:xmin+10,ymin:ymin+10))
   vv1(x0,y0) = median(vv1(xmin:xmin+10,ymin:ymin+10))



endfor


;ii1 = median(ii1,5)
;qq1 = median(qq1,5)
;uu1 = median(uu1,5)
;vv1 = median(vv1,5)




endif

; correct I->QUV cross-talk row by row    
      for kkk = 0,dy_final do begin
         iline = reform(ii1(*,kkk))
         qline = reform(qq1(*,kkk))
         uline = reform(uu1(*,kkk))
         vline = reform(vv1(*,kkk))

; deactivate  I->QUV cross-talk correction for /telcal
         if n_elements(tel_cal) eq 0 then begin
           ; i2q = mean( qline(ccont_wl(0):ccont_wl(1))/iline(ccont_wl(0):ccont_wl(1)) )
           ; i2u = mean( uline(ccont_wl(0):ccont_wl(1))/iline(ccont_wl(0):ccont_wl(1)))
           ; i2v = mean( vline(ccont_wl(0):ccont_wl(1))/iline(ccont_wl(0):ccont_wl(1)))

            i2q = median( qline(ccont_wl(0):ccont_wl(1))/iline(ccont_wl(0):ccont_wl(1)) )
            i2u = median( uline(ccont_wl(0):ccont_wl(1))/iline(ccont_wl(0):ccont_wl(1)))
            i2v = median( vline(ccont_wl(0):ccont_wl(1))/iline(ccont_wl(0):ccont_wl(1)))

            qline = qline - i2q*iline
            uline = uline - i2u*iline
            vline = vline - i2v*iline

            ii2q(kk*nsteps_total+k-scan_range(0),kkk) = i2q
            ii2u(kk*nsteps_total+k-scan_range(0),kkk) = i2u
            ii2v(kk*nsteps_total+k-scan_range(0),kkk) = i2v

         endif

; correct additional V->QU cross-talk
         if n_elements(v2q) ne 0 then begin
            qline = qline - v2q*vline
         endif
         if n_elements(v2u) ne 0 then begin
            uline = uline - v2u*vline
         endif
         
         qq1(*,kkk) = qline
         uu1(*,kkk) = uline
         vv1(*,kkk) = vline

      endfor

; capture intensity spikes along the slit that come from flat field residuals
      if smooth_width gt 1 then begin
  ; spike correction
         ff1a = total(ii1(ccont_wl(0):ccont_wl(1),*),1)/float(ccont_wl(1)-ccont_wl(0)+1.)
         corr1a = ff1a / smooth(ff1a,smooth_width)
              
     ; apply spike correction     
      for kkk = 0,dx_final do begin &  ii1(kkk,*) =  ii1(kkk,*) / corr1a  & qq1(kkk,*) =  qq1(kkk,*) / corr1a  &uu1(kkk,*) =  uu1(kkk,*) / corr1a  &vv1(kkk,*) =  vv1(kkk,*) / corr1a & endfor

      endif

       ibeams_merged(kk,*,*,0) = ii1
       ibeams_merged(kk,*,*,1) = qq1
       ibeams_merged(kk,*,*,2) = uu1
       ibeams_merged(kk,*,*,3) = vv1

  ; do plotting
       if n_elements(plt) ne 0 then begin
          wset,kk
        if dx_final lt 800 then begin

             tvscl,congrid(ii1,dx_final/img_scale_disp,500,/interp) >0 ,0,0 ;< 16000
             tvscl,congrid(qq1/ii1,dx_final/img_scale_disp,500,/interp) >(-.2) < .2,dx_final/img_scale_disp,0
             tvscl,congrid(uu1/ii1,dx_final/img_scale_disp,500,/interp)>(-.2) < .2,dx_final*2/img_scale_disp,0
             tvscl,congrid(vv1/ii1,dx_final/img_scale_disp,500,/interp)>(-.2) < .2,dx_final*3/img_scale_disp,0
          endif else begin
             tvscl,congrid(ii1,dx_final/img_scale_disp,500/img_scale_disp,/interp) >0 ,0,0;< 16000
             tvscl,congrid(qq1/ii1,dx_final/img_scale_disp,500/img_scale_disp,/interp) >(-.2) < .2,dx_final/img_scale_disp,0
             tvscl,congrid(uu1/ii1,dx_final/img_scale_disp,500/img_scale_disp,/interp)>(-.2) < .2,dx_final*0,500/img_scale_disp
             tvscl,congrid(vv1/ii1,dx_final/img_scale_disp,500/img_scale_disp,/interp)>(-.2) < .2,dx_final*1/img_scale_disp,500/img_scale_disp

          endelse
           
       endif

; create overview maps
      ic(kk*nsteps_total+k-scan_range(0),*) = total(ii1(ccont_wl(0):ccont_wl(1),*),1)/float(ccont_wl(1)-ccont_wl(0)+1.)

      totq(kk*nsteps_total+k-scan_range(0),*) = total( abs ( qq1(pos_solar(0):pos_solar(1),*)/ii1(pos_solar(0):pos_solar(1),*)),1)
      totu(kk*nsteps_total+k-scan_range(0),*) = total( abs ( uu1(pos_solar(0):pos_solar(1),*)/ii1(pos_solar(0):pos_solar(1),*)),1)
      totv(kk*nsteps_total+k-scan_range(0),*) = total( abs ( vv1(pos_solar(0):pos_solar(1),*)/ii1(pos_solar(0):pos_solar(1),*)),1)

   endfor

; do plotting
   if n_elements(plt) ne 0 then begin
      wset,4
      tvscl,congrid(ic,xxssize/2.,300,/interp) > 10 ,0,0  ; < 10000.
      tvscl,alog(congrid(totq,xxssize/2.,300,/interp) > .001),xxssize/2.,0
      tvscl,alog(congrid(totu,xxssize/2.,300,/interp) > .001),0,300
      tvscl,alog(congrid(totv,xxssize/2.,300,/interp) > .001),xxssize/2.,300

   endif

; for spectroscopic data, kick out QUV again
   if int_only eq 1 then begin
      ibeams_merged = reform(ibeams_merged(*,*,*,0))
   endif

; write out data
   writeu,outunit,ibeams_merged

; timing information
   print,' Scan steps to go: ',scan_range(1)- k
   newt = systime(1)
   
   secsperstep = newt - oldt
   full_time = secsperstep*(scan_range(1)- k)
   print,' Time per scan step: ', secsperstep,' s'
   print,' Estimated time of termination: ',floor(full_time/3600),' h',floor( (full_time - floor(full_time/3600)*3600)/60 ),' min',round(full_time mod 60),' s'
   
endfor

; close output file
free_lun,outunit

print,'Saving additional maps and info to ',outfile2

; save overview maps and other info
save,ic,totq,totu,totv,sshifts,v2q,v2u,llinepos,pos_solar,dx_final,dy_final,n_slits,int_only,ccont_wl,ii2q,ii2u,ii2v,ttime, ttel_azimuth, ttel_elevation, ttable_angle,filename = outfile2

; close windows
if wdel17 eq 1 then wdelete,17

;stop

end

