pro firs_630_ff,input_dir=input_dir,output_dir=output_dir,no_av = no_av,beam_ang = beam_ang,plt = plt, verbose = verbose, int_limit = int_limit,flat_file = flat_file,dark_file = dark_file,avdc_only = avdc_only,sspeed = sspeed,change_save = change_save,swap = swap,check_shift = check_shift,yyrange = yyrange,pickx = pickx,check_lines = check_lines,nopause = nopause,help = help

;+
;============================================================================
;
;	procedure : firs_630_ff  
;
;	purpose   : calculate gain table for FIRS 630 nm data
;
;	author    : cbeck@nso Apr 2013
;
;==============================================================================

if n_elements(help) ne 0 then begin
	print
	print, "usage:   firs_630_ff,input_dir=input_dir,output_dir=output_dir,no_av = no_av,beam_ang = beam_ang,plt = plt, verbose = verbose, int_limit = int_limit,flat_file = flat_file,dark_file = dark_file,avdc_only = avdc_only,sspeed = sspeed,change_save = change_save,swap = swap,check_shift = check_shift,yyrange = yyrange,help = help"
	print
	print, "	Calculate gain table for FIRS data of the 630 nm channel."
        print, "         "
	print
	print, "	Arguments"
        print, "                none"
        print
        print, "         Keywords :"
        print, "              input_dir: source directory with raw data."
        print, "              output_dir: destination directory for reduced data."
        print, "              /no_av: switch to not averaging raw data, but use previously"
        print, "                      created save files."
        print, "              beam_angle: fltarr(1). Rotation of beams, overrules"
        print, "                       value determined in the routine."
        print, "              /plt : switch on plotting (recommended)"
        print, "              /verbose : switch to activate extensive verbal"
        print, "                        output on the terminal."
        print, "              int_limit: fltarr(1). Intensity threshold to"
        print, "                        localize the beams/slits. Fraction of"
        print, "                        maximal intensity."
        print, "              flat_file: number of flat field file to be reduced"
        print, "                        in the source directory."
        print, "              dark_file: number of dark current file to be used"
        print, "                        in the source directory."
        print, "               /avdc_only: switch to only average the DC file."
        print, "               sspeed :  speeds up the convergence in iterate_gain."
        print, "                         > 0 < ~ 10. Too high value leaves line"
        print, "                         residuals."
        print, "              change_save: directory where the file information"
        print, "                        on the source directory was saved to."
        print, "               /swap   : do a swap endian of the raw data."
        print, "               /check_shift: switch to enforce the manual"
        print, "                        check of the shift between beams."
        print, "                        Needs to be set if offset between upper/"
        print, "                        lower beam is too large to be catched"
        print, "                        by correlation."
        print, "               yyrange : fltarr(4). Height of the two beams along the slit."
        print, "                         [y0,y1,y2,y3] = lower (y0,y1) and upper"
        print, "                         beam (y3,y4)" 
        print, "               /pickx  : manually select slits in x direction"
        print, "               /check_lines: manually select spectral lines"
        print, "               /nopause: ignore all stop/pause commands."
        print, "               /help   : switch to active the help text above."
        print
return
endif
;==============================================================================
;-

if n_elements(int_limit) eq 0 then  int_limit = .3
if n_elements(flat_file) eq 0 then  flat_file = 0
if n_elements(dark_file) eq 0 then  dark_file = 0
if n_elements(sspeed) eq 0 then  sspeed = 2.

; read in file with information on data
erg = read_directory(input_dir,filter ='*.sav')
if n_elements(change_save) ne 0 then erg = read_directory(change_save,filter ='*.sav')

outfile = erg.files
if n_elements(change_save) eq 0 then begin
   gfile = input_dir+outfile(0)  
   erg = check_existence(gfile,/read,/change)
   restore,erg(0)  
endif else begin
   gfile = change_save+outfile(0)
   erg = check_existence(gfile,/read,/change)
   restore,erg(0)  
endelse


;if n_elements(change_save) eq 0 then restore,input_dir+outfile(0) else restore,change_save+outfile(0)


; 16 cal   2 map  1 ff 0 dc

; dark current base file
darkbasefile = bbasefile1(dark_file)

; flat field file
flatbasefile = bbasefile1(flat_file)

; average dark current images
; dark current data files
erg = read_directory(input_dir,filter =darkbasefile+'*')
darkfiles = erg.files(where(erg.types eq 0))

n = float(n_elements(darkfiles))

; skip if no_av is set and use saved file
if n_elements(no_av) eq 0 then begin

   print,' Averaging dc files...',darkbasefile

   ffxread,input_dir+darkfiles(0),dum,hdr
   tt = size(dum)
   if tt(3) eq 8 then begin
      dum(*,*,0) = .5*( dum(*,*,0)+dum(*,*,4))
      dum(*,*,1) = .5*( dum(*,*,1)+dum(*,*,5))
      dum(*,*,2) = .5*( dum(*,*,2)+dum(*,*,6))
      dum(*,*,3) = .5*( dum(*,*,3)+dum(*,*,7))
   endif
   dum = dum(*,*,0:3)
   getiquv,dum,ii,qq,uu,vv
   ii = float(ii)
   dc = ii - ii

   iindex = 0
   for k = 0,n-1 do begin
      error_status = 0 
      ffxread,input_dir+darkfiles(k),dum,hdr,error_status = error_status
      if n_elements(swap) ne 0 then dum = swap_endian(dum)        
      if tt(3) eq 8 then begin
         dum(*,*,0) = .5*( dum(*,*,0)+dum(*,*,4))
         dum(*,*,1) = .5*( dum(*,*,1)+dum(*,*,5))
         dum(*,*,2) = .5*( dum(*,*,2)+dum(*,*,6))
         dum(*,*,3) = .5*( dum(*,*,3)+dum(*,*,7))
      endif
      dum = dum(*,*,0:3)
      getiquv,dum,ii,qq,uu,vv
      ii = float(ii)
      if error_status eq 1 then begin
         dc = dc + ii  
         iindex = iindex + 1.
      endif


   endfor
   dc = dc / iindex
   erg = check_existence(output_dir+'dc.'+darkbasefile+'sav',/write,/change)
   outfffile = erg(0)
   save,dc, outfffile
   
   print,' Saving dc to...',outfffile

   ; jump to end if only averaging of dark is done
   if n_elements(avdc_only) ne 0 then goto,ende

endif

; average flatfield images
; flat field data files
erg = read_directory(input_dir,filter =flatbasefile+'*')
flatfiles = erg.files(where(erg.types eq 0))

n = float(n_elements(flatfiles))

if n_elements(no_av) eq 0 then begin

   print,' Averaging ff files...',flatbasefile

   ffxread,input_dir+flatfiles(0),dum,hdr
   tt = size(dum)
   if tt(3) eq 8 then begin
      dum(*,*,0) = .5*( dum(*,*,0)+dum(*,*,4))
      dum(*,*,1) = .5*( dum(*,*,1)+dum(*,*,5))
      dum(*,*,2) = .5*( dum(*,*,2)+dum(*,*,6))
      dum(*,*,3) = .5*( dum(*,*,3)+dum(*,*,7))
   endif
   dum = dum(*,*,0:3)
   getiquv,dum,ii,qq,uu,vv
   ii = float(ii)      
   ff_img = ii -ii

   iindex = 0.
   
   for k = 0,n-1 do begin 
      error_status = 0 
      ffxread,input_dir+flatfiles(k),dum,hdr,error_status = error_status 
      if n_elements(swap) ne 0 then dum = swap_endian(dum)  
      if tt(3) eq 8 then begin
         dum(*,*,0) = .5*( dum(*,*,0)+dum(*,*,4))
         dum(*,*,1) = .5*( dum(*,*,1)+dum(*,*,5))
         dum(*,*,2) = .5*( dum(*,*,2)+dum(*,*,6))
         dum(*,*,3) = .5*( dum(*,*,3)+dum(*,*,7))
   endif
      dum = dum(*,*,0:3)
      getiquv,dum,ii,qq,uu,vv
      ii = float(ii)
      if error_status eq 1 then begin 
         ff_img = ff_img + ii
         iindex = iindex + 1. 
      endif
   endfor

   ff_img = ff_img / iindex

   erg = check_existence(output_dir+'ff.'+flatbasefile+'sav',/write,/change)
   outfffile = erg(0)
   
   save,ff_img,filename = outfffile
   
   print,' Saving ff to...',outfffile

endif

; restore averaged dc and ff images

erg = check_existence(output_dir+'dc.'+darkbasefile+'sav',/read)
if erg(1) ne 1 then begin
   print,'!firs_630_ff: cannot read ',output_dir+'dc.'+darkbasefile+'sav'
   print,' Please check dc save file.'
   print,' Aborting.'
   return
endif
restore,output_dir+'dc.'+darkbasefile+'sav'

erg = check_existence(output_dir+'ff.'+flatbasefile+'sav',/read)
if erg(1) ne 1 then begin
   print,'!firs_630_ff: cannot read ',output_dir+'dc.'+flatbasefile+'sav'
   print,' Please check ff save file.'
   print,' Aborting.'
   return
endif
restore,output_dir+'ff.'+flatbasefile+'sav'


; output file of gaintable
outfile = output_dir +'gain.'+flatbasefile+'sav'
erg = check_existence(outfile,/write,/change)
outfile = erg(0)

ff_img = float(ff_img) - dc

; catch values below 0
ww = where(ff_img lt 0)
if ww(0) ne -1 then ff_img(ww) = 1.
ww = where(ff_img eq 0)
if ww(0) ne -1 then ff_img(ww) = 1.

ff_img1 = ff_img

; get size of ff image
tt = size(ff_img)

print,'Dimensions of data: ',tt

; switch to distinguish full Stokes and Stokes I only
;if tt(0) eq 2 then 
int_only = 1 

ff_img2 = ff_img

; determine size for display
scale = 2.
device,get_screen_size = xx  

if tt(1)/scale gt xx(0) or tt(2)/scale gt xx(1) then scale = 4.
if tt(1)/scale gt xx(0) or tt(2)/scale gt xx(1) then scale = 6.

print,' Trying automatic detection of slits in x...'

; resize image for determination of beam locations, needs to fit on
; the screen
ff_res = congrid(ff_img2,tt(1)/scale,tt(2)/scale,/interp) 
tt1 = size(ff_res)

; make mask at int_limit * I_max
temp = ff_res
limit = int_limit*max(temp)
temp(0:1,0:1) = limit
temp(tt1(1)-2:tt1(1)-1,tt1(2)-2:tt1(2)-1) = limit

makecontplot,temp,limit,gband = 50, /noshort,mask = mmask,/filled

mmask = median(mmask,15)
ff = where(mmask ne 0)
if ff(0) ne -1 then mmask(ff) = 1

; search for jumps from 0 to 1 in the mask in x direction
ff1 =total(mmask(*,100:150),2)/float(51.)
nn = n_elements(ff1)

; expand by width pixels to left and right to get correct positions
; also for single slit data
width = 3

ff1 = [fltarr(width),ff1,fltarr(width)]
rr = ff1 - ff1

; identify jumps of value from 0 to 1
for k = 0,nn+2*width-2 do begin & rr(k) = abs(ff1(k+1)-ff1(k)) & endfor
nlobes = fltarr(n_elements(ff1))

for k = width,n_elements(rr)-1-width do begin & if rr(k) gt max(rr(k-width:k-1)) and rr(k) gt max(rr(k+1:k+width)) then nlobes(k-width) = k & endfor
ff = where(nlobes ne 0)

; position of beams in x
nlobes = nlobes(ff)-width

; do manual selection of beams in x
if n_elements(pickx) ne 0 then begin

   print
   n_slits = read_keyboard(' How many slits ? ',/number)
   print

   pxx = fltarr(n_slits*2)
   iindex = 0.

   window,0,xs = tt1(1),ys = tt1(2),title = 'Please select location of slit(s) in x'
   tvscl,ff_res

   finished = 0
   while finished ne 1 do begin
      
      wset,0
      cursor,x,y,/device,/change
      tvscl,ff_res
      
      plots,[0,0]+x,[0,tt1(2)],/device
      plots,[0,tt1(1)],[0,0]+y,/device

   for kk = 0,2*n_slits-1 do begin & plots,[0,0]+pxx(kk,0),/device & & plots,[0,0]+pxx(kk),[0,tt1(2)],/device & endfor

      if !mouse.button eq 1 then begin
         pxx(iindex) = x
         print,pxx
         iindex = iindex + 1
         wait,.5
      endif

      if iindex eq 2*n_slits then begin
         erg = read_keyboard('Accept locations y/n ? ')
         if erg ne 'n' then finished = 1
         if erg eq 'n' then begin
             pxx = fltarr(n_slits*2)
             iindex = 0.
         endif
      endif
   endwhile

nlobes = pxx

endif

; determine number of slits
n_slits = 1

if n_elements(nlobes) gt 2 then n_slits = 2
if n_elements(nlobes) gt 4 then n_slits = 3
if n_elements(nlobes) gt 6 then n_slits = 4

; if number of values in x is odd, add maximal x as end position of last beam 
if n_elements(nlobes) eq 3 then nlobes = [nlobes,tt1(1)-1]
if n_elements(nlobes) eq 5 then nlobes = [nlobes,tt1(1)-1]
if n_elements(nlobes) eq 7 then nlobes = [nlobes,tt1(1)-1]

if n_slits eq 4 and n_elements(nlobes) lt 8 then nlobes = [nlobes ,tt1(1)-1]
if n_slits eq 4 and n_elements(nlobes) gt 8 then begin
   nnlobes = nlobes(0:7)
   nnlobes(7) = max(nlobes)
   nlobes = nnlobes
endif
if n_slits eq 4 and n_elements(nlobes) eq 8 then nlobes(7) = tt1(1)-1.


; do plotting
if n_elements(plt) ne 0 then begin

   window,0,title = 'Automatic detection of slits in x.'+num2string(fix(n_slits))+' slit(s) found.'
   plot,rr(5:*)
for k = 0,n_elements(nlobes)-1 do begin & oplot,[0,0]+nlobes(k),[0,.5*max(rr)],color = 240,thick = 3 & endfor


   legend1,text =[' Intensity jumps',' Borders of slit(s) set'],color =[256^3.-1,240],ttextcolor = [256^3.-1,240],pos = 0,linestyle = fltarr(2),charsize = 2.

   if n_elements(nopause) eq 0 then paused

endif

; store determined x positions of the beams
posx = fltarr(n_slits,2)
index_slit = 0.
index = 0.
for k =0,n_elements(nlobes)-1 do begin & posx(index_slit,index) = nlobes(k) & index = (index + 1) & if index eq 2 then index_slit = index_slit + 1 & index = index mod 2 & endfor

print,' Trying automatic detection of beams in y...'

; determine positions in y using derivative
posy = fltarr(n_slits,4)

for k = 0,n_slits-1 do begin & ff =  total(mmask(posx(k,0):posx(k,1),*),1)/float(posx(k,1)-posx(k,0)) & rr =deriv(ff) & erg = min(rr(100:tt1(2)-100),posmin)& erg = max(rr(100:tt1(2)-100),posmax) & posy(k,1) = posmin+100 & posy(k,2) = posmax+100 & endfor

; use median position in y for the beams
posy(*,1)  = posy(*,1)-posy(*,1)+median(posy(*,1))
posy(*,2)  = posy(*,2)-posy(*,2)+median(posy(*,2))

posy(*,0) = 0
posy(*,3) = tt1(2)-1

; overrule: set ranges in y to user-provided values
if n_elements(yyrange) ne 0 then begin
   posy(*,0) = yyrange(0)
   posy(*,1) = yyrange(1)
   posy(*,2) = yyrange(2)
   posy(*,3) = yyrange(3)
endif

; do plotting
if n_elements(plt) ne 0 then begin
   window,0,xs = tt1(1),ys = tt1(2),title = 'Located slits and beams'
   tvscl,ff_res
for k = 0,n_slits-1 do begin & draw_rec,[posx(k,0),posx(k,1)],[posy(k,0),posy(k,1)],/device,thick = 3 & draw_rec,[posx(k,0),posx(k,1)],[posy(k,2),posy(k,3)],/device,thick = 3 & endfor

    if n_elements(nopause) eq 0 then paused
endif

; get sizes of beams in x, use minimal value
dx = fltarr(n_slits)
for k = 0,n_slits-1 do begin & dx(k) = posx(k,1)-posx(k,0) & endfor
dx = min(dx)

; get sizes of beams in y, use minimal value
dy = fltarr(n_slits,2)
for k = 0,n_slits-1 do begin & dy(k,0) = posy(k,1)-posy(k,0) & dy(k,1) = posy(k,3)-posy(k,2) & endfor
dy = min(dy)

; restrict beams to x_i + dx, y_i + dy (from left border in x to x+dx,
; lower boundary in y to y+dy)
for k = 0,n_slits-1 do begin & posx(k,1) =  posx(k,0)+dx & posy(k,1) =  posy(k,0)+dy & posy(k,3) =  posy(k,2)+dy & endfor



if n_elements(plt) ne 0 then begin

   window,0,xs = tt1(1),ys = tt1(2),title = 'Slits/beams clipped to same size'
   tvscl,ff_res
for k = 0,n_slits-1 do begin & draw_rec,[posx(k,0),posx(k,1)],[posy(k,0),posy(k,1)],/device,thick = 3 & draw_rec,[posx(k,0),posx(k,1)],[posy(k,2),posy(k,3)],/device,thick = 3 & endfor

   if n_elements(nopause) eq 0 then paused

endif


; get values in full-sized image
posx = posx * scale
posy = posy * scale
dx = dx * scale
dy = dy * scale

print,'Trying automatic detection of beam rotation...'

; cut out area to determine rotation
area = fltarr(n_slits,dx+1,200)


if int_only eq 0 then begin
for k = 0,n_slits -1 do begin & area(k,*,*) = ff_img1(posx(k,0):posx(k,1),dy-100:dy+199,0)/ff_img1(posx(k,0):posx(k,1),dy-100:dy+99,3) & endfor
endif else begin

if n_slits gt 1 then begin
for k = 0,n_slits-2 do begin & area(k,*,*) = ff_img1(posx(k,0):posx(k,1),dy-100:dy+99)/ff_img1(posx(k+1,0):posx(k+1,1),dy-100:dy+99) & endfor
area(n_slits-1,*,*) = ff_img1(posx(n_slits-1,0):posx(n_slits-1,1),dy-10:dy+189)/ff_img1(posx(n_slits-2,0):posx(n_slits-2,1),dy-100:dy+99) 
endif else begin

k = 0
area(k,*,*) = ff_img1(posx(k,0):posx(k,1),dy-10:dy+189)/reverse(ff_img1(posx(k,0):posx(k,1),dy-10:dy+189),2)

endelse
endelse


;for k = 0,n_slits -1 do begin & area(k,*,*) = ff_img1(posx(k,0):posx(k,1),dy-100:dy+99,0)/ff_img1(posx(k,0):posx(k,1),dy-100:dy+99,3) & endfor

; determine rotation angle
aangle = fltarr(n_slits)
cchi = fltarr(n_slits,100)

if n_elements(plt) ne 0 then window,0,xs = dx+1, ys = 200,title ='Automatic determination of beam rotation'

for k = 0,n_slits-1 do begin & chi = fltarr(100) & for kk = 0,99 do begin & angle = -2.5+kk/20. & a1 = reform( rot(reform(area(k,*,*)),angle,/interp)) & a1r = reverse(a1,1) & chi(kk) =  total( abs (  a1 - a1r ) ) & if n_elements(plt) ne 0 then tvscl,a1-a1r & endfor & erg = min(chi,posmin) & cchi(k,*) = chi & aangle(k) = -2.5+posmin/20. & endfor

if n_slits gt 1 then begin
   if n_slits eq 2 then diff = aangle(1)-aangle(0)
   if n_slits eq 3 then diff = [aangle(1)-aangle(0),aangle(2)-aangle(0),aangle(1)-aangle(2)]
   if n_slits eq 4 then diff = [aangle(1)-aangle(0),aangle(2)-aangle(0),aangle(3)-aangle(0),aangle(1)-aangle(2),aangle(3)-aangle(2),aangle(3)-aangle(1)]
   diff_max = max(abs(diff))

   ; if values differ, request user input
   if diff_max gt 0.3 then begin
      print,'!! firs_630_ff: Warning !! Rotation angles vary by more than 0.3 deg'
      print,'Angles: ',aangle

      window,0,title ='rotation angle'
      !p.multi = [n_slits,1,n_slits]
   for k = 0,n_slits-1 do begin & plot,-2.5+findgen(100)/20.,cchi(k,*) & endfor
      
      ans = read_keyboard('Please enter value to be used: ',/number)
      aangle = aangle -aangle + ans

   endif
endif

; do plotting
if n_elements(plt) ne 0 then begin
   window,0,title ='Beam rotation angle'
   !p.multi = [n_slits,1,n_slits]
for k = 0,n_slits-1 do begin & plot,-2.5+findgen(100)/20.,cchi(k,*) & oplot,[0,0]+aangle(k),[0,max(cchi(k,*))] &  endfor
    if n_elements(nopause) eq 0 then paused
endif

; overrule: set rotation angle to user-provided value
if n_elements(beam_ang) ne 0 then begin
   if n_elements(beam_ang) eq 1 then aangle = fltarr(n_slits)+beam_ang else aangle = beam_ang 
endif

print,'Rotation angle of beams determined to: ',aangle

; switch size between full Stokes and Stokes I only
if int_only eq 0 then ssize = 4 else ssize = 1

; cut out the n_slits beams from the image
ii = fltarr(n_slits,2,dx+1,dy+1,ssize)
for k = 0,n_slits-1 do begin & for kk = 0,ssize-1 do begin &  ff_temp = reform(ff_img1(*,*,kk)) &  & ii(k,0,*,*,kk) = ff_temp(posx(k,0):posx(k,1),posy(k,0):posy(k,1)) & ii(k,1,*,*,kk) = ff_temp(posx(k,0):posx(k,1),posy(k,2):posy(k,3)) & endfor & endfor

; do plotting
if n_elements(plt) ne 0 then begin
for k = 0,n_slits -1 do begin & window,k,xs = 2*dx,ys = dy,title='Upper/lower beams in slit '+num2string(k)+' mod. state 0' & tvscl,ii(k,0,*,*,0),0,0 & tvscl,ii(k,1,*,*,0),dx,0 & endfor
    if n_elements(nopause) eq 0 then paused
endif

print,'Aligning upper and lower beams...'

; correlate upper and lower beam in each slit
eerg = fltarr(n_slits,2)
for k =0,n_slits -1 do begin  & erg = shc( reform(ii(k,0,50:dx-50,50:dy-50,0)),reform(ii(k,1,50:dx-50,50:dy-50,0)),/filt) & eerg(k,*) = erg & print,erg & endfor 

; if displacement between beams varies between slits or /check_shift
; is set, request better start value for correlation and repeat determination

if n_slits gt 1 then begin
   if n_slits eq 2 then diff = eerg(1,1)-eerg(0,1)
   if n_slits eq 3 then diff = [eerg(1,1)-eerg(0,1),eerg(2,1)-eerg(0,1),eerg(1,1)-eerg(2,1)]
   if n_slits eq 4 then diff = [eerg(1,1)-eerg(0,1),eerg(2,1)-eerg(0,1),eerg(3,1)-eerg(0,1),eerg(1,1)-eerg(2,1),eerg(3,1)-eerg(2,1),eerg(3,1)-eerg(1,1)]
   diff_max = max(abs(diff))
   if diff_max gt 10 then begin
      print,'!! firs_630_ff: Warning !! Vertical shifts for beam alignment deviate more than 10 pixels'
      print,'Shifts along slit: ',eerg(*,1)
      ans = read_keyboard('Please enter value to be used: ',/number)
   endif

   if diff_max gt 10 or n_elements(check_shift) ne 0 then begin

   check_shift:

   eerg = fltarr(n_slits,2)
   for k =0,n_slits -1 do begin  & i1 = reform(ii(k,0,*,*,0)) & i2 = reform(ii(k,1,*,*,0)) & i2 = shift(i2,0,ans) & erg = shc( reform(i1(50:dx-50,50:dy-50,0)),reform(i2(50:dx-50,50:dy-50,0)),/filt) & eerg(k,*) = erg+[0,ans] & print,eerg(k,1) & endfor 

      ans = read_keyboard('Accept new shifts: y/n ')
      if ans eq 'n' then begin
         ans = read_keyboard('Please enter value to be used: ',/number)
         goto,check_shift
      endif
   endif

endif

; get maximal shift to restrict beam sizes
shiftx = max( abs(eerg(*,0)))
shifty = max( abs(eerg(*,1)))

; reduce beam size
dx_final = dx - shiftx
dy_final = dy - shifty

; storage array of aligned, equallz sized beams
ii1 = fltarr(n_slits,2,dx_final+1,dy_final+1,ssize)

; final position values to be cut out from observational data, correct
; automatically for the shifts
posx_final = posy-posy
posy_final = posy-posy

for k =0,n_slits-1 do begin 
   
    for kk = 0,ssize-1 do begin
   if eerg(k,0) lt 0 and  eerg(k,1) lt 0  then begin
      ii1(k,0,*,*,kk) = ii(k,0,0:dx_final,0:dy_final,kk)
      ii1(k,1,*,*,kk) = ii(k,1,abs(eerg(k,0)):abs(eerg(k,0))+dx_final,abs(eerg(k,1)):abs(eerg(k,1))+dy_final,kk)

      posx_final(k,0) = 0
      posx_final(k,1) = dx_final
      posx_final(k,2) = abs(eerg(k,0))
      posx_final(k,3) = abs(eerg(k,0))+dx_final
      
      posy_final(k,0) = 0
      posy_final(k,1) = dy_final
      posy_final(k,2) = abs(eerg(k,1))
      posy_final(k,3) = abs(eerg(k,1))+dy_final
      
   endif 
   if eerg(k,0) ge 0 and  eerg(k,1) ge 0  then begin
      ii1(k,1,*,*,kk) = ii(k,1,0:dx_final,0:dy_final,kk)
      ii1(k,0,*,*,kk) = ii(k,0,abs(eerg(k,0)):abs(eerg(k,0))+dx_final,abs(eerg(k,1)):abs(eerg(k,1))+dy_final,kk)

      posx_final(k,2) = 0
      posx_final(k,3) = dx_final
      posx_final(k,0) = abs(eerg(k,0))
      posx_final(k,1) = abs(eerg(k,0))+dx_final
      
      posy_final(k,2) = 0
      posy_final(k,3) = dy_final
      posy_final(k,0) = abs(eerg(k,1))
      posy_final(k,1) = abs(eerg(k,1))+dy_final

   endif 
   
    if eerg(k,0) lt 0 and  eerg(k,1) ge 0  then begin
      ii1(k,0,*,*,kk) = ii(k,0,0:dx_final,abs(eerg(k,1)):abs(eerg(k,1))+dy_final,kk)
      ii1(k,1,*,*,kk) = ii(k,1,abs(eerg(k,0)):abs(eerg(k,0))+dx_final,0:dy_final,kk)

      posx_final(k,0) = 0
      posx_final(k,1) = dx_final
      posx_final(k,2) = abs(eerg(k,0))
      posx_final(k,3) = abs(eerg(k,0))+dx_final
      
      posy_final(k,2) = 0
      posy_final(k,3) = dy_final
      posy_final(k,0) = abs(eerg(k,1))
      posy_final(k,1) = abs(eerg(k,1))+dy_final

   endif 

    if eerg(k,0) ge 0 and  eerg(k,1) lt 0  then begin
      ii1(k,1,*,*,kk) = ii(k,1,0:dx_final,abs(eerg(k,1)):abs(eerg(k,1))+dy_final,kk)
      ii1(k,0,*,*,kk) = ii(k,0,abs(eerg(k,0)):abs(eerg(k,0))+dx_final,0:dy_final,kk)

      posx_final(k,2) = 0
      posx_final(k,3) = dx_final
      posx_final(k,0) = abs(eerg(k,0))
      posx_final(k,1) = abs(eerg(k,0))+dx_final
      
      posy_final(k,0) = 0
      posy_final(k,1) = dy_final
      posy_final(k,2) = abs(eerg(k,1))
      posy_final(k,3) = abs(eerg(k,1))+dy_final

   endif 
endfor

endfor

; do plotting
if n_elements(plt) ne 0 then begin

for k = 0,n_slits -1 do begin & window,k,xs = 2*dx,ys = dy,title='Vertically aligned beams in slit '+num2string(k)+' mod. state 0' & tvscl,ii1(k,0,*,*,0),0,0 & tvscl,ii1(k,1,*,*,0),dx,0 & endfor
    
   if n_elements(nopause) eq 0 then paused

endif

dc1 = reform(dc(*,*,0))

lastx = dx_final+1
firsty = 0
lasty = dy_final+1

; storage arrays for gaintable and reference position of spectral line
gaintable = fltarr(n_slits,2,dx_final+1,dy_final+1,ssize)
pposref = fltarr(n_slits,2,ssize)
gaintablecoarse = fltarr(n_slits,2,dx_final+1,dy_final+1,ssize)

; loop over slits
for k = 0,n_slits -1 do begin 

   ; loop over beams
   for kk =0,1 do begin

      ; loop over modulation states
       for kkkk = 0,ssize-1 do begin

          iref = reform(ii1(k,kk,*,*,kkkk))
 
          ff = total(iref,2)

          if kk eq 0 and kkkk eq 0 then begin
                                ; determine location of spectral
                                ; lines, 4 lines
                                ; (630.15/630.20/630.25/630.27) 

             print,' Trying automatic detection of spectral lines...'

             iline = ff
             iline = abs(iline/float(max(iline))-1)
          
             n = length(iline)
             ilobes = fltarr(n-4)
             threshh = .2
             ilobeindex = 0.
             for kkk = 5,n-5 do begin
                if iline(kkk) gt max(iline(kkk+1:kkk+4)) and iline(kkk) gt max(iline(kkk-4:kkk-1)) and iline(kkk) gt threshh then begin
                   ilobes(ilobeindex) = kkk
                   ilobeindex = ilobeindex+1
                endif
             endfor
             if ilobeindex ne 4 or n_elements(check_lines) ne 0 then begin
                print,' Automatic detection of spectral lines failed.'
                print,' Please select lines manually:'
                window,0,title ='Please mark the 4 lines at 630 nm'
                !p.multi = 0
                plot,iline        
                ilobeindex = 0.
                ilobes = fltarr(n-4)
                while ilobeindex ne 4 do begin
                   wset,0
                   cursor,x,y,/data,/wait
                   if !mouse.button eq 1 then begin
                      ilobes(ilobeindex) = x
                      ilobeindex = ilobeindex+1
                      print,'Line ',ilobeindex
                      print,'at ',x
                      oplot,[x,x],[0,1]
                      wait,.5
                   endif
                   if ilobeindex eq 4 then begin
                      ans = read_keyboard('Selection okay y(es)/n(o) ')
                      if ans eq 'n' then begin
                         ilobeindex = 0
                         plot,iline    
                      endif
                   endif
                endwhile
             endif
             ilobes = ilobes(0:ilobeindex-1)
             ilobes = ilobes(sort(ilobes))
             
             ; set range of telluric line at 630.20
             range_telu = [-15,15]+ilobes(1)
             
             ilo = range_telu(0)
             ihi = range_telu(1)
; set range use for iterative improvement of gaintable            
             wave_lim = [ilobes(0)-30,ilobes(2)+10,sspeed]
          endif

; call routine to calculate gaintable
      create_gaintables_firs_630,ilo,ihi,lastx,firsty,lasty,iref,dc1,gainsave,gaincoarse,posref,wave_lim = wave_lim,verbose =verbose,help = help
;      create_gaintables_firs_630,ilo,ihi,lastx,firsty,lasty,iref,dc1,gainsave,gaincoarse,posref,wave_lim = wave_lim,verbose =verbose

; store reference position of telluric line
      pposref(k,kk,kkkk) = posref

; do plotting
      if n_elements(plt) ne 0 then begin

         window,0,xs = 3*(dx_final+1),ys = dy_final+1,title = 'Gain table of slit '+num2string(k)+', beam '+num2string(kk)
         tvscl,iref,0,0
         tvscl,iref/gainsave,dx_final+1,0
         tvscl,gainsave > .85 < 1.15,2*(dx_final+1),0
          if n_elements(nopause) eq 0 then paused
      endif
      
      gaintable(k,kk,*,*,kkkk) = gainsave
      gaintablecoarse(k,kk,*,*,kkkk) = gaincoarse

  endfor

   endfor


endfor

; try to get different slits to same average intensity 
if n_slits gt 1 then begin
    print,'Balancing the slits...'

    ii11 = ii1/gaintable
    ii11coarse = ii1/gaintablecoarse
    mean_slits = fltarr(n_slits,2,ssize)
    mean_slitscoarse = fltarr(n_slits,2,ssize)

 for k = 0,n_slits -1 do begin & for kk = 0,1 do begin & for kkk = 0,ssize-1 do begin & mean_slits(k,kk,kkk) = mean(ii11(k,kk,50:dx_final-50,50:dy_final-50,kkk)) & endfor & endfor & endfor

 for k = 0,n_slits -1 do begin & for kk = 0,1 do begin & for kkk = 0,ssize-1 do begin & mean_slitscoarse(k,kk,kkk) = mean(ii11coarse(k,kk,50:dx_final-50,50:dy_final-50,kkk)) & endfor & endfor & endfor

        iref = mean(mean_slits)
        irefcoarse = mean(mean_slitscoarse)

    for k = 0,n_slits -1 do begin & for kk = 0,1 do begin & for kkk = 0,ssize-1 do begin & gaintable(k,kk,*,*,kkk) = gaintable(k,kk,*,*,kkk)*mean_slits(k,kk,kkk)/iref & endfor & endfor & endfor
    for k = 0,n_slits -1 do begin & for kk = 0,1 do begin & for kkk = 0,ssize-1 do begin & gaintablecoarse(k,kk,*,*,kkk) = gaintablecoarse(k,kk,*,*,kkk)*mean_slitscoarse(k,kk,kkk)/irefcoarse & endfor & endfor & endfor

        ii11 = ii1/gaintable

        mean_slits1 = mean_slits - mean_slits

     for k = 0,n_slits -1 do begin & for kk = 0,1 do begin & for kkk = 0,ssize-1 do begin & mean_slits1(k,kk,kkk) = mean(ii11(k,kk,50:dx_final-50,50:dy_final-50,kkk)) & endfor & endfor & endfor
;        print,mean_slits
;stop

; do plotting
        if n_elements(plt) ne 0 then begin
            window,0,xs = 4*100*n_slits,ys = dy_final / 4.*2.

        for k = 0,n_slits - 1 do begin & for kk = 0,ssize-1 do begin & tvscl,congrid(reform(ii11(k,0,*,*,kk)),100,dy_final / 4.),kk*100+k*400,0 & tvscl,congrid(reform(ii11(k,1,*,*,kk)),100,dy_final / 4.),kk*100+k*400,dy_final/4. & endfor & endfor
        endif
endif

; save results
print,' Saving gain to...',outfile

save,aangle,dx,dy,dx_final,dy_final,posx,posy,posx_final,posy_final,gaintable,gaintablecoarse,cchi,n_slits,pposref,mean_slits,mean_slitscoarse,mean_slits1,int_only,filename = outfile


if n_elements(plt) ne 0 then wdelete,0,1

;stop

ende:

end

