pro firs_630_obs,plt = plt,v2q = v2q, v2u = v2u,beta = beta,obsindex = obsindex,gainindex = gainindex ,scan_range = scan_range,smooth_width = smooth_width,spec_grad = spec_grad,verbose = verbose,angle = angle,darkindex = darkindex,cal_file = cal_file,change_save = change_save,tel_cal = tel_cal,input_dir = input_dir,output_dir = output_dir,check_lines = check_lines,ccal_file = ccal_file,flat_file = flat_file,swap = swap,tmat_save = tmat_save,ooutfile = ooutfile,wavelength = wavelength,notelcor = notelcor,pos_lines = pos_lines,cont_wl = cont_wl,offang = offang,hair_up = hair_up, hair_down = hair_down,check_hairlines = check_hairlines,no_align = no_align,help = help

;+
;============================================================================
;
;	procedure : firs_630_obs  
;
;	purpose   : reduce FIRS 630 nm data
;
;	author    : cbeck@nso Apr 2013
;
;==============================================================================
;
;	Check number of arguments.
;
;==============================================================================
if n_elements(help) ne 0 then begin
	print
	print, "usage:   firs_630_obs,plt = plt,v2q = v2q, v2u = v2u,beta = beta,obsindex = obsindex,gainindex = gainindex ,scan_range = scan_range,smooth_width = smooth_width,spec_grad = spec_grad,verbose = verbose,angle = angle,darkindex = darkindex,cal_file = cal_file,change_save = change_save,tel_cal = tel_cal,input_dir = input_dir,output_dir = output_dir,check_lines = check_lines,ccal_file = ccal_file,flat_file = flat_file,swap = swap,tmat_save = tmat_save,ooutfile = ooutfile,wavelength = wavelength,notelcor = notelcor,pos_lines = pos_lines,cont_wl = cont_wl,offang = offang,hair_up = hair_up, hair_down = hair_down,check_hairlines = check_hairlines,no_align = no_align,help = help"
	print
	print, "	 Reduce FIRS data of the 630 nm channel."
        print, "         "
	print
	print, "	Arguments"
        print, "                none"
        print
        print, "         Keywords :"
        print, "              /plt : switch on plotting (recommended)"
        print, "              cont_wl : fltarr(2), [l0,l1] : wavelength pixel"
        print, "                        range to use for I->QUV cross-talk"
        print, "                        correction. If not provided, it will"
        print, "                        be set to the range around the telluric"
        print, "                        630.27 nm line."
        print, "              v2q, v2u: fltarr(1) each. Additional cross-talk"
        print, "                        correction coefficient for V->Q, V->U."
        print, "                        Q = Q - v2q * V"
        print, "              pos_lines:fltarr(4),location of spectral lines in pixel"
        print, "              beta    : parasitic light correction;" 
        print, "                        I = I-beta*<I>. Can remove false "
        print, "                        polarisation signal in the telluric lines"
        print, "              obsindex: number of observation file to be reduced"
        print, "                        in the source directory."
        print, "              gainindex:number of gain table to be used for the"
        print, "                        reduction in the target directory"
        print, "              scan_range: fltarr(2), [s0,s1]; scan steps to be"
        print, "                        reduced. If not provided, all steps"
        print, "                        will be reduced."
        print, "              smooth_width: fltarr(1). Width in pixels along"
        print, "                        the slit over which will be smoothed"
        print, "                        to remove spikes from flat field residuals."
        print, "              /spec_grad: switch to activate the removal of"
        print, "                        wavelength gradients. Attention: hard-"
        print, "                        wired range. Use not recommended."
        print, "              /verbose : switch to activate extensive verbal"
        print, "                        output on the terminal."
        print, "              angle   : fltarr(1). Rotation angle for the beams."
        print, "                        Overrules value determined in gain table"
        print, "                        calculation."
        print, "              darkindex:number of dark current to be used for the"
        print, "                        reduction in the target directory"
        print, "              cal_file: number of calibration file to be used"
        print, "                        in the source directory."
        print, "              change_save: directory where the file information"
        print, "                        on the source directory was saved to."
        print, "              /tel_cal :  switch to de-activate telescope correction"
        print, "                        and I -> QUV correction."
        print, "              input_dir: source directory with raw data."
        print, "              output_dir: destination directory for reduced data."
        print, "              smooth_width: ignore, deactivated."
        print, "              /check_lines: switch to enforce the manual selection"
        print, "                         of the spectral lines."
        print, "              ccal_file: use a different response function."
        print, "                         String with the complete filepath and"
        print, "                         filename of the xmat....sav file."
        print, "              flat_file: use a different gain table."
        print, "                         String with the complete filepath and"
        print, "                         filename of the gain.....sav file."
        print, "               /swap   : do a swap endian of the raw data."
        print, "               tmat_save:  String with the complete filepath and"
        print, "                         filename of the T-matrix parameter file."
        print, "               ooutfile: use different filename for output file."
        print, "                         Must include file path."
        print, "               offang: offset angle between telescope and instrument calibration."
        print, "               /notelcor: do not apply telescope correction." 
        print, "               pos_lines:fltarr(2), location of spectral lines in pixel. "
        print, "               hair_up: fltarr(2), upper and lower range in pixels"
        print, "                        around position of hairline."
        print, "               hair_down: fltarr(2), upper and lower range in pixels"
        print, "                        around position of 2nd hairline."
        print, "                        Ignore, currently not used."
        print, "               /check_hairlines: switch to enforce the manual selection"
        print, "               /no_align: do not use hairline for spatial alignment."
        print, "               /help   : switch to active the help text above."
        print
return
endif
;==============================================================================
;-

!p.multi = 0

; smoothing width in pixels along slit to catch spikes
if n_elements(smooth_width) eq 0 then smooth_width = 1
; number of file to be reduced
if n_elements(obsindex) eq 0 then obsindex= 0 
; number of gain table to be used
if n_elements(gainindex) eq 0 then gainindex= 0 
; number of dark current to be used
if n_elements(darkindex) eq 0 then  darkindex = 0
; number of calibration file to be used
if n_elements(cal_file) eq 0 then  cal_file = 0
; filename of save file with telescope parameters
if n_elements(tmat_save) eq 0 then  tmat_save = '/home/cbeck/laptop_teneriffa/telcal_DST/Tmatrix_May2010_4708-14125.idl'

loadct,0,/silent

; wavelength to be used in telescope calibration correction
if n_elements(wavelength) eq 0 then wavelength = 6300.

; restore save file with information on source directory, must be the
; first save file in the directory
if n_elements(change_save) ne 0 then erg = read_directory(change_save,filter ='*.sav') else erg = read_directory(input_dir,filter ='*.sav')

outfile = erg.files
if n_elements(change_save) eq 0 then begin
   gfile = input_dir+outfile(0)  
   erg = check_existence(gfile,/read,/change)
   restore,erg(0)  
endif else begin
   gfile = change_save+outfile(0)
   erg = check_existence(gfile,/read,/change)
   restore,erg(0)  
endelse

;if n_elements(change_save) eq 0 then restore,input_dir+outfile(0) else restore,change_save+outfile(0)

; 16 cal   2 map  1 ff 0 dc

; get dark current
ergdc = read_directory(output_dir,filter ='*dc*')
restore,output_dir + ergdc.files(darkindex)

; get gain table
erggain = read_directory(output_dir,filter ='*gain*')
if n_elements(flat_file) eq 0 then restore,output_dir + erggain.files(gainindex) else restore,flat_file

; reference position in spectrum for spectral lines
posref = mean(pposref)

; spectral offset between the beams, used to get the spectral lines at
; the same place in all beams
oofset = fltarr(n_slits)

for k = 0,n_slits - 1 do begin & oofset(k) = mean(pposref(k,*,*)) - mean(pposref(0,*,*)) & endfor


; get telescope matrix parameters
if n_elements(tel_cal) eq 0 then begin
   restore,tmat_save
   tparams = tt
   tparams(3) = 90.
   if n_elements(offang) ne 0 then tparams(3) = offang

endif

; get instrument calibration 
calbasefile = bbasefile1(cal_file)  
outfile1 = output_dir+'xmat.'+calbasefile+'sav'
if n_elements(ccal_file) eq 0 then restore,outfile1 else restore,ccal_file

; normalize and invert instrument response function X
xxinv = xxmat - xxmat

for kk = 0,n_separations -1 do begin

   for k = 0,n_slits-1 do begin
 
      xmat = reform(xxmat(kk,k,*,*))
      xmat = xmat / xmat(0,0)

      if n_elements(nodemod) eq 0 then begin
         xmat(0,*) = xmat(0,*)+1.
         normaliza=max(xmat(0,*))
         xmat(0,*)=xmat(0,*)/normaliza
      for i=1,3 do begin & xmat(i,*)=xmat(i,*)*xmat(0,*) & endfor
      endif

      dmod=invert(xmat,/double)
      dmod = transpose(dmod)
      xmat = transpose(xmat)

      if n_elements(nodemod) eq 0 then begin
      for i=0,3 do begin & dmod(*,i)=dmod(*,i)*xmat(i,0) & endfor
      endif

      dmod = transpose(dmod)
      xxmat(kk,k,*,*) = xmat
      xxinv(kk,k,*,*) = dmod

   endfor
endfor
ppos = fltarr(n_separations)
for k = 0,n_separations-1 do begin & ppos(k) = (yyr(k)+yyr(k+1))/2. & endfor
full_xmat = fltarr(n_slits,dy_final+1,4,4)
for k = 0,n_slits-1 do begin & xmats = reform(xxmat(*,k,*,*)) & erg = extrapolate_xmat(ppos,xmats,dy_final+1,plt= 0) & full_xmat(k,*,*,*) = erg & endfor
full_xxinv = fltarr(n_slits,dy_final+1,4,4)
for k = 0,n_slits-1 do begin & xmats = reform(xxinv(*,k,*,*)) & erg = extrapolate_xmat(ppos,xmats,dy_final+1,plt= 0) & full_xxinv(k,*,*,*) = erg & endfor


;,plt=plt

;for k = 0,n_slits-1 do begin
;  xmat = reform(xxmat(k,*,*))
;  xmat(0,*) = xmat(0,*)+1.
;  normaliza=max(xmat(0,*))
;  xmat(0,*)=xmat(0,*)/normaliza
;for i=1,3 do begin & xmat(i,*)=xmat(i,*)*xmat(0,*) & endfor
;   dmod=invert(xmat,/double)
;   dmod = transpose(dmod)
;   xmat = transpose(xmat)
;for i=0,3 do begin & dmod(*,i)=dmod(*,i)*xmat(i,0) & endfor
;   dmod = transpose(dmod)
;   xxmat(k,*,*) = xmat
;   xxinv(k,*,*) = dmod
;endfor

; get input raw data file
obsbasefile = bbasefile1(obsindex)
ergobs = read_directory(input_dir,filter ='*'+obsbasefile+'*')
obsfiles = ergobs.files

; number of steps available
nsteps = n_elements(obsfiles)

; re-sort file list when scan was repeated
if nsteps gt 1 then begin

; find numbner of repetitions
   maxrep = fltarr(nsteps)

   for k = 0,nsteps-1 do begin & eerg = string2num(obsfiles(k),divider ='.') & maxrep(k) = eerg(n_elements(eerg)-1) & endfor

      maxrep = max(maxrep)

      if maxrep ne 0 then begin
         obsfiles = strarr(nsteps)
         iiindex = 0.
      for k = 0,maxrep do begin & iindex = k & if iindex lt 10 then maxreptext = '.000'+num2string(fix(iindex)) &  if iindex ge 10 and iindex lt 100 then maxreptext = '.00'+num2string(fix(iindex))  &  if iindex ge 100 and iindex lt 1000 then maxreptext = '.0'+num2string(fix(iindex)) &  if iindex ge 1000 then maxreptext = '.'+num2string(fix(iindex))  & ergobs = read_directory(input_dir,filter ='*'+obsbasefile+'*'+maxreptext) & ffiless = ergobs.files & obsfiles(iiindex:iiindex+n_elements(ffiless)-1) = ffiless & iiindex = iiindex+n_elements(ffiless) & endfor

      endif
   endif


; reduced data file
if n_elements(ooutfile) eq 0 then outfile = output_dir +obsbasefile+'reduced.dat' else outfile = ooutfile
; file of overview maps
outfile2 = outfile+'.sav'

; set scan range to full size or use explicit scan range
if n_elements(scan_range) eq 0 then scan_range = [0,nsteps-1]

; number of steps to be reduced
nsteps_total = scan_range(1)-scan_range(0)+1.

; storage arrays for line positions and image shifts
sshifts = fltarr(n_slits,nsteps_total,2)
llinepos = fltarr(n_slits,nsteps_total,2)

; 2D overview maps I_c, |QUV|
ic = fltarr(nsteps_total*n_slits,dy_final+1)
totq = fltarr(nsteps_total*n_slits,dy_final+1)
totu = fltarr(nsteps_total*n_slits,dy_final+1)
totv = fltarr(nsteps_total*n_slits,dy_final+1)

; do plotting
if n_elements(plt) ne 0 then begin
for k = 0,n_slits -1 do begin & window,k,xs = 4*dx_final,ys = 500,title = 'Stokes spectra slit '+num2string(k) & endfor
window,4,xs = n_slits*nsteps_total*2,ys = 600,title = '2D maps'
endif

; overrule rotation angle from flat field
if n_elements(angle) ne 0 then aangle = aangle - aangle +angle

; switches
erg = 0.
erg1 = 0.

; for spectroscopic data: int_only = 1, is determined from image size
; later on
int_only = 0.

; open output file
openw,outunit,outfile,/get_lun

; loop over scan steps
for k = scan_range(0),scan_range(1) do begin

; get timing for estimate of duration
   oldt = systime(1)

; read in data
error_status = 0.
ffxread,input_dir+obsfiles(k),dum,hdr,error_status = error_status

if error_status eq 0 and k gt scan_range(0) then begin
   print,'! firs_630_obs: file error at step ',k,' file: ',obsfiles(k)
   print,' Replacing data with previous scan step.'
   dum = dumold
endif

; save last image as backup for missing files (NSO raid has bad file
; pointers!) missing step will be replaced with pervious step
  
   if error_status eq 1 then begin
      dumold = dum
   endif
   
; do swap endian if needed
if n_elements(swap) ne 0 then dum = swap_endian(dum)
dum = float(dum)


; check size of image in raw data. If only spectroscopy, blow up to
; QUV == 0
 ee = size(dum)
 if k eq scan_range(0) then print,'Dimensions of data: ',ee
 
 if ee(3) eq 8 then begin
    dum(*,*,0) = .5*( dum(*,*,0)+dum(*,*,4))
    dum(*,*,1) = .5*( dum(*,*,1)+dum(*,*,5))
    dum(*,*,2) = .5*( dum(*,*,2)+dum(*,*,6))
    dum(*,*,3) = .5*( dum(*,*,3)+dum(*,*,7))
    dum = dum(*,*,0:3)
 endif
 
 if ee(0) eq 2 then begin
    ddum = fltarr(ee(1),ee(2),4)
    ddum(*,*,0) = dum
    ddum(*,*,1) = dum
    ddum(*,*,2) = dum
    ddum(*,*,3) = dum
    dum = ddum
    int_only = 1.
 endif
     
   getiquv,dum,ii,qq,uu,vv
   ii = float(ii)
   ii = ii - float(dc)
   ff = where(ii lt 0)
   if ff(0) ne -1 then ii(ff) = 1.
   ff = where(finite(ii) ne 1)
   if ff(0) ne -1 then ii(ff) = 1.
  
   dum(*,*,0) = ii
   dum(*,*,1) = float(qq)
   dum(*,*,2) = float(uu)
   dum(*,*,3) = float(vv)
    

;if int_only eq 1 and k eq scan_range(0) then begin
; If only spectroscopy, blow up dark current to QUV == 0
;   ddc =  fltarr(ee(1),ee(2),4)
;   eee = size(ddc)
;   ddd = size(dc)
;   if total(ddd-eee) ne 0 then begin
;      ddc(*,*,0) = dc
;      ddc(*,*,1) = dc
;      ddc(*,*,2) = dc
;      ddc(*,*,3) = dc
;   endif else ddc = dc
;   dc = ddc

; If only spectroscopy, blow up gaintable to QUV == 0
;   ww = size(gaintable)
;   ggaintable = fltarr(ww(1),ww(2),ww(3),ww(4),4)
;   www = size(ggaintable)
;   if total(ww-www) ne 0 then begin
;      ggaintable(*,*,*,*,0) = gaintable
;      ggaintable(*,*,*,*,1) = gaintable
;      ggaintable(*,*,*,*,2) = gaintable
;      ggaintable(*,*,*,*,3) = gaintable
;      endif else ggaintable = gaintable
;   gaintable = ggaintable
;endif
;if n_elements(nodemod) ne 0 then begin
;   getiquv,dum,ii,qq,uu,vv
;   dum(*,*,0) = ii
;   dum(*,*,1) = qq
;   dum(*,*,2) = uu
;   dum(*,*,3) = vv
;   dum(*,*,0) = dum(*,*,0) - dc(*,*,0)
;endif else dum = dum - dc
; subtract dark current
; 

; define array to store beams  
 ibeams = fltarr(n_slits,2,dx+1,dy+1,4)

; cut out the part beams for n slits
for kk = 0,n_slits-1 do begin &  ii_temp = reform(dum(*,*,0)) & ibeams(kk,0,*,*,0) = ii_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1))  & ibeams(kk,1,*,*,0) = ii_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & qq_temp = reform(dum(*,*,1)) & ibeams(kk,0,*,*,1) = qq_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,1) = qq_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & uu_temp = reform(dum(*,*,2)) & ibeams(kk,0,*,*,2) = uu_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,2) = uu_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & vv_temp = reform(dum(*,*,3)) & ibeams(kk,0,*,*,3) = vv_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,3) = vv_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & endfor

; cut out the aligned part beams for n slits
   ibeams1 = fltarr(n_slits,2,dx_final+1,dy_final+1,4)

for kk = 0,n_slits-1 do begin & ibeams1(kk,0,*,*,*) =  ibeams(kk,0,posx_final(kk,0):posx_final(kk,1), posy_final(kk,0): posy_final(kk,1),*) & ibeams1(kk,1,*,*,*) =  ibeams(kk,1,posx_final(kk,2):posx_final(kk,3), posy_final(kk,2): posy_final(kk,3),*) & endfor

; apply gain correction
for kk = 0,n_slits -1 do begin &  for kkk = 0,3 do begin & ibeams1(kk,0,*,*,kkk) =  ibeams1(kk,0,*,*,kkk) / gaintable(kk,0,*,*) & ibeams1(kk,1,*,*,kkk) =  ibeams1(kk,1,*,*,kkk) / gaintable(kk,1,*,*)  & endfor & endfor

; apply gain correction
; for kk = 0,n_slits -1 do begin &  for kkk = 0,3 do begin & ibeams1(kk,0,*,*,kkk) =  ibeams1(kk,0,*,*,kkk) /gaintable(kk,0,*,*,kkk) & ibeams1(kk,1,*,*,kkk) =  ibeams1(kk,1,*,*,kkk) / gaintable(kk,1,*,*,kkk)  & endfor & endfor

; apply beam rotation
 for kk = 0,n_slits -1 do begin &  for kkk = 0,3 do begin & ibeams1(kk,0,*,*,kkk) =  rot(reform(ibeams1(kk,0,*,*,kkk)),aangle(kk),/interp) & ibeams1(kk,1,*,*,kkk) =  rot(reform(ibeams1(kk,1,*,*,kkk)),aangle(kk),/interp)  & endfor & endfor

; at first scan step, determine location of spectral lines

if ( (k-scan_range(0))) eq 0 then begin

 ; determine location of spectral lines

; average profile
 iline = total(reform(ibeams1(0,0,*,*,0)),2)
 iline = abs(iline/float(max(iline))-1)

; search for lines automatically 
 n = length(iline)
 ilobes = fltarr(n-4)
 threshh = .2
 ilobeindex = 0.
 for kkk = 5,n-5 do begin
    if iline(kkk) gt max(iline(kkk+1:kkk+4)) and iline(kkk) gt max(iline(kkk-4:kkk-1)) and iline(kkk) gt threshh then begin
       ilobes(ilobeindex) = kkk
       ilobeindex = ilobeindex+1
    endif
 endfor
 
 if n_elements(verbose) ne 0 then print,ilobeindex,' spectral lines found.'

 del17 = 0.


; location of solar lines for |QUV| maps
 if n_elements(pos_lines) ne 0 then begin
    ilobes = pos_lines
    ilobeindex = n_elements(ilobes)

 endif
 
; mark center of lines manually
 if ilobeindex ne 4 or n_elements(check_lines) ne 0 then begin
    print,' Automatic detection of spectral lines failed.'
    print,' Please select lines manually:'
    window,17
    del17 = 1.
    !p.multi = 0
    plot,iline 
    ilobeindex = 0.
    ilobes = fltarr(n-4)
    while ilobeindex ne 4 do begin
       wset,17
       cursor,x,y,/data,/wait
       if !mouse.button eq 1 then begin
          ilobes(ilobeindex) = x
          ilobeindex = ilobeindex+1
          print,'Line ',ilobeindex
          print,'at ',x
          oplot,[x,x],[0,1]
          wait,.5
       endif
       if ilobeindex eq 4 then begin
          ans = read_keyboard('Selection okay y(es)/n(o) ')
          if ans eq 'n' then begin
             ilobeindex = 0
             plot,iline    
          endif
       endif
    endwhile
 endif
 ilobes = ilobes(0:ilobeindex-1)
 ilobes = ilobes(sort(ilobes))

 pos_lines = ilobes

; location of telluric lines at 630.27 nm 
 range_telu = [-15,15]+ilobes(1)
 
 ilo = range_telu(0)
 ihi = range_telu(1)

 pos_solar = [ilobes(2)-50,ilobes(2)+50.]

      if n_elements(hair_up) eq 0 and n_elements(no_align) eq 0 then begin

; average profile
         iline = total(transpose(reform(ibeams1(0,0,*,*,0)*gaintable(0,0,*,*))),2)

         if max(iline) eq 0 then iline = iline + 1.
         iline = abs(iline/float(max(iline))-1)
   
                                ; search for hair lines automatically 
         n = length(iline)
         ilobes = fltarr(n-4)
         threshh = .5
         ilobeindex = 0.
         for kkk = 5,n-5 do begin
            if iline(kkk) gt max(iline(kkk+1:kkk+4)) and iline(kkk) gt max(iline(kkk-4:kkk-1)) and iline(kkk) gt threshh then begin
               ilobes(ilobeindex) = kkk
               ilobeindex = ilobeindex+1
            endif
         endfor
   
         if n_elements(verbose) ne 0 then print,ilobeindex,'  hairline found.'
   
         wdel17 = 0.
         ans = 1.
; mark center of lines manually
         if ilobeindex eq 0 or ilobeindex gt 2 or n_elements(check_hairlines) ne 0 then begin
            print
            print,' Automatic detection of hairlines failed.'
            ans = read_keyboard('How many hairlines 1/2 : ',/number)
            if ans eq 'nonumber' or ans eq 0 then ans = 1.
            print,' Please select hair lines manually:'
            window,17,title ='Please mark hair lines'
            wdel17 = 1.
            !p.multi = 0
            plot,iline 
            ilobeindex = 0.
            ilobes = fltarr(n-4)
            while ilobeindex ne ans do begin
               wset,17
               cursor,x,y,/data,/wait
               if !mouse.button eq 1 then begin
                  ilobes(ilobeindex) = x
                  ilobeindex = ilobeindex+1
                  print,'Hairline ',ilobeindex
                  print,'at ',x
                  oplot,[x,x],[0,1]
                  wait,.5
               endif
               if ilobeindex eq ans then begin
                  aans = read_keyboard('Selection okay y(es)/n(o) ')
                  if aans eq 'n' then begin
                     ilobeindex = 0
                     plot,iline    
                  endif
               endif
            endwhile
         endif
         ilobes = ilobes(0:ilobeindex-1)
         ilobes = ilobes(sort(ilobes))
   
   ; location of spectral hair line to be used for alignment purposes
         if ans eq 1 then hair_up = [ilobes(0)-5,ilobes(0)+5] else begin
            hair_up = [ilobes(1)-5,ilobes(1)+5]
            hair_down = [ilobes(0)-5,ilobes(0)+5]
            if hair_down(0) lt 0 then hair_down(0) = 0
         endelse
         if hair_up(1) gt dy_final then hair_up(1) = dy_final

         if n_elements(plt) ne 0 then begin
            window,17,title ='Hairlines'
            plot,iline
            oplot,[0,0]+hair_up(0),[0,2]
            oplot,[0,0]+hair_up(1),[0,2]
            if ans eq 2 then begin
               oplot,[0,0]+hair_down(0),[0,2]
               oplot,[0,0]+hair_down(1),[0,2]
            endif
            print,ans,' hairlines found.'
            if n_elements(nopause) eq 0 then paused
         endif
      endif
endif

; correction of spectral intensity gradients (optional)
if n_elements(spec_grad) ne 0 then begin

xx = [ilobes(0)-80+findgen(40),ilobes(1)-25+findgen(12),ilobes(1)+12+findgen(12),ilobes(3)+10+findgen(35)]

for kk = 0,n_slits-1 do begin & for kkk = 0,1 do begin & for kkkk = 0,3 do begin & ii = reform(ibeams1(kk,kkk,*,*,kkkk)) & for kkkkk = 0,dy_final do begin & iline = reform(ii(*,kkkkk)) & ergg = poly_fit(xx+oofset(kk),iline(xx+oofset(kk)),4,/double) & yy = ergg(0)+ergg(1)*findgen(dx_final+1)+ergg(2)*findgen(dx_final+1)^2.+ergg(3)*findgen(dx_final+1)^3.+ergg(4)*findgen(dx_final+1)^4. & yy = yy /mean(yy) & iline = iline / yy & ii(*,kkkkk) = iline & endfor & ibeams1(kk,kkk,*,*,kkkk) = ii & endfor & endfor & endfor 

endif

;  remove the curvature of the spectrum using the telluric line at
;  630.27 nm
for kk = 0,n_slits -1 do begin & i1 = reform(ibeams1(kk,0,*,*,0)) & q1 = reform(ibeams1(kk,0,*,*,1))& u1 = reform(ibeams1(kk,0,*,*,2))& v1 = reform(ibeams1(kk,0,*,*,3)) &  i2 = reform(ibeams1(kk,1,*,*,0)) &  q2 = reform(ibeams1(kk,1,*,*,1)) &  u2 = reform(ibeams1(kk,1,*,*,2)) &  v2 = reform(ibeams1(kk,1,*,*,3)) &  deskew1,i1,q1,u1,v1,ilo+oofset(kk),ihi+oofset(kk),0,dy_final,0,dy_final,3,/third_order & deskew1,i2,q2,u2,v2,ilo+oofset(kk),ihi+oofset(kk),0,dy_final,0,dy_final,3,/third_order & ibeams1(kk,0,*,*,0) = i1 & ibeams1(kk,1,*,*,0) = i2 & ibeams1(kk,0,*,*,1) = q1 & ibeams1(kk,1,*,*,1) = q2 & ibeams1(kk,0,*,*,2) = u1 & ibeams1(kk,1,*,*,2) = u2 & ibeams1(kk,0,*,*,3) = v1 & ibeams1(kk,1,*,*,3) = v2  & endfor

poslines = fltarr(n_slits,2,dy_final+1,4)
for kk = 0,n_slits -1 do begin & for kkk = 0,0 do begin & for kkkkk = 0,1 do begin & i1 = reform(ibeams1(kk,kkkkk,*,*,kkk)) & for kkkk = 0,dy_final do begin & iline = reform(i1(*,kkkk)) & ergg = min(iline(ilo+oofset(kk):ihi+oofset(kk)),posmin) & lpff,iline(posmin+ilo+oofset(kk)-5:posmin+ilo+oofset(kk)+5),ppos & poslines(kk,kkkkk,kkkk,kkk) = ppos+posmin+ilo+oofset(kk)-5 & endfor & endfor & endfor & endfor

for kk =  0,n_slits -1 do begin & ref_curve = reform(poslines(kk,0,*,0)) & for kkk = 0,1 do begin & for kkkk = 0,3 do begin & ccurve = reform(poslines(kk,kkk,*,0)) & diff = ref_curve - ccurve &  ergg = poly_fit(findgen(dy_final+1),diff,5,/double) &  yy = 0. & for oo = 0.,5 do begin & yy = yy + findgen(dy_final+1)^float(oo)*ergg(oo) & endfor & i1 = reform(ibeams1(kk,kkk,*,*,kkkk)) & sshift_img,i1,yy,0,dy_final-1 & ibeams1(kk,kkk,*,*,kkkk) = i1 & endfor & endfor & endfor

; check position of telluric line afterwards
poslines = fltarr(n_slits,2,dy_final+1,4)
for kk = 0,n_slits -1 do begin & for kkk = 0,0 do begin & for kkkkk = 0,1 do begin & i1 = reform(ibeams1(kk,kkkkk,*,*,kkk)) & for kkkk = 0,dy_final do begin & iline = reform(i1(*,kkkk)) & ergg = min(iline(ilo+oofset(kk):ihi+oofset(kk)),posmin) & lpff,iline(posmin+ilo+oofset(kk)-5:posmin+ilo+oofset(kk)+5),ppos & poslines(kk,kkkkk,kkkk,kkk) = ppos+posmin+ilo+oofset(kk)-5 & endfor & endfor & endfor & endfor

for kk = 0,n_slits -1 do begin & llinepos(kk,k-scan_range(0),0) = mean(poslines(kk,0,*,0)) & llinepos(kk,k-scan_range(0),1) = mean(poslines(kk,1,*,0)) & endfor



; do precise subpixel alignment using hairlines

if n_elements(no_align) eq 0 then begin

   for kk = 0,n_slits -1 do begin 

      im1a = reform(ibeams1(kk,0,*,*,0))
      im2a = reform(ibeams1(kk,0,*,*,1))
      im3a = reform(ibeams1(kk,0,*,*,2))
      im4a = reform(ibeams1(kk,0,*,*,3))

      im1b = reform(ibeams1(kk,1,*,*,0))
      im2b = reform(ibeams1(kk,1,*,*,1))
      im3b = reform(ibeams1(kk,1,*,*,2))
      im4b = reform(ibeams1(kk,1,*,*,3))

; align beams with subpixel precision
      i1atemp  = im1a*gaintable(kk,0,*,*)
      i1atemp  = transpose(i1atemp)

      im1atemp = transpose(im1a)
      im2atemp = transpose(im2a)
      im3atemp = transpose(im3a)
      im4atemp = transpose(im4a)
     
      i1btemp  = im1b*gaintable(kk,1,*,*)
      i1btemp  = transpose(i1btemp)

      im1btemp = transpose(im1b)
      im2btemp = transpose(im2b)
      im3btemp = transpose(im3b)
      im4btemp = transpose(im4b)

; align in spatial direction with hair line 
      deskew1,i1atemp, im2atemp, im3atemp, im4atemp,hair_up(0),hair_up(1),0,dx_final,0,dx_final,2,/straight
      i1atemp  = im1a*gaintable(kk,0,*,*)
      i1atemp  = transpose(i1atemp)
      dum =  i1atemp -  i1atemp 
      deskew1,i1atemp,im1atemp,dum,dum,hair_up(0),hair_up(1),0,dx_final,0,dx_final,2,/straight
  
      wave_shift_spinor,i1atemp,i1btemp, im2btemp, im3btemp, im4btemp,hair_up(0),hair_up(1),0,dx_final,2,/straight
      i1btemp  = im1b*gaintable(kk,1,*,*)
      i1btemp  = transpose(i1btemp)
      wave_shift_spinor,i1atemp,i1btemp,im1btemp,dum,dum,hair_up(0),hair_up(1),0,dx_final,2,/straight

; align in spectral direction with spectral line
      im1a = transpose(im1atemp)
      im2a = transpose(im2atemp)
      im3a = transpose(im3atemp)
      im4a = transpose(im4atemp)
     
      im1b = transpose(im1btemp)
      im2b = transpose(im2btemp)
      im3b = transpose(im3btemp)
      im4b = transpose(im4btemp)
 ;      wave_shift_spinor,im1a,im1b,im2b,im3b,im4b,ilo,ihi,0,dy_final,2

      ibeams1(kk,0,*,*,0) = im1a 
      ibeams1(kk,0,*,*,1) = im2a 
      ibeams1(kk,0,*,*,2) = im3a 
      ibeams1(kk,0,*,*,3) = im4a 

      ibeams1(kk,1,*,*,0) = im1b 
      ibeams1(kk,1,*,*,1) = im2b 
      ibeams1(kk,1,*,*,2) = im3b 
      ibeams1(kk,1,*,*,3) = im4b 
  
   endfor

endif 


; storage array for merged beams
   ibeams_merged =  fltarr(n_slits,dx_final+1,dy_final+1,4)

; apply demodulation
   for kk = 0,n_slits -1 do begin
      ;xinv = reform(xxinv(kk,*,*))
      ;dmod = xinv

      xinv = reform(full_xxinv(kk,*,*,*))
      dmod = xinv


      im1a = reform(ibeams1(kk,0,*,*,0))
      im2a = reform(ibeams1(kk,0,*,*,1))
      im3a = reform(ibeams1(kk,0,*,*,2))
      im4a = reform(ibeams1(kk,0,*,*,3))
            
;      im1b = reform(ibeams1(kk,0,*,*,0))
;      im2b = reform(ibeams1(kk,0,*,*,0))
;      im3b = reform(ibeams1(kk,0,*,*,0))
;      im4b = reform(ibeams1(kk,0,*,*,0))

      im1b = reform(ibeams1(kk,1,*,*,0))
      im2b = reform(ibeams1(kk,1,*,*,1))
      im3b = reform(ibeams1(kk,1,*,*,2))
      im4b = reform(ibeams1(kk,1,*,*,3))

      i1a = im1a
      i2a = im2a
      i3a = im3a
      i4a = im4a
      
      i1b = im1b
      i2b = im2b
      i3b = im3b
      i4b = im4b
      
      if n_elements(no_align) ne 0 then begin
         erg = shc(i1a,i1b,/filt,/interp)
         if total(abs(erg)) lt 10 then begin
            i1b = shift_frac(i1b,erg(0),erg(1),'C')
            i2b = shift_frac(i2b,erg(0),erg(1),'C')
            i3b = shift_frac(i3b,erg(0),erg(1),'C')
            i4b = shift_frac(i4b,erg(0),erg(1),'C')
         endif
      endif

; merge the beams to convert to IQUV
      ii1 = (i1a+i1b)/2.
      qq1 = (i2a-i2b)/2.
      uu1 = (i3a-i3b)/2.
      vv1 = (i4a-i4b)/2.

; capture intensity spikes along the slit that come from flat field residuals
      if smooth_width eq 4642. then begin

         if n_slits gt 1 then begin
            posslit = mean(poslines(kk,*,*,*))
            if n_elements(cont_wl) eq 0 then ccont_wl = [posslit - 15,posslit + 15] else ccont_wl = cont_wl
            
         endif else begin

            if n_elements(cont_wl) eq 0 then ccont_wl = [ilobes(1)-15,ilobes(1)+15] else ccont_wl = cont_wl
         endelse

  ; spike correction
      ff1a = total(im1a(ccont_wl(0):ccont_wl(1),*),1)/float(ccont_wl(1)-ccont_wl(0)+1.)
      ff2a = total(im2a(ccont_wl(0):ccont_wl(1),*),1)/float(ccont_wl(1)-ccont_wl(0)+1.)
      ff3a = total(im3a(ccont_wl(0):ccont_wl(1),*),1)/float(ccont_wl(1)-ccont_wl(0)+1.)
      ff4a = total(im4a(ccont_wl(0):ccont_wl(1),*),1)/float(ccont_wl(1)-ccont_wl(0)+1.)
      ff1b = total(im1b(ccont_wl(0):ccont_wl(1),*),1)/float(ccont_wl(1)-ccont_wl(0)+1.)
      ff2b = total(im2b(ccont_wl(0):ccont_wl(1),*),1)/float(ccont_wl(1)-ccont_wl(0)+1.)
      ff3b = total(im3b(ccont_wl(0):ccont_wl(1),*),1)/float(ccont_wl(1)-ccont_wl(0)+1.)
      ff4b = total(im4b(ccont_wl(0):ccont_wl(1),*),1)/float(ccont_wl(1)-ccont_wl(0)+1.)

      corr1a = ff1a / smooth(ff1a,smooth_width)
      corr2a = ff2a / smooth(ff2a,smooth_width)
      corr3a = ff3a / smooth(ff3a,smooth_width)
      corr4a = ff4a / smooth(ff4a,smooth_width)
      corr1b = ff1b / smooth(ff1b,smooth_width)
      corr2b = ff2b / smooth(ff2b,smooth_width)
      corr3b = ff3b / smooth(ff3b,smooth_width)
      corr4b = ff4b / smooth(ff4b,smooth_width)
 
; apply spike correction     
       for kkk = 0,dx_final do begin &  im1a(kkk,*) =  im1a(kkk,*) / corr1a  &im2a(kkk,*) =  im2a(kkk,*) / corr2a  &im3a(kkk,*) =  im3a(kkk,*) / corr3a  &im4a(kkk,*) =  im4a(kkk,*) / corr4a  & im1b(kkk,*) =  im1b(kkk,*) / corr1b  & im2b(kkk,*) =  im2b(kkk,*) / corr2b  & im3b(kkk,*) =  im3b(kkk,*) / corr3b  &im4b(kkk,*) =  im4b(kkk,*) / corr4b & endfor
       endif

      ii1_temp = ii1-ii1
      qq1_temp = ii1-ii1
      uu1_temp = ii1-ii1
      vv1_temp = ii1-ii1

; apply inverse response function
      for kkk =0,dy_final do begin
         ii1_temp(*,kkk) = full_xxinv(kk,kkk,0,0)*ii1(*,kkk)+full_xxinv(kk,kkk,1,0)*qq1(*,kkk)+full_xxinv(kk,kkk,2,0)*uu1(*,kkk)+full_xxinv(kk,kkk,3,0)*vv1(*,kkk)
         qq1_temp(*,kkk) = full_xxinv(kk,kkk,0,1)*ii1(*,kkk)+full_xxinv(kk,kkk,1,1)*qq1(*,kkk)+full_xxinv(kk,kkk,2,1)*uu1(*,kkk)+full_xxinv(kk,kkk,3,1)*vv1(*,kkk)
         uu1_temp(*,kkk) = full_xxinv(kk,kkk,0,2)*ii1(*,kkk)+full_xxinv(kk,kkk,1,2)*qq1(*,kkk)+full_xxinv(kk,kkk,2,2)*uu1(*,kkk)+full_xxinv(kk,kkk,3,2)*vv1(*,kkk)
         vv1_temp(*,kkk) = full_xxinv(kk,kkk,0,3)*ii1(*,kkk)+full_xxinv(kk,kkk,1,3)*qq1(*,kkk)+full_xxinv(kk,kkk,2,3)*uu1(*,kkk)+full_xxinv(kk,kkk,3,3)*vv1(*,kkk)
      endfor

      ii1 = ii1_temp
      qq1 = qq1_temp
      uu1 = uu1_temp
      vv1 = vv1_temp

      vold = vv1
      uold = uu1
      qold = qq1
      

; apply telescope correction
              if n_elements(notelcor) eq 0 and n_elements(tel_cal) eq 0 then begin
; if n_elements(tel_cal) eq 0 then begin

         tel_azimuth = FXPAR(hdr,'DST_AZ')
         tel_elevation = FXPAR(hdr,'DST_EL')
         table_angle = FXPAR(hdr,'DST_TBL')
         tel_p = FXPAR(hdr,'DST_PEE')

;         print
;         print, table_angle,tel_p,tel_azimuth,tel_elevation,tparams(3)
;         print

; get telescope matrix
         tmtx  = ibis_tmtx(tparams, wavelength, tel_azimuth, tel_elevation, table_angle)
         tmtx = tmtx/tmtx(0,0)

; strange flip of sign...
;         vv1 = -vv1

         Tinv = INVERT(tmtx,/double) 
         Tinv = float(Tinv)

; apply inverse telescope matrix      
         ii2  = Tinv(0,0)*ii1+Tinv(1,0)*qq1+Tinv(2,0)*uu1+Tinv(3,0)*vv1
         qq2  = Tinv(0,1)*ii1+Tinv(1,1)*qq1+Tinv(2,1)*uu1+Tinv(3,1)*vv1
         uu2  = Tinv(0,2)*ii1+Tinv(1,2)*qq1+Tinv(2,2)*uu1+Tinv(3,2)*vv1
         vv2  = Tinv(0,3)*ii1+Tinv(1,3)*qq1+Tinv(2,3)*uu1+Tinv(3,3)*vv1
     

; correct QU for parallactic angle
         DST_earth_latitude = 32.786 * !dtor
         r_frame_sphtri, DST_earth_latitude , tel_azimuth * !dtor, tel_elevation * !dtor,siteha, sitedc, neg_parallactic
         parallactic = -neg_parallactic

         rotn = !pi + parallactic - tel_p * !dtor
         crot = cos(-2.*rotn) 
         srot = sin(-2.*rotn)

         temp0 =  crot*qq2 + srot*uu2
         temp1 = -srot*qq2 + crot*uu2

         ii1 = ii2
         qq1 = temp0
         uu1 = temp1
         vv1 = vv2

      endif
  
; remove intensity < 0
      ff = where(ii1 lt 0)
      if ff(0) ne -1 then ii1(ff) = 0.

; store calibrated IQUV 
      ibeams_merged(kk,*,*,0) = ii1
      ibeams_merged(kk,*,*,1) = qq1
      ibeams_merged(kk,*,*,2) = uu1
      ibeams_merged(kk,*,*,3) = vv1
   endfor

; correct for I -> QUV cross talk
   for kk = 0,n_slits -1 do begin
     
      ii1 = reform( ibeams_merged(kk,*,*,0))

                                ; correct I for parasitic light. Must be
                                ; done before I->QUV cross-talk
                                ; correction !
      if n_elements(beta) ne 0 then ii1 = ii1 - beta*mean(ii1)

      qq1 = reform( ibeams_merged(kk,*,*,1))
      uu1 = reform( ibeams_merged(kk,*,*,2))
      vv1 = reform( ibeams_merged(kk,*,*,3))

; define wavelength range for I->QUV cross-talk correction
       if n_slits gt 1 then begin
          posslit = mean(poslines(kk,*,*,*))
          if n_elements(cont_wl) eq 0 then ccont_wl = [posslit - 15,posslit + 15] else ccont_wl = cont_wl
       endif else begin
          if n_elements(cont_wl) eq 0 then ccont_wl = [ilobes(1)-15,ilobes(1)+15] else ccont_wl = cont_wl
       endelse

; correct I->QUV cross-talk row by row    
      for kkk = 0,dy_final do begin
         iline = float(reform(ii1(*,kkk)))
         qline = reform(qq1(*,kkk))
         uline = reform(uu1(*,kkk))
         vline = reform(vv1(*,kkk))

; deactivate  I->QUV cross-talk correction for /telcal
if n_elements(tel_cal) eq 0 then begin
      
         i2q = mean(qline(ccont_wl(0):ccont_wl(1))/iline(ccont_wl(0):ccont_wl(1)))
         i2u = mean(uline(ccont_wl(0):ccont_wl(1))/iline(ccont_wl(0):ccont_wl(1)))
         i2v = mean(vline(ccont_wl(0):ccont_wl(1))/iline(ccont_wl(0):ccont_wl(1)))
           
         qline = qline - i2q*iline
         uline = uline - i2u*iline
         vline = vline - i2v*iline

endif

; correct additional V->QU cross-talk
         if n_elements(v2q) ne 0 then begin
            qline = qline - v2q*vline
         endif
         if n_elements(v2u) ne 0 then begin
            uline = uline - v2u*vline
         endif

         qq1(*,kkk) = qline
         uu1(*,kkk) = uline
         vv1(*,kkk) = vline

      endfor

; shift all beams to the same spectral position
      if n_slits gt 1 then begin
;         posslit = mean(poslines(kk,*,*,*))
         posslit = mean(poslines(kk,*,*,0))

         ii1 = shift_frac(ii1,posref- posslit  ,0,'C')
         qq1 = shift_frac(qq1,posref- posslit,0,'C')
         uu1 = shift_frac(uu1,posref- posslit ,0,'C')
         vv1 = shift_frac(vv1,posref- posslit ,0,'C')
         
      endif

      ibeams_merged(kk,*,*,0) = ii1
      ibeams_merged(kk,*,*,1) = qq1
      ibeams_merged(kk,*,*,2) = uu1
      ibeams_merged(kk,*,*,3) = vv1

; do plotting
      if n_elements(plt) ne 0 then begin
         wset,kk
         tvscl,congrid(ii1,dx_final,500,/interp) >0 <10000,0,0
         tvscl,congrid(qq1/ii1,dx_final,500,/interp) >(-.2) < .2,dx_final,0
         tvscl,congrid(uu1/ii1,dx_final,500,/interp)>(-.2) < .2,dx_final*2,0
         tvscl,congrid(vv1/ii1,dx_final,500,/interp)>(-.2) < .2,dx_final*3,0
      endif

; create overview maps
      ic(kk*nsteps_total+k-scan_range(0),*) = total(ii1(ccont_wl(0):ccont_wl(1),*),1)/float(ccont_wl(1)-ccont_wl(0)+1.)

      totq(kk*nsteps_total+k-scan_range(0),*) = total( abs ( qq1(pos_solar(0):pos_solar(1),*)/ii1(pos_solar(0):pos_solar(1),*)),1)
      totu(kk*nsteps_total+k-scan_range(0),*) = total( abs ( uu1(pos_solar(0):pos_solar(1),*)/ii1(pos_solar(0):pos_solar(1),*)),1)
      totv(kk*nsteps_total+k-scan_range(0),*) = total( abs ( vv1(pos_solar(0):pos_solar(1),*)/ii1(pos_solar(0):pos_solar(1),*)),1)
   endfor

; do plotting
   if n_elements(plt) ne 0 then begin
      wset,4
      tvscl,congrid(ic,n_slits*nsteps_total,300,/interp) > 10 < 10000.,0,0
      tvscl,alog(congrid(totq,n_slits*nsteps_total,300,/interp) > 1),n_slits*nsteps_total,0
      tvscl,alog(congrid(totu,n_slits*nsteps_total,300,/interp) > 1),0,300
      tvscl,alog(congrid(totv,n_slits*nsteps_total,300,/interp) > 1),n_slits*nsteps_total,300
   endif

; for spectroscopic data, kick out QUV again
   if int_only eq 1 then begin      
      ibeams_merged = reform(ibeams_merged(*,*,*,0))
   endif

; write out data
   writeu,outunit,ibeams_merged

; timting information
   print,' Scan steps to go: ',scan_range(1)- k
   newt = systime(1)
   secsperstep = newt - oldt
   full_time = secsperstep*(scan_range(1)- k)
   print,' Time per scan step: ', secsperstep,' s'
   print,' Estimated time of termination: ',floor(full_time/3600),' h',floor( (full_time - floor(full_time/3600)*3600)/60 ),' min',round(full_time mod 60),' s'
   
endfor

; close output file
free_lun,outunit

; save overview maps and other info
save,ic,totq,totu,totv,sshifts,v2q,v2u,llinepos,pos_solar,dx_final,dy_final,n_slits,int_only,ccont_wl,filename = outfile2

; delete windows
if del17 eq 1 then wdelete,17

end

