pro firs_cal,spat_res =spat_res ,no_av = no_av,plt =plt,ilimit = ilimit,retard = retard,theta_err = theta_err,theta_pol = theta_pol,gainindex = gainindex,darkindex = darkindex,cal_file = cal_file,change_save = change_save,input_dir = input_dir,output_dir = output_dir,flat_file = flat_file,swap = swap,add_cal = add_cal,nverbose = nverbose,cchi = cchi,n_separations = n_separations,nopause = nopause,quit_column = quit_column,help = help

;+
;============================================================================
;
;	procedure : firs_cal 
;
;	purpose   : calculate FIRS instrument response 
;
;	author    : cbeck@nso Apr 2013
;
;==============================================================================
;
;	Check number of arguments.
;
;==============================================================================
if n_elements(help) ne 0 then begin
	print
	print, "usage:   firs_cal,spat_res =spat_res ,no_av = no_av,plt =plt,ilimit = ilimit,retard = retard,theta_err = theta_err,theta_pol = theta_pol,gainindex = gainindex,darkindex = darkindex,cal_file = cal_file,change_save = change_save,input_dir = input_dir,output_dir = output_dir,flat_file = flat_file,swap = swap,add_cal = add_cal,nverbose = nverbose,cchi = cchi,n_separations = n_separations,nopause = nopause,quit_column = quit_column,help = help"
	print
	print, "	 Calculate FIRS polarimetric instrument response."
        print, "         "
	print
	print, "	Arguments"
        print, "                none"
        print
        print, "         Keywords :"
        print, "              /plt : switch on plotting (recommended)"

        print, "              spat_res: use subfields in spectra to get"
        print, "                      spatially resolved X matrices"
        print, "                      spat_res=[nx,by]; nx,ny number of subfields" 
        print, "                      in x, y dimension. For tests only."
        print, "              /no_av: do not average the measurements, use"
        print, "                      previously saved polarization vectors"
        print, "              ilimit: required intensity limit to use pixels"
        print, "                      in the calculation, fraction of imean"
        print, "              retard: retardance value to be used in degree"
        print, "              theta_err: offset angle between retarder and"
        print, "                      polarizer, in degree"
        print, "              theta_pol: position angle of polarizer during"
        print, "                      the calibration, in degree"
        print, "              add_cal: number of additional +-QU calibration"
        print, "                      files to be used in the source directory."
        print, "              /nverbose: suppress terminal output"
        print, "              cchi   : chi square value of additional +-QU"
        print, "                       vectors, measures reproduction of input"
        print, "              gainindex:number of gain table to be used for the"
        print, "                        reduction in the target directory"
        print, "              darkindex:number of dark current to be used for the"
        print, "                        reduction in the target directory"
        print, "              cal_file: number of calibration file to be used"
        print, "                        in the source directory."
        print, "              change_save: directory where the file information"
        print, "                        on the source directory was saved to."
        print, "              input_dir: source directory with raw data."
        print, "              output_dir: destination directory for reduced data."
        print, "              flat_file: use a different gain table."
        print, "                         String with the complete filepath and"
        print, "                         filename of the gain.....sav file."
        print, "               /swap   : do a swap endian of the raw data."
        print, "               n_separations: number of subdivisions along the slit."
        print, "               /nopause: ignore all stop/pause commands."
        print, "               /quit_column:switch to replace missing column at"
        print, "                         middle of IR CCD chip. Cosmetics."
        print, "               /help   : switch to active the help text above."
        print
return
endif
;==============================================================================
;-


; set default file indices
;if n_elements(med_with) eq 0 then  med_width = 2
if n_elements(darkindex) eq 0 then  darkindex = 0
if n_elements(cal_file) eq 0 then  cal_file = 0
if n_elements(gainindex) eq 0 then gainindex = 0

; set minimum and maximum intensity limits
if n_elements(ilimit) eq 0 then ilimit = .5
if n_elements(maxi)  eq 0 then maxi = 1.5 

; number of subdivisions along the slit
if n_elements(n_separations) eq 0 then n_separations = 10.

; default ICU properties
if n_elements(retard) eq 0 then retard = 90.
if n_elements(theta_err) eq 0 then theta_err = 0.
if n_elements(theta_pol) eq 0 then theta_pol = 0.


if n_elements(change_save) ne 0 then erg = read_directory(change_save,filter ='*.sav') else erg = read_directory(input_dir,filter ='*.sav')

; restore file with data information
outfile = erg.files
if n_elements(change_save) eq 0 then begin
   gfile = input_dir+outfile(0)  
   erg = check_existence(gfile,/read,/change)
   restore,erg(0)  
endif else begin
   gfile = change_save+outfile(0)
   erg = check_existence(gfile,/read,/change)
   restore,erg(0)  
endelse

;if n_elements(change_save) eq 0 then restore,input_dir+outfile(0) else restore,change_save+outfile(0)

; restore averaged dark current file
ergdc = read_directory(output_dir,filter ='*dc*')

erg = check_existence(output_dir + ergdc.files(darkindex),/read,/change)
restore,erg(0)

;output_dir + ergdc.files(darkindex)

; restore flat file, but it is not used anyway
if n_elements(flat_file) eq 0 then begin
   erggain = read_directory(output_dir,filter ='*gain*')
   erg = check_existence(output_dir + erggain.files(gainindex),/read,/change)
;output_dir + erggain.files(gainindex)
   restore,erg(0)
endif else begin
   erg = check_existence(flat_file,/read,/change)
   restore,erg(0)
endelse

; calibration data

calbasefile = bbasefile1(cal_file)
erg = read_directory(input_dir,filter =calbasefile+'*')
calfiles = erg.files(where(erg.types eq 0))
n = float(n_elements(calfiles))

; output files for observed Stokes vectors and response function
outfile = output_dir+'polvec.'+calbasefile+'sav'
outfile1 = output_dir+'xmat.'+calbasefile+'sav'

erg = check_existence(outfile,/write,/change)
outfile = erg(0)
erg = check_existence(outfile1,/write,/change)
outfile1 = erg(0)

; do plotting
if n_elements(plt) ne 0 then begin
   window,0,title ='Observed calibration measurements'
   laodct,41
   device,decomposed = 0
endif


yy = floor(dy_final/n_separations)
yyr = findgen(dy_final/yy)*yy
yyr = [yyr,dy_final]

; read in calibration data
if n_elements(no_av) eq 0 then begin

   print,' Reading in calibration  files...',calbasefile

   ffxread,input_dir+calfiles(0),dum,hdr
   cal_img = dum - dum
   tt = size(dum)

; do spatially resolved response function
   cal_curves = fltarr(n_separations,n_slits,n,4)

; loop over files
   for k = 0,n-1 do begin
      ffxread,input_dir+calfiles(k),dum,hdr
      if n_elements(swap) ne 0 then dum = swap_endian(dum)
      dum = float(dum)
      
      if k eq 0 then print,'Dimensions of data: ',tt

      if tt(3) eq 8 then begin
         dum(*,*,0) = .5*( dum(*,*,0)+dum(*,*,4))
         dum(*,*,1) = .5*( dum(*,*,1)+dum(*,*,5))
         dum(*,*,2) = .5*( dum(*,*,2)+dum(*,*,6))
         dum(*,*,3) = .5*( dum(*,*,3)+dum(*,*,7))
      endif
      dum = dum(*,*,0:3)
   
      if n_elements(quit_column) ne 0 then begin


      for kk = 0.,39 do begin & dum(495+kk,*,*) = reform(dum(494,*,*)) + kk/39.*( reform(dum(535,*,*)-dum(494,*,*)) ) & endfor

      endif

      getiquv,dum,ii,qq,uu,vv
      ii = float(ii)
      ii = ii - float(dc)
      ff = where(ii lt 0)
      if ff(0) ne -1 then ii(ff) = 1.
      ff = where(finite(ii) ne 1)
      if ff(0) ne -1 then ii(ff) = 1.
  
; cut out beams 
      ibeams = fltarr(n_slits,2,dx+1,dy+1,4)

for kk = 0,n_slits-1 do begin &  ii_temp = ii & ibeams(kk,0,*,*,0) = ii_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1))  & ibeams(kk,1,*,*,0) = ii_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & qq_temp = qq & ibeams(kk,0,*,*,1) = qq_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,1) = qq_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & uu_temp = uu & ibeams(kk,0,*,*,2) = uu_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,2) = uu_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & vv_temp = vv & ibeams(kk,0,*,*,3) = vv_temp(posx(kk,0):posx(kk,1),posy(kk,0):posy(kk,1)) & ibeams(kk,1,*,*,3) = vv_temp(posx(kk,0):posx(kk,1),posy(kk,2):posy(kk,3)) & endfor


     ibeams1 = fltarr(n_slits,2,dx_final+1,dy_final+1,4)

  for kk = 0,n_slits-1 do begin & ibeams1(kk,0,*,*,*) =  ibeams(kk,0,posx_final(kk,0):posx_final(kk,1), posy_final(kk,0):posy_final(kk,1),*) & ibeams1(kk,1,*,*,*) =  ibeams(kk,1,posx_final(kk,2):posx_final(kk,3), posy_final(kk,2): posy_final(kk,3),*) & endfor



;if n_elements(ngain) eq 0 then begin
   ;for kk = 0,n_slits -1 do begin &  for kkk = 0,3 do begin & ibeams1(kk,0,*,*,kkk) =  ibeams1(kk,0,*,*,kkk) /gaintable(kk,0,*,*,kkk) & ibeams1(kk,1,*,*,kkk) =  ibeams1(kk,1,*,*,kkk) / gaintable(kk,1,*,*,kkk)  & endfor & endfor

;endif
; for kk = 0,n_slits -1 do begin &  for kkk = 0,3 do begin & ibeams1(kk,0,*,*,kkk) =  rot(reform(ibeams1(kk,0,*,*,kkk)),aangle(kk),/interp) & ibeams1(kk,1,*,*,kkk) =  rot(reform(ibeams1(kk,1,*,*,kkk)),aangle(kk),/interp)  & endfor & endfor
;for kk = 0,n_slits -1 do begin & ii = reform(ibeams1(kk,0,*,*,0)) & ff = where(ii lt mean(ii)*.4 or ii gt mean(ii)*1.75) & ii(ff) = mean(ii) & ibeams1(kk,0,*,*,0) = ii & ii = reform(ibeams1(kk,1,*,*,0))  & ii(ff) = mean(ii) & ibeams1(kk,1,*,*,0) = ii & for kkk = 1,3 do begin & ii = reform(ibeams1(kk,0,*,*,kkk)) & ii(ff) = mean(ii) & ibeams1(kk,0,*,*,kkk) = ii & ii = reform(ibeams1(kk,1,*,*,kkk)) & ii(ff) = mean(ii) & ibeams1(kk,1,*,*,kkk) = ii  & endfor & endfor

; merge the two beams
     ibeams_merged =  fltarr(n_slits,dx_final+1,dy_final+1,4)


for kk = 0,n_slits -1 do begin & for kkk = 0,3 do begin & if kkk eq 0 then  ibeams_merged(kk,*,*,kkk) = .5*(ibeams1(kk,0,*,*,kkk) + ibeams1(kk,1,*,*,kkk) ) else ibeams_merged(kk,*,*,kkk) =   .5* ( ibeams1(kk,0,*,*,kkk) - ibeams1(kk,1,*,*,kkk) ) /  ibeams_merged(kk,*,*,0) & endfor & endfor
 


ii = reform(ibeams_merged(0,*,*,0))


for kk = 0,n_slits-1 do begin  & temp_i = reform(ibeams_merged(kk,*,*,0))  & temp_q = reform(ibeams_merged(kk,*,*,1)) & temp_u = reform(ibeams_merged(kk,*,*,2)) & temp_v = reform(ibeams_merged(kk,*,*,3)) &  rr = abs(temp_i - median(temp_i,15)) &  ff = where(rr gt .25 * median(temp_i)) & mask = temp_i - temp_i & mask(ff) = 1. & for kkk =0,n_separations-1 do begin & i = temp_i(*,yyr(kkk):yyr(kkk+1))  & q = temp_q(*,yyr(kkk):yyr(kkk+1))  & u = temp_u(*,yyr(kkk):yyr(kkk+1)) & v = temp_v(*,yyr(kkk):yyr(kkk+1)) & mmask = mask(*,yyr(kkk):yyr(kkk+1)) & ff = where(mmask eq 0) & cal_curves(kkk,kk,k,0) = median(i(ff)) & cal_curves(kkk,kk,k,1) = median(q(ff)) & cal_curves(kkk,kk,k,2) = median(u(ff)) & cal_curves(kkk,kk,k,3) = median(v(ff)) & endfor & endfor


; do plotting
if n_elements(plt) ne 0 then begin

      cchars = 1.
      erase
      !p.multi = [4,2,2]
      plot,cal_curves(0,0,*,0),/nodata,charsize = cchars 

   for kkk = 0,n_separations-1 do begin & for kk = 0,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,0),color =40*(kk+kkk)+10,thick = 2 & endfor & endfor
  
      plot,cal_curves(0,0,*,1),/nodata,yr = [-1,1],/ystyle,charsize = cchars
     for kkk = 0,n_separations-1 do begin & for kk = 0,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,1),color =40*(kk+kkk)+10,thick = 2 & endfor & endfor

; for kk = 1,n_slits-1 do begin  oplot,cal_curves(kk,*,1),color =40*kk & endfor
      
      plot,cal_curves(0,0,*,2),/nodata,yr = [-1,1],/ystyle,charsize = cchars
                                ; for kk = 1,n_slits-1 do begin
                                ; oplot,cal_curves(kk,*,2),color
                                ; =40*kk & endfor
     for kkk = 0,n_separations-1 do begin & for kk = 0,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,2),color =40*(kk+kkk)+10,thick = 2 & endfor & endfor

      plot,cal_curves(0,0,*,3),/nodata,yr = [-1,1],/ystyle,charsize = cchars
     for kkk = 0,n_separations-1 do begin & for kk = 0,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,3),color =40*(kk+kkk)+10,thick = 2 & endfor & endfor
  
        wait,.1
   
endif

endfor

print,' Saving observed calibration measurements to ',outfile

if n_elements(add_cal) eq 0 then save,cal_curves,filename = outfile else  save,cal_curves,add_calvalues,filename = outfile

endif

pp = 0
if pp eq 1 then begin

cchars =.7
set_plot,'PS'
device,filename = '/export/cbeck/work/DKIST/telcal/FIRS_sequence.ps',xs = 17,ys = 14,/color,bits_per_pixel = 8

      !p.multi = [4,2,2]
      plot,cal_curves(0,0,*,0),/nodata,yr = [0,max(cal_curves(0,0,*,0))+500],/ystyle,title = 'Stokes I',charsize = cchars, ytitle = 'counts',xtitle = 'calibration position'

   for kkk = 0,n_separations-1 do begin & for kk = 0,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,0),color =(40*(kk+kkk)+10) mod 254 & endfor & endfor
  
      plot,cal_curves(0,0,*,1),/nodata,yr = [-1,1],/ystyle,title = 'Stokes Q',charsize = cchars, ytitle = 'Q/I',xtitle = 'calibration position'
     for kkk = 0,n_separations-1 do begin & for kk = 0,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,1),color =(40*(kk+kkk)+10) mod 254 & endfor & endfor

        plot,cal_curves(0,0,*,2),/nodata,yr = [-1,1],/ystyle,title = 'Stokes U',charsize = cchars, ytitle = 'U/I',xtitle = 'calibration position'
   for kkk = 0,n_separations-1 do begin & for kk = 0,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,2),color =(40*(kk+kkk)+10) mod 254 & endfor & endfor

      plot,cal_curves(0,0,*,3),/nodata,yr = [-1,1],/ystyle,title = 'Stokes V',charsize = cchars, ytitle = 'V/I',xtitle = 'calibration position'
     for kkk = 0,n_separations-1 do begin & for kk = 0,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,3),color =(40*(kk+kkk)+10) mod 254 & endfor & endfor
  


device,/close
set_plot,'X'

stop

endif

print,' Restoring observed calibration measurements from ',outfile

; restore saved polarization vectors
restore,outfile

for kkk = 0,n_separations-1 do begin & for kk = 0,n_slits-1 do begin & cal_curves(kkk,kk,*,0) = cal_curves(kkk,kk,*,0)/mean(cal_curves(kkk,kk,*,0)) & endfor & endfor

; (n_separations,n_slits,n,4)

; do plotting
if n_elements(plt) ne 0 then begin
   window,0,title ='Polarization degree in '+num2string(n_slits)+' slits'
   !p.multi = 0.
if n_slits gt 1 then begin
plot,sqrt( (cal_curves(0,0,*,1))^2.+ (cal_curves(0,0,*,2))^2.+ (cal_curves(0,0,*,3))^2.),/nodata,yr =[0,1.3],/ystyle
for k = 0,n_slits-1 do begin & oplot,sqrt( (cal_curves(0,k,*,1))^2.+ (cal_curves(0,k,*,2))^2.+ (cal_curves(0,k,*,3))^2.),color = k*40+10. & endfor
endif else begin
   plot,sqrt( (cal_curves(0,0,*,1))^2.+ (cal_curves(0,0,*,2))^2.+ (cal_curves(0,0,*,3))^2.)
endelse
   if n_elements(nopause) eq 0 then paused
 
endif


; create input polarization 

; Calculate Stokes input from calibration unit

; Non-ideal polarizer
; calculation of dichroism
px = 1.
py = 0.005
sq = sqrt(px*py)
a = ( px + py )/2.
c = ( px - py )/2.
polmat = setmtxright(a,c,0,0,c,a,0,0,0,0,0,0,0,0,0,0)/( (px + py )/2.)

; Retardance of cal. unit of the observation wavelength
; Non-ideal waveplate

; Offset angle between POL and RET (presently not used)

; retardance
retard = retard*!pi/180.

; (Non)-ideal retarder
b = 0.

; calculation of dichroism
px = 1 + b/2.
py = 1 - b/2.

sq = sqrt(px*py)
a = ( px + py )/2.
c = ( px - py )/2.
retmat = setmtxright(a,c,0,0,c,a,0,0,0,0,sq*cos(retard),sq*sin(retard),0,0,-sq*sin(retard),sq*cos(retard))

theta_err = theta_err*!pi/180.

stokesin = fltarr(n,4)
; calculate Stokes vector
for t =0,n-1 do begin

       ; thetaret = (-55.+t*5.)*!pi/180.

        thetaret = (0+t*10.)*!pi/180.

        thetapol = theta_pol*!pi/180.

; rotation of polarizer
        c = cos(2*thetapol)
        s = sin(2*thetapol)
        rotext1 = setmtxright(1,0,0,0,0,c,s,0,0,-s,c,0,0,0,0,1)
        rotext2 = setmtxright(1,0,0,0,0,c,-s,0,0,s,c,0,0,0,0,1)

; rotation of retarder
        c1 = cos(2*thetaret)
        s1 = sin(2*thetaret)
        rotint1 = setmtxright(1,0,0,0,0,c1,s1,0,0,-s1,c1,0,0,0,0,1)
        rotint2 = setmtxright(1,0,0,0,0,c1,-s1,0,0,s1,c1,0,0,0,0,1)

; rotation between both elements     
        c2 = cos(2*theta_err)
        s2 = sin(2*theta_err)
        rot_err = setmtxright(1,0,0,0,0,c2,s2,0,0,-s2,c2,0,0,0,0,1)

; calculate optical train
        calibmat = rotint2 ## retmat ## rotint1 ## rot_err ## rotext2 ## polmat ## rotext1

stokesvec = (calibmat ## [1,0,0,0])
stokesvec =  stokesvec/stokesvec(0)

stokesin(t,*) = stokesvec
endfor


; do plotting
if n_elements(plt) ne 0 and n_elements(no_av) ne 0 then begin

      window,0,title ='Example of observed calibration measurements'
      !p.multi = [4,2,2]
      plot,cal_curves( n_separations/2,0,*,0)
   for kkk = n_separations/2,n_separations/2 do begin &   for kk = 1,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,0),color =40*kk & endfor & endfor


      plot,cal_curves(n_separations/2,0,*,1)
   for kkk =  n_separations/2,n_separations/2 do begin & for kk = 1,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,1),color =40*kk & endfor & endfor
      
      plot,cal_curves( n_separations/2,0,*,2)
  for kkk =  n_separations/2,n_separations/2 do begin &  for kk = 1,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,2),color =40*kk & endfor & endfor

      plot,cal_curves( n_separations/2,0,*,3)
   for kkk =  n_separations/2,n_separations/2 do begin & for kk = 1,n_slits-1 do begin  oplot,cal_curves(kkk,kk,*,3),color =40*kk & endfor & endfor

  
if n_elements(nopause) eq 0 then paused

endif



print,'Inverting observations and Stokes input...'

; get response function X
xxmat = fltarr(n_separations,n_slits,4,4)

if n_elements(add_cal) ne 0 then cchi = fltarr(4)

for kkk =0,n_separations -1 do begin 

for k = 0,n_slits -1 do begin
   
   stokesout = reform(cal_curves(kkk,k,*,*))
   
; do matrix inversion between input and output
matrix_inversion,transpose(stokesin),transpose(stokesout),xmat,error


   xmat = xmat / xmat(0,0)

   xxmat(kkk,k,*,*) = xmat
   xinitial = xmat

; invert (forward) response
xinv = invert(xinitial,/double)

if n_elements(nverbose) eq 0 then begin

print
print,'Subdivision number ',kkk,' in slit ',k
print
print,'Xmat: '
print,xmat
xmat = transpose(xmat)
print
;rr = check_physicality(xmat)
;if rr(0) le 0 then print,' !! firs_630_cal: Warning !! Unphysical X matrix.'
;if rr(0) ge 0 then print,' firs_630_cal: X matrix seems to be ok...'
print
print,'Inverse'
print,xinv
print 
print,'Efficiencies I, Q, U, V, total QUV:'
print,effic(xmat),format='(5f8.3)'
print
xmat = transpose(xmat)
print,sqrt(total( xmat(*,1:3)^2.,2))
print
print,total( sqrt(total( xmat(*,1:3)^2.,2))^2.)
;stop

endif

erg = effic(xmat)
if n_elements(nverbose) eq 0 then print,' '

if max(erg(1:3)) gt 0.866 then begin
   print
   print,' !! firs_cal: Warning !! Efficiencies above maximal possible value of 0.866. Something could be wrong...' 
print
endif

; do plotting
if n_elements(plt) ne 0 then begin

totout = stokesout

ii1  = xinv(0,0)*stokesout(*,0)+xinv(1,0)*stokesout(*,1)+xinv(2,0)*stokesout(*,2)+xinv(3,0)*stokesout(*,3)
ii2  = xinv(0,1)*stokesout(*,0)+xinv(1,1)*stokesout(*,1)+xinv(2,1)*stokesout(*,2)+xinv(3,1)*stokesout(*,3)
ii3  = xinv(0,2)*stokesout(*,0)+xinv(1,2)*stokesout(*,1)+xinv(2,2)*stokesout(*,2)+xinv(3,2)*stokesout(*,3)
ii4  = xinv(0,3)*stokesout(*,0)+xinv(1,3)*stokesout(*,1)+xinv(2,3)*stokesout(*,2)+xinv(3,3)*stokesout(*,3)

i1 = xinitial(0,0)*stokesin(*,0)+xinitial(1,0)*stokesin(*,1)+xinitial(2,0)*stokesin(*,2)+xinitial(3,0)*stokesin(*,3)
i2 = xinitial(0,1)*stokesin(*,0)+xinitial(1,1)*stokesin(*,1)+xinitial(2,1)*stokesin(*,2)+xinitial(3,1)*stokesin(*,3)
i3 = xinitial(0,2)*stokesin(*,0)+xinitial(1,2)*stokesin(*,1)+xinitial(2,2)*stokesin(*,2)+xinitial(3,2)*stokesin(*,3)
i4 = xinitial(0,3)*stokesin(*,0)+xinitial(1,3)*stokesin(*,1)+xinitial(2,3)*stokesin(*,2)+xinitial(3,3)*stokesin(*,3)



window,0,title = 'Observed cal. measurements and X * input, slit '+num2string(k+1)

cchars= 1.

!p.multi = [4,2,2]
plot,stokesout(*,0),/yno,charsize = cchars
oplot,i1,psym = -1,color = 240
plot,stokesout(*,1),/yno,charsize = cchars
oplot,i2,psym = -1,color = 240
plot,stokesout(*,2),/yno,charsize = cchars
oplot,i3,psym = -1,color = 240
plot,stokesout(*,3),/yno,charsize = cchars
oplot,i4,psym = -1,color = 240

legend1,text =[' Observed calibration curves',' X * Stokes input'],color =[ 256^3.-1,240],ttextcolor = [ 256^3.-1,240],pos = 1,linestyle = fltarr(2),charsize = 1.

window,1,title = 'Created input and X!E-1!N * observed cal.measurements, slit '+num2string(k+1)


;set_plot,'PS'
;device,filename = '/export/cbeck/work/DKIST/pol_workshop_may15/offset_angle_firs_3.ps',xs = 16,ys = 16,bits_per_pixel = 8,/color

!p.multi = [4,2,2]
plot,stokesin(*,0),/yno,yrange = [.9,1.1],/ystyle,title = 'Stokes I',charsize = cchars
oplot,ii1,psym = -1,color = 240
plot,stokesin(*,1),/yno,title = 'Stokes Q',charsize = cchars
oplot,ii2,psym = -1,color = 240
plot,stokesin(*,2),/yno,title = 'Stokes U',charsize = cchars
oplot,ii3,psym = -1,color = 240
plot,stokesin(*,3),/yno,title = 'Stokes V',charsize = cchars
oplot,ii4,psym = -1,color = 240

;legend1,text =[' Stokes input',' X!E-1!N * observed calibration curves'],color =[0,240],ttextcolor = [0,240],pos = [.1,.88],linestyle = fltarr(2),charsize = 1.,/normal,/no_box

legend1,text =[' Stokes input',' X!E-1!N * observed calibration curves'],color =[256^3.-1,240],ttextcolor = [256^3.-1,240],pos = 1,linestyle = fltarr(2),charsize = 1.

;device,/close
;set_plot,'X'
;stop

if n_elements(nopause) eq 0 then paused


endif


endfor

endfor

if n_elements(nverbose) eq 0 then print,'Saving to: ',outfile1

; save results
if n_elements(add_cal) eq 0 then save,theta_err,xxmat,retard,yyr,n_separations,filename = outfile1 else save,theta_err,xxmat,retard,cchi,add_stokesout,yyr,n_separations,filename = outfile1

if n_elements(add_cal) ne 0 then begin
   print
   print,total(cchi/4.),retard/!pi*180.
endif

end

