pro get_firs_dataprop, input_dir,change_save = change_save,help= help

;+
;============================================================================
;
;	procedure : get_firs_dataprop
;
;	purpose   : get settings of observations on one day
;
;	author    : cbeck@nso Apr 2013
;
;==============================================================================

if n_elements(help) ne 0 then begin
	print
	print, "usage:   get_firs_dataprop, input_dir,change_save = change_save,help= help"
	print
	print, "	 Get settings of all FIRS observations in one directory."
        print, "         "
	print
	print, "	Arguments"
        print, "                input_dir"
        print
        print, "         Keywords :"
        print, "              change_save: directory where the file information"
        print, "                        on the source directory will be saved to."
        print, "                          To be used if you do not have write"
        print, "                          permission in source directory."
        print, "               /help   : switch to active the help text above."
        print
return
endif
;==============================================================================
;-

; input directory, set to current if not provided
if n_elements(input_dir) eq 0 then begin
cd,current = input_dir
input_dir = input_dir+'/'
endif

; read in all FIRS files
erg = read_directory(input_dir,filter ="firs.*")
files = erg.files(where(erg.types eq 0))

ww = erg.types
if ww(0) eq '' then begin
print,'!get_firs_dataprop: No FIRS data files found in: ',input_dir
print,'Aborting.'
return
stop
endif

; storage arrays for times etc.
n = n_elements(files)

times  = fltarr(n,4)
types  = fltarr(n)
date   = fltarr(n,3)
expos    = fltarr(n)
acc    = fltarr(n)
llambda = fltarr(n)
n_steps= fltarr(n)
timestart = fltarr(n,4)
timeend = fltarr(n,4)
camera = fltarr(n)
bbasefile = strarr(n)

ccomment = strarr(n)

; loop over files
for k = 0,n-1 do begin
   
   erg = string2num(files(k),divider = '.')
   year = string2num(strmid(files(k),7, 4))
   month= string2num(strmid(files(k),11, 2))
   day  = string2num(strmid(files(k),13, 2))
   hour = string2num(strmid(files(k),16, 2))
   min = string2num(strmid(files(k),18, 2))
   sec = string2num(strmid(files(k),20, 2))

   date(k,0) = year
   date(k,1) = month
   date(k,2) = day

   times(k,0) = hour
   times(k,1) = min
   times(k,2) = sec
   times(k,3) = hour+min/60.+sec/3600.

; find base file names, extensions are unimportant
   base_file = strmid(files(k),0, 23)
   bbasefile(k) = base_file
   erg = read_directory(input_dir,filter =base_file+'*')
   camera(k) =  strmid(files(k),5, 1)

; check how many files belong to the observation
   n_steps(k) = n_elements(where(erg.types eq 0))

; get start time and header info
   ffxread,input_dir+files(k),dat,hdr,/NODATA    
   expos(k) = FXPAR(hdr,'EXP_TIME')
   acc(k)   = FXPAR(hdr,'SUMS')
   llambda(k) = string2num(FXPAR(hdr,'WAVELENG'),divider ='In')
   jump =  n_elements(where(erg.types eq 0))
   time = string2num(FXPAR(hdr,'OBS_STAR',COUNT = obs_star_exist),divider ='T:.')
   aa =  FXPAR(hdr,'COMMENT')
   bb = ''
   for ll = 0,n_elements(aa)-1 do begin & bb = bb +' '+aa(ll) & endfor
   ccomment(k) = bb

   if n_elements(time) lt 2 then    time = string2num(FXPAR(hdr,'DST_TIME',COUNT = obs_star_exist),divider ='T:.')

   timestart(k,0:2) = time(1:3)
   timestart(k,3) = time(1)+time(2)/60.+time(3)/3600.

; get end time
   ffxread,input_dir+files(k+jump-1),dat,hdr,/NODATA    
   time = string2num(FXPAR(hdr,'OBS_STAR',COUNT = obs_star_exist),divider ='T:.')

   if n_elements(time) lt 2 then    time = string2num(FXPAR(hdr,'DST_TIME',COUNT = obs_star_exist),divider ='T:.')

   timeend(k,0:2) = time(1:3)
   timeend(k,3) = time(1)+time(2)/60.+time(3)/3600.

; jump to next base file
   k = k + jump -1 

endfor

; sort timing etc. info to base files only
ff = where(reform(date(*,0)) ne 0)

n1 = n_elements(ff)

times1  = fltarr(n1,4)
timestart1  = fltarr(n1,4)
timeend1  = fltarr(n1,4)
date1   = fltarr(n1,3)

times1(*,0) = times(ff,0)
times1(*,1) = times(ff,1)
times1(*,2) = times(ff,2)
times1(*,3) = times(ff,3)
timestart1(*,0) = timestart(ff,0)
timestart1(*,1) = timestart(ff,1)
timestart1(*,2) = timestart(ff,2)
timestart1(*,3) = timestart(ff,3)
timeend1(*,0) = timeend(ff,0)
timeend1(*,1) = timeend(ff,1)
timeend1(*,2) = timeend(ff,2)
timeend1(*,3) = timeend(ff,3)

date1(*,0) = date(ff,0)
date1(*,1) = date(ff,1)
date1(*,2) = date(ff,2)
types1      = types(ff)
exp1        = expos(ff)
acc1        = acc(ff)
lambda1     = llambda(ff)
n_steps1    = n_steps(ff)
camera1     = camera(ff)
bbasefile1 = bbasefile(ff)
ccomment = ccomment(ff)

; output text file
outfile = input_dir+num2string(fix(date1(0,0)))+num2string(fix(date1(0,1)))+num2string(fix(date1(0,2)))+'.'+num2string(fix(camera1(0)))+'.txt'

if n_elements(change_save) ne 0 then outfile = change_save+num2string(fix(date1(0,0)))+num2string(fix(date1(0,1)))+num2string(fix(date1(0,2)))+'.'+num2string(fix(camera1(0)))+'.txt'

; check write permission
erg = check_existence(outfile,/write,/change)
outfile = erg(0)


print
print,' Text file with file information in: ',outfile

; open output text file
openw,outunit,outfile,/get_lun

printf,outunit,'No.  File base               date      start    - end time ? exp.  naccum steps'

for k = 0,n1-1 do begin
iindex = k

printf,outunit,num2string(fix(iindex))+' Op '+bbasefile1(k)+' '+num2string(fix(date1(k,0)))+'/'+num2string(fix(date1(k,1)))+'/'+num2string(fix(date1(k,2)))+' '+num2string(fix(timestart1(k,0)))+':'+num2string(fix(timestart1(k,1)))+':'+num2string(fix(timestart1(k,2)))+' - '+num2string(fix(timeend1(k,0)))+':'+num2string(fix(timeend1(k,1)))+':'+num2string(fix(timeend1(k,2)))+' '+num2string(fix(types1(k)))+' '+num2string(exp1(k))+' '+num2string(fix(acc1(k)))+' '+num2string(fix(n_steps1(k)))+' '+ccomment(k)
endfor
; close output text file
free_lun,outunit

; idl save file with other parameters
outfile = input_dir+num2string(fix(date1(0,0)))+num2string(fix(date1(0,1)))+num2string(fix(date1(0,2)))+'.'+num2string(fix(camera1(0)))+'.sav'

if n_elements(change_save) ne 0 then outfile = change_save+num2string(fix(date1(0,0)))+num2string(fix(date1(0,1)))+num2string(fix(date1(0,2)))+'.'+num2string(fix(camera1(0)))+'.sav'

; check write permission
erg = check_existence(outfile,/write,/change)
outfile = erg(0)

print
print,' Saving file information to: ',outfile

; save results
save,times1 ,timestart1,timeend1,date1,types1 ,exp1,acc1,lambda1,n_steps1,camera1,bbasefile1,filename =outfile


end
