
PRO MAIN_FIRS_REDUCE,config_str,CALS_ONLY=CALS_ONLY

  COMMON PORT4_BLOCK, mm_inv_port4

  ;;-----------------------------------------------------------------
  ;; SET UP IMAGE AND DEVICE SYSTEM VARIABLES FOR PLOT_IMAGE CALLS
  ;;-----------------------------------------------------------------
  
  DEVICELIB
  IMAGELIB

  output_dir = config_str.output_dir
  scan_date  = config_str.scan_date
  dark_name  = config_str.dark_name
  lambda     = config_str.lambda
  scan_dir   = config_str.scan_dir
  scan_name  = config_str.scan_name
  flat_dir   = config_str.flat_dir
  flat_name  = config_str.flat_name
  grid_file  = config_str.grid_file 
  lin_file   = config_str.lin_file
  pcal_name  = config_str.pcal_name
  pcal_Lang  = config_str.pcal_Lang
  polc_dir   = config_str.polc_dir
  slit_width = config_str.slit_width
  target     = config_str.target  

  julian     = SYSTIME(/JULIAN)
  CALDAT, julian, Month, Day, Year, Hour, Minute, Second      
  start_time = STRING(hour,FORMAT = '(I02)') + $
    STRING(minute,FORMAT = '(I02)') + $
    STRING(FIX(second),FORMAT = '(I02)') 
  journal_filename = output_dir + scan_date + '.' + scan_name + '.'+start_time+'.logfile'
  IF (config_str.journal_tag EQ 1) THEN BEGIN
      PRINT,' '
      PRINT,' THE OUTPUT OF THIS REDUCTION PROCESS WILL'
      PRINT,' BE LOGGED IN THIS FILE: ',journal_filename
      JOURNAL,journal_filename
      PRINT,' '
      PRINT,' Type .c to continue' 
      STOP
  ENDIF

  
  PRINT,'       ****************************************************'
  PRINT,'                        Scan Reduction Inputs              '
  PRINT,'       ****************************************************'
  PRINT,'       SCAN DATE:  ',config_str.scan_date
  PRINT,'       SCAN NAME:  ',config_str.scan_name
  PRINT,'       FLAT NAME:  ',config_str.flat_name
  PRINT,'       DARK NAME:  ',config_str.dark_name
  PRINT,'       PCAL NAME:  ',config_str.pcal_name
  PRINT,'       TARGET:     ',config_str.target
  PRINT,'       LAMBDA:     ',config_str.lambda
  PRINT,'       SLIT WIDTH: ',config_str.slit_width
  PRINT,'       SCAN DIR:   ',config_str.scan_dir
  PRINT,'       DARK DIR:   ',config_str.dark_dir
  PRINT,'       FLAT DIR:   ',config_str.flat_dir
  PRINT,'       POLC DIR:   ',config_str.polc_dir
  PRINT,'       OUTPUT DIR: ',config_str.output_dir
  PRINT,'       ****************************************************'

  ;;**********************************************************************
  ;;          REDUCTION PROCEDURE FOR FIRS IR HELIUM I 1083 SCANS
  ;;
  ;;  This reduction method can be used for scanned data acquired with the
  ;;  multi-slit Facility Infrared Spectropolarimeter (FIRS) located at
  ;;  Port 4 of the National Solar Observatory's Dunn Solar Telescope
  ;;  (DST) located on Sacramento Peak, New Mexico, USA.
  ;;
  ;;  VERSION HISTORY:
  ;;    2010        - Originally written by Sarah Jaeggli, Univ of Hawaii
  ;;    2011 FEB 22 - Reduction method adapted by Tom Schad
  ;;    2012 OCT 16 - Various Updates Including:
  ;;                --> use config structure instead of common block
  ;;                --> reduce beams entirely independently and save the results for each
  ;;                beam separately
  ;;                --> write the reduced data differently so the
  ;;                program can run without overloading memory use
  ;;**********************************************************************
   
  ;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  ;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  ;;          GENERATION OF DATA CALIBRATION FILES
  ;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  ;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  PRINT,'----------------------------------------------------------'
  PRINT,'----------------------------------------------------------'
  PRINT,' ACQUIRING THE DARK CURRENT CALIBRATION'
  PRINT,'----------------------------------------------------------'
  PRINT,'----------------------------------------------------------'

  FIRS_GET_DARK,config_str

  PRINT,'----------------------------------------------------------'
  PRINT,'----------------------------------------------------------'
  PRINT,' NOW FIND PARAMETERS DETERMINING APPROXIMATE SLIT' 
  PRINT,' LOCATIONS, SKEW ROTATION VALUE, ETC'
  PRINT,'----------------------------------------------------------'
  PRINT,'----------------------------------------------------------'

  FIRS_GET_SLIT_SKEW,config_str

  PRINT,'----------------------------------------------------------'
  PRINT,'----------------------------------------------------------'
  PRINT,' NOW GET AVERAGE FLAT FIELD FRAMES AND ACQUIRE SLIT CURVATURE '
  PRINT,' ' 
  PRINT,' NOTE THAT FLATS WILL NOT BE DEMODULATED AT THIS POINT    '
  PRINT,' '
  PRINT,' ALSO NOTE THAT THE SLIT CURVATURE CORRECTION WILL   '
  PRINT,' NOT BE IMPLEMENTED ON THE FLATS YET !!                   '
  PRINT,' ' 
  PRINT,' THREE DIFFERENT FLATS WILL BE GIVEN :                    '
  PRINT,' 1. FLAT         --> FLAT FRAMES AVERAGED                 '
  PRINT,' 2. FLAT_SKEWCOR --> FLAT FRAMES, SKEW CORRECTED          '
  PRINT,' 2. FLATRM       --> FLAT FRAMES, SKEW CORRECTED, LINES REMOVED '
  PRINT,'----------------------------------------------------------'
  PRINT,'----------------------------------------------------------'

  PRINT,' TYPE .c TO CONTINUE'
  
  nflats = N_ELEMENTS(config_str.flat_name)
  PRINT,' '
  PRINT,' NUMBER OF FLATS SPECIFIED: ' + STRING(nflats,FORMAT = '(I2)')
  FOR n = 0,nflats-1 DO BEGIN
      PRINT,' GETTING FLAT NUMBER ', STRTRIM(n+1,2), ' OF ', STRTRIM(nflats,2)
      FIRS_GET_FLAT,config_str,n
  ENDFOR

  ;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  ;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  ;;        READING, CALIBRATING, AND COMPILING SCANNED DATA
  ;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  ;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  IF NOT KEYWORD_SET(CALS_ONLY) THEN BEGIN

      PRINT,'----------------------------------------------------------'
      PRINT,' CALCULATING THE INSTRUMENTAL POLARIMETRIC RESPONSE       '
      PRINT,'----------------------------------------------------------'
      
      FIRS_POLCAL_MULTI,config_str
      
      ;;================================================
      ;; RESTORE DARK
      ;;================================================

      dark_savfile = output_dir + scan_date+'.'+dark_name+'.darkcal'
      RESTORE,dark_savfile
      
      ;;================================================
      ;; RESTORE LINEARITY FILE IF IT EXISTS
      ;;================================================

      IF (FILE_TEST(lin_file)) THEN BEGIN
          
          PRINT,'----------------------------------------------'
          PRINT,' Linearity File exists !! '
          PRINT,' Filename:  ',lin_file
          PRINT,' Restoring Linearity File'
          PRINT,'----------------------------------------------'
          
          RESTORE, lin_file
          
      ENDIF ELSE PRINT,'---> NO LINEARITY FILE FOUND!!!'
    
      ;;================================================
      ;; GET NUMBER OF SLITS FROM SLIT_SKEW CAL FILE
      ;; TO CREATE STRING NAMES FOR REDUCED SLIT FILES
      ;; ---------> ALSO GET FIRS_SKEW ANGLE <----------
      ;;================================================

      skew_savfile = output_dir + scan_date + '.skewcal'
      RESTORE,skew_savfile
      PRINT,' '
      PRINT,' Number of slits: ', num_slits
      nos = STRTRIM(STRING(INDGEN(num_slits)+1),2)

      ;;==============================================================
      ;; RESTORE POLCAL FILE, INVERT, AND RESIZE, LOAD IN COMMON BLOCK
      ;;==============================================================

      pcal_savfile = output_dir + scan_date + '.' + pcal_name[0] + '.polcal'

      RESTORE, pcal_savfile,/VERB

      IF (lambda EQ '630') THEN BEGIN
          NX = 2048.
          NY = 2048.
      ENDIF ELSE BEGIN
          NX = 1024.
          NY = 1024.
      ENDELSE

      mm_arr_inv  = mm_arr
      FOR i = 0,(NX/4.)-1 DO FOR j = 0,(NY/4.)-1 DO MM_ARR_INV[i,j,*,*] = INVERT(REFORM(MM_ARR[i,j,*,*]))
      mm_inv = FLTARR(NX,NY,4,4)
      FOR i = 0,3 DO FOR j = 0,3 DO mm_inv[*,*,i,j] = REBIN(REFORM(mm_arr_inv[*,*,i,j]),NX,NY)
        
      mm_inv_port4 = mm_inv
  
      ;;=============================================================
      ;; FIND THE FLAT FILES AND READ ONE FILE FOR HEADER INFORMATION
      ;;=============================================================
 
      fskew_demod = FLTARR(NX,NY,4,nflats)
      frm_demod   = FLTARR(NX,NY,4,nflats)
      flat_time   = STRARR(nflats)
      flat_names_sav = flat_name

      FOR n = 0,nflats-1 DO BEGIN

          PRINT,' RESTORING FLAT NUMBER ',STRTRIM(n+1,2),' OF ', STRTRIM(nflats,2)

          flat_name = flat_names_sav[n]

          flat_savfile = output_dir + scan_date+'.'+flat_name+'.flatcal'
          RESTORE,flat_savfile

          IF (lambda EQ '630') THEN camera_no = FIX(1.) ELSE camera_no = FIX(2.)
          flat_file = flat_dir+'firs.'+strtrim(camera_no,2)+'.*'+flat_name+'*'
          FILEN = FILE_SEARCH(flat_file, COUNT = nflats_tmp)
          FILEN = FIRS_FILE_ORDER(FILEN)
          FXREAD,filen[nflats_tmp/2.],dat,hdr_flat,/NODATA

          ;;================================================
          ;; DEMODULATE THE FLAT
          ;;================================================

          PRINT,' DEMODULATING FLAT NUMBER ',STRTRIM(n+1,2),' OF ', STRTRIM(nflats,2)

          fskew_demod_n = FIRS_XT_DEMOD(flat_skewcor,hdr_flat,FLOAT(lambda),config_str.t_file,/SOLAR)
          frm_demod_n   = FIRS_XT_DEMOD(flatrm,hdr_flat,FLOAT(lambda),config_str.t_file,/SOLAR)      

          PRINT,' --- Demodulation AND XT correction complete --'

          ;;================================================
          ;; NORMALIZE THE FLAT FIELDS TO MEDIAN VALUE IN
          ;; STOKES I -- DO THIS SLIT BY SLIT (no?)
          ;;================================================
	
          mask_lbl = LABEL_REGION(mask)
          mask_norm = mask
          frm_sI = frm_demod_n[*,*,0]
          mask_norm[*,*] = MEDIAN(frm_sI[WHERE(mask eq 1.)])
      
          FOR state = 0,3 DO fskew_demod_n[*,*,state] = fskew_demod_n[*,*,state]/mask_norm
          FOR state = 0,3 DO frm_demod_n[*,*,state] = frm_demod_n[*,*,state]/mask_norm 

          ;;================================================
          ;; FIX FLAT FOR I->Q,U,V CROSSTALK
          ;;================================================

          fskew_demod_n = FIRS_I2X_REMOVE(fskew_demod_n,mask)
          frm_demod_n   = FIRS_I2X_REMOVE(frm_demod_n,mask)

          ;;================================================
          ;; GET TIME FOR THE FLAT MIDDLE TIME
          ;; WARNING:  DIFFERENT TIME TAGS EXIST FOR DATA 
          ;;           PRE/POST JAN 2012. 
          ;;================================================
          
          ftime = FXPAR(hdr_flat,'TIME_OBS',COUNT = time_obs_exist)
          ftime = FXPAR(hdr_flat,'OBS_STAR',COUNT = obs_star_exist)
          IF (time_obs_exist EQ 1) THEN $
            ftime = STRCOMPRESS(STRING(FXPAR(hdr_flat,'DATE_OBS'))) + ' ' + $
            STRCOMPRESS(STRING(FXPAR(hdr_flat,'TIME_OBS')))
          IF (obs_star_exist EQ 1) THEN ftime = STRCOMPRESS(STRING(FXPAR(hdr_flat,'OBS_STAR')))
          ftime = ANYTIM2UTC(ftime,/CCSDS)
          ftime = STRMID(ftime,11,8)
          
          flat_time[n]         = ftime
          fskew_demod[*,*,*,n] = fskew_demod_n
          frm_demod[*,*,*,n]   = frm_demod_n

      ENDFOR

      ;;================================================
      ;; FIND AND ENSURE PROPER ORDER OF THE SCAN FILES
      ;;================================================

      IF (lambda EQ '630') THEN camera_no = FIX(1.) ELSE camera_no = FIX(2.)

      scan_file = scan_dir + 'firs.' + strtrim(camera_no,2) + '.' + scan_date + '.' + scan_name+'*'
  
      filen = FILE_SEARCH(scan_file, COUNT = no_scans)
      filen = FIRS_FILE_ORDER(FILEN,loop_tags = loop_tags, repeat_tags = repeat_tags)

      PRINT,' '
      PRINT,' Scan Date.............................',STRTRIM(scan_date,2)
      PRINT,' Scan Name.............................',STRTRIM(scan_name,2)
      PRINT,' Total Number of Scan Files Found......',STRTRIM(no_scans,2)
      PRINT,' Number of Repeats Found...............',STRTRIM(FIX(MAX(repeat_tags) + 1),2)
      PRINT,' '
      
      FOR nr = 0,MAX(repeat_tags) DO PRINT,' LOOPS IN REPEAT NUMBER '+ STRTRIM(FIX(nr+1),2) + $
        '..............' + STRTRIM(N_ELEMENTS(where(repeat_tags EQ nr)),2)

      nos_repeat = STRTRIM(STRING(INDGEN(MAX(repeat_tags+1))),2)

      ;;================================================
      ;; Initiate the HyperStral Stokes Data Structure
      ;; This is the routine where the slit window is defined
      ;;================================================

      PRINT,'MEMORY USE BEFORE DATA ARRAY ALLOCATION : ',(MEMORY(/CURRENT)*9.53674e-7)/1000.,' GIGABYTES'

      hssd2_bot = FIRS_HSSD_INIT(config_str)
      hssd2_top = FIRS_HSSD_INIT(config_str)   

      PRINT,'MEMORY USE AFTER: ',(MEMORY(/CURRENT)*9.53674e-7)/1000.,' GIGABYTES'
      
      ;;================================================
      ;; Data Reduction Loop - All Repeats, Slits and Steps
      ;;================================================

      FOR rn = 0,MAX(repeat_tags) DO BEGIN 

          ;; get the files for the individual scan whether repeated or not
          filen_repeat = filen[WHERE(repeat_tags EQ rn, no_loops)]

          FOR k = 0,num_slits-1 DO BEGIN  
              
              FOR i = 0,no_loops-1 DO BEGIN
                  
                  systime0 = systime(/seconds)
                  
                  PRINT,'Processing scan step ', STRTRIM(i,2), ' in slit ', STRTRIM(k+1), ' and repeat: ',STRTRIM(rn+1,2)
                  
                  ;;=================================================
                  ;; READ FILE, CORRECT FOR DARK CURRENT AND LINEARITY
                  ;; FIX IR CAMERA BAD PIXELS, IF NECESSARY
                  ;;=================================================

                  systime00 = SYSTIME(/SECONDS)
                  dat = FLOAT(READFITS(filen_repeat[i],hdr)) - FLOAT(dark)
                  IF (N_ELEMENTS(photons) GT 0.) THEN dat = PHOTONS(dat) ELSE dat = dat
                  ;; for intensity only data
                  IF (N_ELEMENTS(SIZE(dat,/DIMENSIONS)) EQ 2.) then dat = [[[dat]],[[dat]],[[dat]],[[dat]]]
                  IF (camera_no EQ 2.) THEN FOR p = 0,3 do dat[0,0,p] = FIX_VIRGO1K(dat(*,*,p), /SCI)
                  PRINT,'    load time: ',SYSTIME(/SECONDS) - systime00
                  
                  ;;=================================================
                  ;; DEMODULATE DATA FILE FOR ENTIRE POLARIMETRIC SYSTEM
                  ;;=================================================
     
                  systime00 = SYSTIME(/SECONDS)
                  dat_demod = FIRS_XT_DEMOD(dat,hdr,FLOAT(lambda),config_str.t_file,/SOLAR)
                  PRINT,'   demod time: ',SYSTIME(/SECONDS) - systime00
                  
                  ;;=================================================
                  ;; CORRECT DATA FOR SKEW ANGLE
                  ;;=================================================

                  systime00 = systime(/seconds)
                  FOR state = 0,3 DO dat_demod[0,0,state] = $
                    ROT(dat_demod[*,*,state],firs_skew,1,0.,0.,/PIVOT,/INTERP,CUBIC = -1,MISSING = 0.)
                  PRINT,'    skew time: ',systime(/seconds) - systime00
                  
                  ;;=================================================
                  ;; CORRECT DATA FOR I->Q,U,V CROSSTALK
                  ;;=================================================

                  dat_demod = FIRS_I2X_REMOVE(dat_demod,mask)
              
                  ;;=================================================
                  ;; FLAT FIELD CORRECTION OF DATA
                  ;;=================================================
                  
                  ;; first make interpolated flat if more than one flat exists
                  ;; if there is only one flat,
                  ;; firs_flat_interpol will return it  
                  
                  scan_obs_time = FXPAR(hdr,'TIME_OBS',COUNT = time_obs_exist)
                  scan_obs_time = FXPAR(hdr,'OBS_STAR',COUNT = obs_star_exist)
                  IF (time_obs_exist EQ 1) THEN $
                    scan_obs_time = STRCOMPRESS(STRING(FXPAR(hdr,'DATE_OBS'))) + ' ' + $
                    STRCOMPRESS(STRING(FXPAR(hdr,'TIME_OBS')))
                  IF (obs_star_exist EQ 1) THEN BEGIN
                                ;; get both start and end time 
                                ;; take middle of these times
                      obs_time1 = STRCOMPRESS(STRING(FXPAR(hdr,'OBS_STAR')))
                      obs_time2 = STRCOMPRESS(STRING(FXPAR(hdr,'OBS_END')))
                      time_diff = (ANYTIM2TAI(obs_time2) - ANYTIM2TAI(obs_time1))
                      scan_obs_time = ANYTIM2UTC(ANYTIM2TAI(obs_time1) + time_diff/2.,/CCSDS)
                  ENDIF

                  scan_obs_time = STRMID(scan_obs_time,11,8)
                     
                  fskew_demod_int = FIRS_FLAT_INTERPOL(fskew_demod,flat_time,scan_obs_time)
                  frm_demod_int   = FIRS_FLAT_INTERPOL(frm_demod,flat_time,scan_obs_time)

                                ;; PROBABLY SHOULD GET RID OF FRINGES
                                ;; IN FLATS EARLIER, PRIOR TO
                                ;; INTERPOLATION AND BEFORE MAIN LOOP
                                    
                                ;; PUT FRINGE REMOVAL METHOD IN
                                ;; HERE..APPLY TO DATA..USE MODEL OF
                                ;; FRINGES DEFINE BY FLATS?

                  systime00 = systime(/seconds)
                  IF (config_str.flat_subtract EQ 1) THEN BEGIN                       
                      ;; first subtract off flat field relative polarization fringes q,u,v
                      FOR state = 1,3 DO dat_demod[0,0,state] = $
                        (dat_demod[*,*,state]/dat_demod[*,*,0]) - (fskew_demod_int[*,*,state]/fskew_demod_int[*,*,0])
                      ;; multiply again by stks i to acquire Q,U,V (unnormalized)
                      FOR state = 1,3 DO dat_demod[0,0,state] = dat_demod[*,*,state]*dat_demod[*,*,0]
                  ENDIF
                  ;; divide by stks i flat (remember to preserve relative polarization!!!)
                  FOR state = 0,3 DO dat_demod[0,0,state] = dat_demod[*,*,state]/(frm_demod_int[*,*,0]>0<2.)
                  ;; multiply by mask to get rid of bad values
                  FOR state = 0,3 DO dat_demod[0,0,state]=dat_demod[*,*,state]*mask
                  PRINT,'    flat time: ',systime(/seconds) - systime00
                  
                  ;;=================================================
                  ;; CUT OUT DATA SLIT SHIFTED FOR SLIT CURVATURE
                  ;;=================================================
                  
                  systime00 = systime(/seconds)
                  dat_slit = FIRS_DATA_GET_SLIT(dat_demod, k, hdr,config_str)
                  PRINT,'slit get time: ',systime(/seconds) - systime00
                  
                  ;;=================================================
                  ;; SPLIT UP INTO SEPARATE VALUES TO FEED UPDATE
                  ;;=================================================
                  
                  systime00 = SYSTIME(/SECONDS)
                  
                  I_bot = dat_slit[*,*,0,0]
                  Q_bot = dat_slit[*,*,1,0]
                  U_bot = dat_slit[*,*,2,0]
                  V_bot = dat_slit[*,*,3,0]
                  
                  I_top = dat_slit[*,*,0,1]
                  Q_top = dat_slit[*,*,1,1]
                  U_top = dat_slit[*,*,2,1]
                  V_top = dat_slit[*,*,3,1]
                  
                  ;;==================================================
                  ;; DISPLAY SOME DATA - CAVEAT: BEFORE COMBINATION
                  ;;==================================================
                 
                  bot_disp = [ [BYTSCL((V_bot/I_bot)-MEDIAN(V_bot/I_bot),-0.05,0.05)],$
                               [BYTSCL((U_bot/I_bot)-MEDIAN(U_bot/I_bot),-0.05,0.05)],$
                               [BYTSCL((Q_bot/I_bot)-MEDIAN(Q_bot/I_bot),-0.05,0.05)],$
                               [BYTSCL(I_bot, 0,MEDIAN(I_bot)+0.2*MEDIAN(I_bot))]]
                  
                  top_disp = [ [BYTSCL((V_top/I_top)-MEDIAN(V_top/I_top),-0.05,0.05)],$
                               [BYTSCL((U_top/I_top)-MEDIAN(U_top/I_top),-0.05,0.05)],$
                               [BYTSCL((Q_top/I_top)-MEDIAN(Q_top/I_top),-0.05,0.05)],$
                               [BYTSCL(I_top, 0,MEDIAN(I_top)+0.2*MEDIAN(I_top))]]
             
                  sz_disp = SIZE(bot_disp,/DIMENSIONS)
                  sz_screen = MIN(GET_SCREEN_SIZE())
                  disp_ratio = (0.5*sz_screen)/MAX(sz_disp)
                  
                  IF (i EQ 0) OR  (!d.window EQ -1) THEN BEGIN
                      WINDOW,0,XS = sz_disp[0]*disp_ratio, YS = sz_disp[1]*disp_ratio, TITLE = ' BOTTOM BEAM DATA '
                      WINDOW,2,XS = sz_disp[0]*disp_ratio, YS = sz_disp[1]*disp_ratio, TITLE = ' TOP BEAM DATA '
                  ENDIF
                  
                  WSET,0
                  TVSCL,FREBIN(bot_disp,sz_disp[0]*disp_ratio,sz_disp[1]*disp_ratio)
                  WSET,2
                  TVSCL,FREBIN(top_disp,sz_disp[0]*disp_ratio,sz_disp[1]*disp_ratio)
                  
                  print,' display time: ',systime(/seconds) - systime00
                  
                  ;;==================================================
                  ;; SAVE THE BEAMS INTO TWO DIFFERENT HSSD STRUCTURE
                  ;;==================================================
                  
                  hssd2_bot = FIRS_HSSD2_UPDATE(hssd2_bot,hdr,i,I_bot,Q_bot,U_bot,V_bot)
                  hssd2_top = FIRS_HSSD2_UPDATE(hssd2_top,hdr,i,I_top,Q_top,U_top,V_top)
                  
                  PRINT,'   total time:',systime(/seconds)-systime0

              ENDFOR

              PRINT,' SAVING EACH BEAM TO SAVE FILES '
      
              PRINT,'----------------------------------------------------------'
              PRINT,' '
              hssd_sav_name = output_dir+scan_date+'.'+scan_name+'.'+nos(k)+'.r'+nos_repeat(rn)+'.bot.hssd'
              PRINT,'Saving processed data in ' + hssd_sav_name+ ', please wait...'
              SAVE, hssd2_bot, FILENAME = hssd_sav_name
              PRINT,' '
              PRINT,' '
              hssd_sav_name = output_dir+scan_date+'.'+scan_name+'.'+nos(k)+'.r'+nos_repeat(rn)+'.top.hssd'
              PRINT,'Saving processed data in ' + hssd_sav_name+ ', please wait...'
              SAVE, hssd2_top, FILENAME = hssd_sav_name
              PRINT,' '             
              PRINT,'----------------------------------------------------------'
              
              IF (N_ELEMENTS(hssd2_bot) EQ 0.) OR (N_ELEMENTS(hssd2_top) EQ 0) THEN BEGIN 
      
                  PRINT,' RESTORING EACH BEAM'
                  hssd_bot_name = output_dir+scan_date+'.'+scan_name+'.'+nos(k)+'.r'+nos_repeat(rn)+'.bot.hssd'
                  hssd_top_name = output_dir+scan_date+'.'+scan_name+'.'+nos(k)+'.r'+nos_repeat(rn)+'.top.hssd'
                  
                  RESTORE,hssd_bot_name,/VERB
                  RESTORE,hssd_top_name,/VERB
                  
              ENDIF
              
              slit_no = k
              
              ;;================================================================
              ;; COMBINE BEAMS HERE, CORRECT FOR I->X CROSSTALK, AND SCAN DRIFT
              ;;================================================================
              
              PRINT,' ALIGNING AND COMBINING THE POLARIZED BEAMS '
              hssd2 = FIRS_BEAM_COMBINE(hssd2_bot,hssd2_top,slit_no,config_str)
              hssd2_bot = 0.  ;; memory
              hssd2_top = 0.  ;; memory

              ;;================================================================
              ;; ACCOUNT FOR SPECTROGRAPH DRIFT (i.e. SCAN DRIFT).
              ;; --> TELLURIC LINES NEED TO BE USED FOR THIS PURPOSE.
              ;; --> NO TE LINES ARE AVAILABLE FOR 1565, THUS A COMPARISON TO
              ;;     THE SOLAR ROTATION VALUE WILL BE GIVEN
              ;;================================================================
              
              hssd2 = FIRS_SCAN_DRIFT(hssd2, slit_no, fskew_demod, frm_demod, hdr_flat,config_str)
              
              ;;================================================================
              ;; CREATE QUICKLOOK MAPS FOR HSSD STRUCTURE
              ;;================================================================
              
              hssd2 = FIRS_HSSD_MAPS(hssd2,config_str)
              
              ;;================================================================
              ;; PREPARE AND SAVE HSSD2 STRUCTURE
              ;;================================================================
              
              PRINT,'----------------------------------------------------------'
              PRINT,' '
              hssd_sav_name = output_dir+scan_date+'.'+scan_name+'.'+nos(k)+'.r'+nos_repeat(rn)+'.hssd
              PRINT,'Saving processed data in ' + hssd_sav_name+ ', please wait...'
              SAVE, hssd2, FILENAME = hssd_sav_name
              PRINT,' '
              PRINT,'----------------------------------------------------------'
              
          ENDFOR
          
      ENDFOR

  ENDIF
      
PRINT,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'
PRINT,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'
PRINT,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'
PRINT,'               REDUCTION COMPLETE                  '
PRINT,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'
PRINT,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'
PRINT,'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'

IF (config_str.journal_tag EQ 1) THEN JOURNAL

END

