function string2num,str,verbose = verbose,divider = divider

;+
;==============================================================================
;
;	function :  string2num
;
;	purpose  :  return the value of a string as double if possible
;                   takes whitespace, '/', ',' or ':' as separator
;
;       calls    : strsplit.pro, strtrim.pro, strcompress.pro,
;                  double.pro, where.pro (idl-implemented),
;                  length.pro, num2string.pro
;
;       written  :  cbeck@KIS 06/02
;
;==============================================================================
;
;	Check number of parameters.
;
;==============================================================================
if n_params() ne 1 then begin
	print
	print, "usage:  num = string2num(str, /verbose )"
	print
	print, "      Return content of a string as double precision floating."
        print, "      Returns vector, if more than one number seperated"
        print, "      by blanks or  /,;: is found."
        print, "      Attention : The number 42424242424242.42 will be cut"
        print, "      out of the string, so be careful with that number !!"
        print, "      If no number or ONLY '42424242424242.42' is found, the"
        print, "      the output is the string 'nonumber' ."
	print
        print, "	Arguments"
        print, "                Input :"
        print, "		str   : string, may contain more than one"
        print, "                        number."
        print
        print, "                Output:"
        print, "                num   : number/vector of numbers."
        print
        print, "        Keywords:"
        print, "                /verbose : give additional run time info"
        print
        print, "        Example: str = '12,324:55.34,abcd76 45' gives"
        print, "        res=[12,324,55.34,45]"
        print, "        while str = '12,324:55.34,76abcd 45' gives"
        print, "        res=[12,324,55.34,76,45]"
        print
	return,0
endif
;==============================================================================
;-

;==============================================================================
; check for substrings and store in erg, get number of substrings
;==============================================================================

ergg = 42424242424242.42

for k = 0,n_elements(str)-1 do begin
    str1 = str(k)

    if n_elements(divider) eq 0 then erg = strsplit(str1,' ;:/,_',/extract) else begin
        erg = strsplit(str1,divider,/extract)
    endelse
    n = n_elements(erg)
    erg1 = fltarr(n)
    if n_elements(verbose) eq 1 then  begin
        print,num2string(n)+' substring(s) found.'
    endif

;==============================================================================
; convert entries to double if possible
;==============================================================================
    for s =0,n-1 do begin
        valid = 0
        ON_IOERROR, bad_substring

        erg1(s)  = double(strtrim(strcompress(erg(s)),2))
        valid = 1

;==============================================================================
; remove bad substrings, i.e. not convertible to floating numbers
;==============================================================================
        bad_substring: if valid ne 1 then begin
            if n_elements(verbose) eq 1 then  begin
                print,'! string2num: bad substring ',erg(s),' removed !'
            endif    
            erg1(s) = '42424242424242.42'
        endif
        
    endfor

ff = where(erg1 ne 42424242424242.42)
if ff(0) eq -1 then erg1 = 'nonumber' else erg1 = erg1(ff)

;print,s,k
;print,erg1
;paused

if ff(0) ne -1 then ergg = [ergg,erg1]


endfor

ff = where(ergg ne 42424242424242.42)
if ff(0) ne -1 then ergg = ergg(ff) else  ergg = 'nonumber'


;==============================================================================
;	Return floating values of the substrings.
;==============================================================================
return, ergg

end
