pro wave_shift_firs,ia,ib,qb,ub,vb,ilo,ihi,firsty,lasty,repetitions,plt=plt,hairs = hairs,straight = straight,new = new

;+
;============================================================================
;
;	procedure : wave_shift
;
;	purpose  : calculate position of line cores for images ia and ib,
;	           fit polynomial of 2nd order on curve, shift all
;	           images of ib,qb,ub,vb for differences of the
;	           polynomials, repeat procedure 
; 
;       calls    : lpff.pro, sshift_img.pro
;
;       Hardwired: offsets of CA spectrally and spatially, line
;       64f. This has to be modified, if CA data is 'good',
;       i.e. hairline can be detected and spectral lines in the part
;       beams are not far displaced.
;
;	written  : cbeck@kis 4/03
;
;==============================================================================
;
;	Check number of arguments.
;
;==============================================================================
if (n_params() lt 10) then begin
	print
	print, "usage:  wave_shift,ia,ib,qb,ub,vb,ilo,ihi,firsty,lasty"
        print, "                  ,repetitions,plt=plt"
	print
	print, "	Shift all images (I,Q,U,V) to minimal difference of"
        print, "        line core postions of spectral line in I."
	print
	print, "	Arguments"
        print, "                Input :"
	print, "		ia,ib     : Stokes I intensity images A/B"
        print, "                qb,ub,vb  : polarization images of B"
        print, "                ilo,ihi   : range of spectral line to be"
        print, "                            used in aligning"
        print, "                firsty,lasty : useable range along slit"
        print, "                repetitions  : number of iterative repetitions"
        print, "                               to improve accuracy"
        print
        print, "        Keywords"
        print, "                plt      : plot line cores before and"
        print, "                           after shift"
        print, "                hairs    : if set, shift only once by the"
        print, "                           value of hairs(CA data)"
        print, "                straight : use straight line instead of"
        print, "                           polynomial of 2nd degree"
	print
        print, "        Additional : for CA first shift in spectral dimension"
        print, "                     is value of filt_stepsz from op. header"
        print
        return
endif
;==============================================================================
;-

llinerange = 5
llinerangeca = 3

;@polis_op_hdr.com
line_offset = 0


;if p_demod eq 2 then line_offset = filt_stepsz   ;12.
;if p_demod eq 2 and n_elements(hairs) ne 0 then line_offset = hairs ;-5.5
;if p_demod eq 1 and n_elements(new) ne 0 then line_offset = filt_stepsz

if n_elements(plt) ne 0 then begin
window,17
!p.multi = [2,1,2]
plot,ia(0:ihi+5,45),/ynozero,title=' I+ with range'
oplot,[ilo,ilo],[0,40000]
oplot,[ihi,ihi],[0,40000]
plot,ib(0:ihi+5,45),/ynozero,title=' I- with range'
oplot,[ilo,ilo]-line_offset,[0,40000]
oplot,[ihi,ihi]-line_offset,[0,40000]
!p.multi = 0
;paused
wait,1.5
endif

;--------------------------------------------------
; loop over number of repetitions
;--------------------------------------------------
for rep = 0,repetitions do begin 

if rep gt 0 then line_offset = 0

;--------------------------------------------------
; get position of O_2 line core in Stokes I A/B
;
; find minimum intensity inside range ilo:ihi
; apply lpff.pro around minimum
;--------------------------------------------------
    posa_range = fltarr(lasty-firsty+1)
    posb_range = fltarr(lasty-firsty+1)

    for k = firsty,lasty do begin
        gg = min(ia(ilo:ihi,k-firsty),posmin)
        posa_range(k-firsty) = posmin
        ggg = min(ib(ilo-line_offset:ihi-line_offset,k-firsty),posmin)
        posb_range(k-firsty) = posmin-line_offset

if n_elements(plt) ne 0 and k eq lasty/2 then begin
erase
!p.multi = [3,1,3]
plot,ia(ilo:ihi,k-firsty),/ynozero,title=' I+, row : '+num2string(lasty/2)
plot,ib(ilo-line_offset:ihi-line_offset,k-firsty),/ynozero,title=' I-, row : '+num2string(lasty/2)
plot,ib(*,k-firsty),/ynozero,title=' both, row : '+num2string(lasty/2)
oplot,[ilo-line_offset,ilo-line_offset],[0,20000]
oplot,[ihi-line_offset,ihi-line_offset],[0,20000]
oplot,ia(*,k-firsty),linestyle = 2
oplot,[ilo,ilo],[0,20000],linestyle = 2
oplot,[ihi,ihi],[0,20000],linestyle = 2
!p.multi = 0
;paused
wait,1.5
endif

    endfor
    posa = fltarr(lasty-firsty+1)
    posb = fltarr(lasty-firsty+1)

;--------------------------------------------------
; get line core position with lpff.pro
;--------------------------------------------------
    for k = firsty,lasty do begin

if n_elements(hairs) eq 0 then begin 
        
        lpff,ia(ilo+posa_range(k-firsty)-llinerange:ilo+posa_range(k-firsty)+llinerange,k-firsty),pos
        posa(k-firsty) = pos+posa_range(k-firsty)-llinerange
        
        lpff,ib(ilo+posb_range(k-firsty)-llinerange:ilo+posb_range(k-firsty)+llinerange,k-firsty),pos
        posb(k-firsty) = pos+posb_range(k-firsty)-llinerange
endif else begin
        lpff,ia(ilo+posa_range(k-firsty)-llinerangeca:ilo+posa_range(k-firsty)+llinerangeca,k-firsty),pos
        posa(k-firsty) = pos+posa_range(k-firsty)-llinerangeca
        
        lpff,ib(ilo+posb_range(k-firsty)-llinerangeca:ilo+posb_range(k-firsty)+llinerangeca,k-firsty),pos
        posb(k-firsty) = pos+posb_range(k-firsty)-llinerangeca
endelse
    endfor

;if tt(0) ne -1 then begin
;    for hh = 0,length(tt)-1 do begin
;        posb(tt(hh))= posb(tt(hh)-1)
;    endfor
;endif

gg  = where(finite(posa) eq 0)
gg1 = where(finite(posa) ne 0)

gg2 = where(finite(posb) eq 0)
gg3 = where(finite(posb) ne 0)

if gg(0)  ne -1 and gg1(0) ne -1 then posa(gg)  = mean(posa(gg1))
if gg2(0) ne -1 and gg3(0) ne -1 then posb(gg2) = mean(posb(gg3))

;-------------------------------------------------------------
; fit polynomial of 2nd degree to position of core along slit
;-------------------------------------------------------------
    range_fit = findgen(lasty-firsty+1)+firsty 

if n_elements(straight) eq 0 then begin
    erga = poly_fit(range_fit(1:lasty-firsty-1),posa(1:lasty-firsty-1)+ilo,5,/double)
    ergb = poly_fit(range_fit(1:lasty-firsty-1),posb(1:lasty-firsty-1)+ilo,5,/double)

    yy_a = erga(0)+erga(1)*range_fit+erga(2)*(range_fit)^2.+erga(3)*(range_fit)^3.+erga(4)*(range_fit)^4.+erga(5)*(range_fit)^5.
    yy_b = ergb(0)+ergb(1)*range_fit+ergb(2)*(range_fit)^2.+ergb(3)*(range_fit)^3.+ergb(4)*(range_fit)^4.+ergb(5)*(range_fit)^5.
endif

if n_elements(straight) ne 0 then begin
     erga = poly_fit(range_fit(1:lasty-firsty-1),posa(1:lasty-firsty-1)+ilo,1,/double)
     ergb = poly_fit(range_fit(1:lasty-firsty-1),posb(1:lasty-firsty-1)+ilo,1,/double)
     yy_a = erga(0)+erga(1)*range_fit
     yy_b = ergb(0)+ergb(1)*range_fit
endif

;--------------------------------------------------------------------
; get difference of line cores
;--------------------------------------------------------------------
    ydiff = yy_a-yy_b

;if rep ge 0 then begin
;ydiff = posa-posb
;kk = where(abs(ydiff) gt 10)
;if kk(0) ne -1 then ydiff(kk) = 0
;window,0
;plot,ydiff,/ynozero
;window,1
;!p.multi = [4,1,4]
;plot,posa,/ynozero
;plot,posb,linestyle=3,/ynozero
;plot,yy_a,linestyle=2,/ynozero
;plot,yy_b,linestyle=1,/ynozero
;paused
;endif


;if p_demod eq 2 and n_elements(hairs) ne 0 then ydiff = ydiff-ydiff+line_offset

;--------------------------------------------------------------------
; give warning if too high shift values occur
;--------------------------------------------------------------------
rr= mean(ydiff)
if rr gt 7.5 then begin   ; 3.5
    print,' wave_shift.pro: '
    print,' Warning ! Very high mean shift value of '+num2string(rr)+' pixels.'
    print,' Suggestion: check range of spectral lines/hairlines.'
    if n_elements(plt) eq 0 then begin
        print,' Do you want plotting to be switched on ?'
        ans = read_keyboard(' (y)es or (n)o :')
        if ans eq 'y' then plt = 1
    endif
endif

;--------------------------------------------------
; do plotting
;--------------------------------------------------
    if n_elements(plt) ne 0 then begin
        erase
        plot,range_fit,posa+ilo,/ynozero,title = 'Line cores A/B, shift : '+num2string(rep)
        oplot,range_fit,posb+ilo,linestyle = 2
        oplot,range_fit,yy_a
        oplot,range_fit,yy_b
        print,mean(ydiff),' mean difference of line cores in pixel'
        wait,1.5
;        paused
    endif

;-------------------------------------------------------------------------
; shift all images of camera B with the difference in line core positions
;-------------------------------------------------------------------------

sshift_img,ib,ydiff,firsty,lasty-1
sshift_img,qb,ydiff,firsty,lasty-1
sshift_img,ub,ydiff,firsty,lasty-1
sshift_img,vb,ydiff,firsty,lasty-1

endfor

;--------------------------------------------------
; display final result (same steps as above)
;--------------------------------------------------
if n_elements(plt) ne 0 then begin

    posa_range = fltarr(lasty-firsty+1)
    posb_range = fltarr(lasty-firsty+1)

    for k = firsty,lasty do begin
        gg = min(ia(ilo:ihi,k-firsty),posmin)
        posa_range(k-firsty) = posmin
        ggg = min(ib(ilo-line_offset:ihi-line_offset,k-firsty),posmin)
        posb_range(k-firsty) = posmin-line_offset
    endfor
    posa = fltarr(lasty-firsty+1)
    posb = fltarr(lasty-firsty+1)

;--------------------------------------------------
; get line core position with lpff.pro
;--------------------------------------------------
    for k = firsty,lasty do begin

if n_elements(hairs) eq 0 then begin 
        lpff,ia(ilo+posa_range(k-firsty)-llinerange:ilo+posa_range(k-firsty)+llinerange,k-firsty),pos
        posa(k-firsty) = pos+posa_range(k-firsty)-llinerange
        lpff,ib(ilo+posb_range(k-firsty)-llinerange:ilo+posb_range(k-firsty)+llinerange,k-firsty),pos
        posb(k-firsty) = pos+posb_range(k-firsty)-llinerange
endif else begin
        lpff,ia(ilo+posa_range(k-firsty)-llinerangeca:ilo+posa_range(k-firsty)+llinerangeca,k-firsty),pos
        posa(k-firsty) = pos+posa_range(k-firsty)-llinerangeca
        lpff,ib(ilo+posb_range(k-firsty)-llinerangeca:ilo+posb_range(k-firsty)+llinerangeca,k-firsty),pos
        posb(k-firsty) = pos+posb_range(k-firsty)-llinerangeca
endelse

    endfor

gg  = where(finite(posa) eq 0)
gg1 = where(finite(posa) ne 0)

gg2 = where(finite(posb) eq 0)
gg3 = where(finite(posb) ne 0)

if gg(0)  ne -1 then posa(gg)  = mean(posa(gg1))
if gg2(0) ne -1 then posb(gg2) = mean(posb(gg3))

;--------------------------------------------------
; fit polynomial to position of core along slit
;--------------------------------------------------
    range_fit = findgen(lasty-firsty+1)+firsty 

if n_elements(straight) eq 0 then begin
    erga = poly_fit(range_fit(1:lasty-firsty-1),posa(1:lasty-firsty-1)+ilo,5,/double)
    ergb = poly_fit(range_fit(1:lasty-firsty-1),posb(1:lasty-firsty-1)+ilo,5,/double)
    yy_a = erga(0)+erga(1)*range_fit+erga(2)*(range_fit)^2.+erga(3)*(range_fit)^3.+erga(4)*(range_fit)^4.+erga(5)*(range_fit)^5.
    yy_b = ergb(0)+ergb(1)*range_fit+ergb(2)*(range_fit)^2.+ergb(3)*(range_fit)^3.+ergb(4)*(range_fit)^4.+ergb(5)*(range_fit)^5.
endif
if n_elements(straight) ne 0 then begin
     erga = poly_fit(range_fit(1:lasty-firsty-1),posa(1:lasty-firsty-1)+ilo,1,/double)
     ergb = poly_fit(range_fit(1:lasty-firsty-1),posb(1:lasty-firsty-1)+ilo,1,/double)
     yy_a = erga(0)+erga(1)*range_fit
     yy_b = ergb(0)+ergb(1)*range_fit
endif

;--------------------------------------------------
; do plotting
;--------------------------------------------------
    plot,range_fit,posa+ilo,/ynozero,title = 'Line cores A/B, final shift'
    oplot,range_fit,posb+ilo,linestyle = 2
    oplot,range_fit,yy_a
    oplot,range_fit,yy_b,psym = 1
    print,mean(ydiff),' mean difference of line cores in pixel'
;    paused
    wait,1.5
endif

if n_elements(plt) ne 0 then begin
window,17
!p.multi = [2,1,2]
plot,ia(0:ihi+50,45),/ynozero
oplot,[ilo,ilo],[0,40000]
oplot,[ihi,ihi],[0,40000]
plot,ib(0:ihi+50,45),/ynozero
oplot,[ilo,ilo],[0,40000]
oplot,[ihi,ihi],[0,40000]
;paused
wait,1.5
!p.multi = 0
wdelete,17
endif

;--------------------------------------------------
; Done.
;--------------------------------------------------

end
